#include<iostream>
#include<fstream>
#include<cstdlib>
using namespace std;


#include"defo.h"
#include"zgfiles.h"
#include"coordc.h"
#include"multiint.h"


void LinCC::reset_to_null()
// sets dim=0 and deletes arrays describing  the transformation
{
  dim=0;
  delete [] origin_shift;
  origin_shift=NULL;
  delete [] m;
  m=NULL;
  delete [] mInv;
  mInv=NULL;

  delete [] intv_origin_shift;
  intv_origin_shift=NULL;
  delete [] intv_m;
  intv_m=NULL;
  delete [] intv_mInv;
  intv_mInv=NULL;
}

void LinCC::create_arrays()
{
  if(dim>0){
    origin_shift=new double[dim+1];
    m=new doublematrix1[dim+1];
    mInv=new doublematrix1[dim+1];

    intv_origin_shift=new interval[dim+1];
    intv_m=new intervalmatrix1[dim+1];
    intv_mInv=new intervalmatrix1[dim+1];
  } else{
    origin_shift=NULL;
    m=NULL;
    mInv=NULL;



    intv_origin_shift=NULL;
    intv_m=NULL;
    intv_mInv=NULL;
  }
}



LinCC::~LinCC()
{
  reset_to_null();
}

LinCC & LinCC::operator=(const LinCC & s)
// an copy/assignment operator
{
  if(dim != s.dim){
    reset_to_null();
    dim=s.dim;
    create_arrays();
  }
  copyvector(s.origin_shift,origin_shift,dim);
  copymatrix(s.m,m,dim);
  copymatrix(s.mInv,mInv,dim);

  copyvector(s.intv_origin_shift,intv_origin_shift,dim);
  copymatrix(s.intv_m,intv_m,dim);
  copymatrix(s.intv_mInv,intv_mInv,dim);
  return *this;
}


void LinCC::doubletointerval(int orthogonal)
{
  for(int i=1;i<=dim;i++)
    intv_origin_shift[i]=origin_shift[i];
  if(!orthogonal){
    for(int i=1;i<=dim;i++)
      for(int j=1;j<=dim;j++){
        intv_m[i][j]=m[i][j];
        intv_mInv[i][j]=mInv[i][j];
      }
  }
  if(orthogonal){
//    interval q[MAX_DIM+1][MAX_DIM+1];
    intervalmatrix1 *q;
    q=new intervalmatrix1[dim+1];
    for(int i=1;i<=dim;i++)
      for(int j=1;j<=dim;j++)
        q[i][j]=m[i][j];

    if(orthogonalize(q,dim,dim)){
      for(int i=1;i<=dim;i++)
        for(int j=1;j<=dim;j++){
          intv_m[i][j]=q[i][j];
          intv_mInv[i][j]=q[j][i];
        }
    } else{
      #if defined(WIN95) && defined(KRAK)
        rootFrame.Clear();
      #endif
      cerr << "doubletointerval - unable to orthogonalize a matrix"  << endl;
      #if defined(WIN95) && defined(KRAK)
        waitBt();
      #endif
      exit(EXIT_FAILURE);
    }
    delete [] q;
  }
}

void LinCC::intervaltodouble()
{
  int i;
  for(i=1;i<=dim;i++)
    origin_shift[i]=mid(intv_origin_shift[i]).right_b();
  for(i=1;i<=dim;i++)
    for(int j=1;j<=dim;j++){
      m[i][j]=mid(intv_m[i][j]).right_b();
      mInv[i][j]=mid(intv_mInv[i][j]).right_b();
    }

}


void LinCC::shiftorigin(const interval IP[])
// function updates the origin intv_origin_shift=IP
{
  for(int i=1;i<=dim;i++)
    intv_origin_shift[i]=IP[i];
  intervaltodouble();
}

LinCC::LinCC(int ldim,const interval IP[],const interval IM[][MAX_DIM+1],
          const interval InvIM[][MAX_DIM+1])
// initiate structure with interval data
{
  dim=ldim;
  create_arrays();
  for(int i=1;i<=dim;i++)
    intv_origin_shift[i]=IP[i];

  for(int i=1;i<=dim;i++)
    for(int j=1;j<=dim;j++){
      intv_m[i][j]=IM[i][j];
      intv_mInv[i][j]=InvIM[i][j];
    }
  intervaltodouble();
  if(!testinverse()){
    cerr << "LinCC::LinCC - inverse matrix test failed" << endl;
    // WARNING. we prefer not to stop the programm, because sometimes we just want the computations to continue
    #if defined(WIN95) && defined(KRAK)
      waitBt();
    #endif
  }
}


LinCC::LinCC(int ldim,const interval IM[][MAX_DIM+1],
         const interval InvIM[][MAX_DIM+1])
// initiates structure with interval data
// P(=origin_shift)  is assumed to be 0
{
  dim=ldim;
  create_arrays();
  for(int i=1;i<=dim;i++)
    intv_origin_shift[i]=0.0;

  for(int i=1;i<=dim;i++)
    for(int j=1;j<=dim;j++){
      intv_m[i][j]=IM[i][j];
      intv_mInv[i][j]=InvIM[i][j];
    }
  intervaltodouble();
  if(!testinverse()){
    cerr << "LinCC::LinCC - inverse matrix test failed" << endl;
     // WARNING. we prefer not to stop the programm, because sometimes we just want the computations to continue
    #if defined(WIN95) && defined(KRAK)
      waitBt();
    #endif
  }
}


LinCC::LinCC(int ldim,const interval IP[])
//  M is assumed to be identity matrix
{
  dim=ldim;
  create_arrays();
  for(int i=1;i<=dim;i++)
    intv_origin_shift[i]=IP[i];

  for(int i=1;i<=dim;i++)
    for(int j=1;j<=dim;j++){
      if(i==j){
        intv_m[i][j]=1.0;
        intv_mInv[i][j]=1.0;
      }
      else{
        intv_m[i][j]=0.0;
        intv_mInv[i][j]=0.0;
      }
    }
  intervaltodouble();
}


LinCC::LinCC(int ldim,const double P[])
//  M is assumed to be identity matrix
{
  dim=ldim;
  create_arrays();
  int orthogonal=1;
  for(int i=1;i<=dim;i++)
    origin_shift[i]=P[i];

  for(int i=1;i<=dim;i++)
    for(int j=1;j<=dim;j++){
      if(i==j){
        m[i][j]=1.0;
        mInv[i][j]=1.0;
      }
      else{
        m[i][j]=0.0;
        mInv[i][j]=0.0;
      }
    }
  doubletointerval(orthogonal);
}





LinCC::LinCC()
{
  dim=0;
  create_arrays();
}

LinCC::LinCC(int ldim)
// the identity transformation
{
  dim=ldim;
  create_arrays();
  for(int i=1; i<=dim; i++){
    origin_shift[i]=0.0;
    intv_origin_shift[i]=0.0;
  }
  identity(m,dim);
  identity(intv_m,dim);
  identity(mInv,dim);
  identity(intv_mInv,dim);
}


LinCC::LinCC(const LinCC & lincc)
{
  dim=lincc.dim;
  create_arrays();
  copyvector(lincc.origin_shift,origin_shift,dim);
  copymatrix(lincc.m,m,dim);
  copymatrix(lincc.mInv,mInv,dim);

  copyvector(lincc.intv_origin_shift,intv_origin_shift,dim);
  copymatrix(lincc.intv_m,intv_m,dim);
  copymatrix(lincc.intv_mInv,intv_mInv,dim);
}


double LinCC::coord(const double x[],int i) const
 // y=m*(x - origin_shift)
{
  double ret=0;
  for(int j=1;j<=dim; j++) ret+=m[i][j]*(x[j]-origin_shift[j]);
  return(ret);
}


interval LinCC::coord(const interval x[],int i) const
 // y=m*(x - origin_shift)
{
  interval ret=0;
  for(int j=1;j<=dim; j++) ret+=intv_m[i][j]*(x[j]-intv_origin_shift[j]);
  return(ret);
}



void LinCC::coordtran(const double x[],double y[]) const
   // y=m*(x-origin_shift)
{
  double yt[MAX_DIM+1];   // we use this additional variable to be able
                         // to call this function with x and y pointing on the same
                         // table
  for(int i=1;i<=dim; i++){
    double t=0.0;
    for(int j=1;j<=dim;j++)
      t+=m[i][j]*(x[j]-origin_shift[j]);

    yt[i]=t;
  }
  for(int i=1;i<=dim;i++)
    y[i]=yt[i];
}

void LinCC::coordtran(const interval x[],interval y[]) const
   // y=m*(x-origin_shift)
{
  interval yt[MAX_DIM+1];   // we use this additional variable to be able
                         // to call this function x and y pointing on the same
                         // table
  for(int i=1;i<=dim; i++){
    interval t=0.0;
    for(int j=1;j<=dim;j++)
      t+=intv_m[i][j]*(x[j]-intv_origin_shift[j]);
    yt[i]=t;
  }
  for(int i=1;i<=dim;i++)
    y[i]=yt[i];
}

void LinCC::coordinvtran(const double y[],double x[]) const
   // x=mInv*y + origin_shift
{
  double xt[MAX_DIM+1];   // we use this additional variable to be able
                         // to call this function with x and y pointing on the same
                         // table
  for(int i=1;i<=dim;i++){
    double t=0.0;
    for(int j=1;j<=dim; j++)
      t+=mInv[i][j]*y[j];
    xt[i]=t + origin_shift[i];
  }
  for(int i=1;i<=dim;i++)
    x[i]=xt[i];
}

void LinCC::coordinvtran(const interval y[],interval x[]) const
   // x=mInv*y + origin_shift
{
  interval xt[MAX_DIM+1];   // we use this additional variable to be able
                         // to call this function with x and y pointing on the same
                         // table
  for(int i=1;i<=dim;i++){
    interval t=0.0;
    for(int j=1;j<=dim; j++)
      t+=intv_mInv[i][j]*y[j];
    xt[i]=t + intv_origin_shift[i];
  }
  for(int i=1;i<=dim;i++)
    x[i]=xt[i];
}

void LinCC::coordtran(const interval x[],const interval Cx[][MAX_DIM+1],
       const interval rx[],interval y[],interval Cy[][MAX_DIM+1],interval ry[]) const
     // the transformation of the set x+Cx*rx in old coordinates
     //  on output the set in new coordinates is given by
     //  y+Cy*ry = m(x - origin_shift) + m*Cx*rx
     //   y=m(x-origin_shift),  Cy=m*Cx ,   ry=rx
{
  coordtran(x,y); // y=m(x-origin_shift)
  ArrArr(intv_m,Cx,Cy,dim);
  copyvector(rx,ry,dim);
}




void LinCC::coordinvtran(const interval y[],const interval Cy[][MAX_DIM+1],
      const interval ry[],
       interval x[],interval Cx[][MAX_DIM+1],interval rx[]) const
     // the transformation of the set y+Cy*ry in new coordinates
     //  on output the set in old coordinates is given by
     //   x+Cx*rx=mInv*y + origin_shift + mInv*Cy*ry
     //    x= mInv*y + origin_shift,   Cx=mInv*Cy,  rx=ry
{
  coordinvtran(y,x);    // x=mInv*y + origin_shift
  ArrArr(intv_mInv,Cy,Cx,dim);
  copyvector(ry,rx,dim);
}






   // transformation of tangent vectors: from old to new
void LinCC::vectcoordtran(const double x[],double y[]) const
   // y=m*x
{
  multAV(m,x,y,dim);
}
void LinCC::vectcoordtran(const interval x[],interval y[])  const
   // y=m*x
{
  multAV(intv_m,x,y,dim);
}


   // transformation of tangent vectors: from new to old
void LinCC::vectcoordinvtran(const double y[],double x[]) const
  // x=mInv*y
{
  multAV(mInv,y,x,dim);
}

void LinCC::vectcoordinvtran(const interval y[],interval x[]) const
// x=mInv*y
{
  multAV(intv_mInv,y,x,dim);
}



void LinCC::printCC(const char name[]) const
  // procedure prints to file coordinate change data
{
  ofstream report(name);
  int i;
  report << "dim=" << dim << endl;
  report << "y=M(x - P)" << endl;
  report << "Translation vector - P" << endl;
  for(i=1;i<=dim; i++)
    report << origin_shift[i] << "  " ;
  report << endl << endl;;


  report << "Matrix M" << endl;
  for(i=1;i<=dim; i++){
    for(int j=1; j<=dim; j++)
      report << m[i][j] << "  ";
    report << endl;
  }


  report << endl << endl;;


  report << "Matrix M^\\{-1\\}" << endl;
  for(i=1;i<=dim; i++){
    for(int j=1; j<=dim; j++)
      report << mInv[i][j] << "  ";
    report << endl;
  }



}



LinCC LinCC::inverse() const
{
  LinCC invlincc(dim);
  int i;
  for(i=1;i<=dim;i++){
    double t=0.0;
    interval It=0.0;
    for(int j=1;j<=dim;j++){
      t+=m[i][j]*origin_shift[j];
      It+=intv_m[i][j]*intv_origin_shift[j];
    }
    (invlincc.origin_shift)[i]=-t;
    (invlincc.intv_origin_shift)[i]=-It;
  }

  for(i=1;i<=dim; i++)
    for(int j=1; j<=dim; j++){
      invlincc.m[i][j]=mInv[i][j];
      invlincc.mInv[i][j]=m[i][j];

      invlincc.intv_m[i][j]=intv_mInv[i][j];
      invlincc.intv_mInv[i][j]=intv_m[i][j];

    }

  return(invlincc);
}


void permutecoordc(LinCC *lincc,int per[])
// the procedure permutes the variables in the <lincc>
// the permutation is given as a new ordering  of variables(per[1],per[2],...)
//  for example: per={2,3,....}
//      means 2-nd coordinate become 1
//            3-rd become second and so on
// on output : the coordinate change works as follows:
// y_i (new) = y_{per[i]}(old)
{

  // a preparation of the permutation matrix
  int mper[MAX_DIM+1][MAX_DIM+1];
  int imper[MAX_DIM+1][MAX_DIM+1];
  for(int i=1;i<=lincc->dim;i++)
    for(int j=1; j<= lincc->dim;j++){
      mper[i][j]=0;
      imper[i][j]=0;
    }

  for(int i=1;i<=lincc->dim;i++){
    mper[i][per[i]]=1;
    imper[per[i]][i]=1;
  }
  // end of preparation


  // y(new)=mper * y(old)=mper* m * (x - origin_shift)
  //   hence y(new)= (mper * m) *(x-przes)
  //   x= mac^{-1}*imper y(new) + origin_shift

  double mtemp[MAX_DIM+1][MAX_DIM+1];
//  interval imtemp[][MAX_DIM+1];
  intervalmatrix1 *imtemp;
  imtemp=new intervalmatrix1[lincc->dim+1];

  // transforming of lincc->m
  for(int i=1;i<=lincc->dim;i++)
    for(int j=1;j<=lincc->dim;j++)
      for(int k=1;k<=lincc->dim;k++){
        if(mper[i][k]==1){
          mtemp[i][j]=(lincc->m)[k][j];
          imtemp[i][j]=(lincc->intv_m)[k][j];
        }
      }

  for(int i=1;i<=lincc->dim;i++)
    for(int j=1;j<=lincc->dim; j++){
      (lincc->m)[i][j]=mtemp[i][j];
      (lincc->intv_m)[i][j]=imtemp[i][j];
    }




  // transforming  lincc->mInv
  for(int i=1;i<=lincc->dim;i++)
    for(int j=1;j<=lincc->dim;j++)
      for(int k=1;k<=lincc->dim;k++){
        if(imper[k][j]==1){
          mtemp[i][j]=(lincc->mInv)[i][k];
          imtemp[i][j]=(lincc->intv_mInv)[i][k];
        }
      }

  for(int i=1;i<=lincc->dim;i++)
    for(int j=1;j<=lincc->dim; j++){
      (lincc->mInv)[i][j]=mtemp[i][j];
      (lincc->intv_mInv)[i][j]=imtemp[i][j];
    }

  delete [] imtemp;
}

int LinCC::testinverse() const
  // function returns 1 - if the identify matrix \subset intv_m*intv_mInv
  //                               and  (reversed order) the identify matrix \subset intv_mInv*intv_m
  //                 0 - otherwise
{
  int ret=1;

  for(int i=1;i<=dim;i++)
    for(int j=1;j<=dim;j++){
      interval t=0.0;
      for(int k=1;k<=dim;k++)
        t+=intv_m[i][k]*intv_mInv[k][j];
      if(i==j)
        if(!t.contains(1.0)) ret=0;
      if(i!=j)
        if(!t.contains(0.0)) ret=0;
    }


  for(int i=1;i<=dim;i++)
    for(int j=1;j<=dim;j++){
      interval t=0.0;
      for(int k=1;k<=dim;k++)
        t+=intv_mInv[i][k]*intv_m[k][j];

      if(i==j)
        if(!t.contains(1.0)) ret=0;
      if(i!=j)
        if(!t.contains(0.0)) ret=0;
    }

  return(ret);
}

int LinCC::savebin(FILE *fbin,int whatsave) const
  // saves the coordinate change to a binary file
  // whatsave=0; only double's are saved
  //         =1; only intervals
  //         =2; everything
  //// structure of the file:
  //  whatsave,dim,(doubles:origin_shift,m,mInv),(intervals:intv_origin_shift,intv_m,intv_mInv)
{
  fwrite(&whatsave,sizeof(int),1,fbin);
  fwrite(&dim,sizeof(int),1,fbin);
  if((whatsave==0) || (whatsave==2)){  // saving doubles
    fwrite(origin_shift+1,sizeof(double),dim,fbin);

    // saving m
    for(int i=1;i<=dim;i++)
      fwrite(m[i]+1,sizeof(double),dim,fbin);

    // saving mInv
    for(int i=1;i<=dim;i++)
      fwrite(mInv[i]+1,sizeof(double),dim,fbin);
  }

  if((whatsave==1) || (whatsave==2)){  // saving intervals
    fwrite(intv_origin_shift+1,sizeof(interval),dim,fbin);

    // saving intv_m
    for(int i=1;i<=dim;i++)
      fwrite(intv_m[i]+1,sizeof(interval),dim,fbin);

    // saving intv_mInv
    for(int i=1;i<=dim;i++)
      fwrite(intv_mInv[i]+1,sizeof(interval),dim,fbin);
  }
  return 1;
}

int LinCC::savebin(const char *filename,int whatsave) const
{
  FILE* fhandle;
  fhandle=fopen(filename,"wb");
  int ret=(fhandle!=NULL);
  if(ret){
    ret=savebin(fhandle,whatsave);
    fclose(fhandle);
  }else{
    cerr << "cannot open " << filename << endl;
    exit(EXIT_FAILURE);
  }
  return ret;
}


int LinCC::readbin(FILE *fbin,bool recomp_inverse /*=false*/)
 // reads (really reinitiates) coordinate change from the file <fbin>
   // if  recomp_inverse, then the inverse in *plincc is recomputed
   //  to make sure that computation is rigorous, can be of use, when
   // computing things first on the computer on which we don't have correct rounding
{
  reset_to_null();

  int whatsave;
  fread(&whatsave,sizeof(int),1,fbin);
  fread(&dim,sizeof(int),1,fbin);
  create_arrays();
  if((whatsave==0) || (whatsave==2)){  // reading doubles
    fread(origin_shift+1,sizeof(double),dim,fbin);

    // reading m
    for(int i=1;i<=dim;i++)
      fread(m[i]+1,sizeof(double),dim,fbin);

    // reading mInv
    for(int i=1;i<=dim;i++)
      fread(mInv[i]+1,sizeof(double),dim,fbin);
  }

  if((whatsave==1) || (whatsave==2)){  // reading intervals
    fread(intv_origin_shift+1,sizeof(interval),dim,fbin);

    // reading intv_m
    for(int i=1;i<=dim;i++)
      fread(intv_m[i]+1,sizeof(interval),dim,fbin);

    // reading intv_mInv
    for(int i=1;i<=dim;i++)
      fread(intv_mInv[i]+1,sizeof(interval),dim,fbin);
  }

  if(whatsave==1) intervaltodouble();
  if(whatsave==0) doubletointerval(1);

  if(recomp_inverse){
    if(!precond_inverse(intv_m,intv_mInv,dim)){
       cerr << "can't compute the inverse "<< endl;
       #if defined(KRAK) && defined(WIN95)
       waitBt();
       #endif
       exit(EXIT_FAILURE);
    }
    intervaltodouble();
  }

  return 1;
}

int LinCC::readbin(const char *filename,bool recomp_inverse/*=false*/)
 // reads (really reinitiates) coordinate change from the file <fbin>
  // if  recomp_inverse, then the inverse in *plincc is recomputed
   //  to make sure that computation is rigorous, can be of use, when
   // computing things first on the computer on which we don't have correct rounding
{
  FILE* fhandle;
  fhandle=fopen(filename,"rb");
  int ret=(fhandle!=NULL);
  if(ret){
    ret=readbin(fhandle,recomp_inverse);
    fclose(fhandle);
  }else{
    cerr << "cannot open " << filename << endl;
      #if defined(KRAK) && defined(WIN95)
           waitBt();
       #endif
    exit(EXIT_FAILURE);
  }
  return ret;
}


int LinCC::savetext(ostream &ftext,int whatsave) const
  // saves the coordinate change to a text file
  // whatsave=0; only double's are saved
  //         =1; only intervals
  //         =2; everything
  //// structure of the file:
  //  whatsave,dim,(doubles:origin_shift,m,mInv),(intervals:intv_origin_shift,intv_m,intv_mInv)
{

  int writeint=1; // 1   -interval ; 2 - center + interval

  ftext << whatsave << " what is saved: ";
  if(whatsave==0)
    ftext << " only doubles are saved ";
  if(whatsave==1)
    ftext << " only intervals are saved ";
  if(whatsave==2)
    ftext << " doubles and intervals saved ";
  ftext << endl;

  ftext << dim <<   "  dimension " << endl;
  if((whatsave==0) || (whatsave==2)){  // saving doubles
    ftext << "shift - double " << endl;
    saveAllCoord(ftext,origin_shift,dim,0);

    ftext << "m - transition from old to new coordinates " << endl;
    saveArray(ftext,m,dim);


    ftext << "mInv - transition from  new  to old coordinates " << endl;
    saveArray(ftext,mInv,dim);
  }

  if((whatsave==1) || (whatsave==2)){  // saving intervals
    ftext << "shift - interval " << endl;
    saveAllCoord(ftext,intv_origin_shift,dim,writeint);

    ftext << "m - transition from old to new coordinates " << endl;
    saveArray(ftext,intv_m,dim,writeint);


    ftext << "mInv - transition from  new  to old coordinates " << endl;
    saveArray(ftext,intv_mInv,dim,writeint);
  }

  return 1;
}

int LinCC::readtext(istream &ftext)
  // reads (really reinitiates) coordinate change from the file <fbin>
{
  char trash[240];

  int whatsave;
  ftext >> whatsave;
  ftext.getline(trash,240); // to eat endl

  ftext >> dim ;
  ftext.getline(trash,240); // to eat endl
  if((whatsave==0) || (whatsave==2)){  // reading doubles
    ftext.getline(trash,240); //  to eat "shift - double "
    readvector(ftext,origin_shift,dim);
    ftext.getline(trash,240); // to eat endl

    ftext.getline(trash,240); // to eat  "m - transition from old to new coordinates " << endl;
    readmatrix(ftext,m);


    ftext.getline(trash,240); // to eat  "mInv - transition from  new  to coordinates " << endl;
    readmatrix(ftext,mInv);
  }

  if((whatsave==1) || (whatsave==2)){  //  reading intervals
    ftext.getline(trash,240); //  to eat "shift - interval "
    readvector(ftext,intv_origin_shift,dim);
    ftext.getline(trash,240); // to eat endl

    ftext.getline(trash,240); // to eat  "m - transition from old to new coordinates " << endl;
    readmatrix(ftext,intv_m);


    ftext.getline(trash,240); // to eat  "mInv - transition from  new  to coordinates " << endl;
    readmatrix(ftext,intv_mInv);
  }

  if(whatsave==1) intervaltodouble();
  if(whatsave==0) doubletointerval(1);

  return 1;
}



void transLinearPart(const interval lin[][MAX_DIM+1],const LinCC &linccIn,
      const LinCC &linccOut,     interval translin[][MAX_DIM+1])
// this function does coordinate transformation of linear map
//  <lin>, in new coordinates on input - given by linccIn
//  and output - linccOut
//  hence it returns the linear matrix reprsenting the map
//       linccOut \circ lin \circ linccIn^{-1}
//
//  This function makes sense for transforming derivative matrix of Poincare
//  map

{
//  translin=(linccOut.m)*lin*(linccIn.mInv)
  intervalmatrix1 *temp;
  temp=new intervalmatrix1[linccIn.dim+1];


  ArrArr(linccOut.intv_m,lin,temp,linccIn.dim);
  ArrArr(temp,linccIn.intv_mInv,temp,linccIn.dim);
  copymatrix(temp,translin,linccIn.dim);

  delete [] temp;
}

void transLinearPart(const double lin[][MAX_DIM+1],const LinCC &linccIn,
     const LinCC &linccOut,     double translin[][MAX_DIM+1])
// this function does coordinate transformation of linear map
//  <lin>, in new coordinates on input - given by linccIn
//  and output - linccOut
//  hence it returns the linear matrix representing the map
//       linccOut \circ lin \circ linccIn^{-1}
//  This function makes sense for transforming derivative matrix of Poincare
//  map
{
//  translin=(linccOut.m)*lin*(linccIn.mInv)
  doublematrix1 *temp;
  temp=new doublematrix1[linccIn.dim+1];



  ArrArr(linccOut.m,lin,temp,linccIn.dim);
  ArrArr(temp,linccIn.mInv,temp,linccIn.dim);
  copymatrix(temp,translin,linccIn.dim);

  delete [] temp;
}


void transLinearPart(const double lin[][MAX_DIM+1],const LinCC &lincc,
     double translin[][MAX_DIM+1])
// this function does coordinate transformation of linear map
//  <lin>, On output <translin> contains <lin> in new coordinates
//  the fact that <lin> in new coordinates might be not a linear
//  map (this is the case when the origin s shifted) is ignored
//  This function makes sense for transforming derivative matrix of Poincare
//  map  or the derivative of the vector field around fixed point
{

  //  translin=(lincc.m)*lin*(lincc.mInv)
  transLinearPart(lin,lincc,lincc,translin);
}


void transLinearPart(const interval lin[][MAX_DIM+1],const LinCC &lincc,
     interval translin[][MAX_DIM+1])
// this function does coordinate transformation of linear map
//  <lin>, On output <translin> contains <lin> in new coordinates
//  the fact that <lin> in new coordinates might be not a linear
//  map (this is the case when the origin s shifted) is ignored
//  This function makes sense for transforming derivative matrix of Poincare
//  map  or the derivative of the vector field around fixed point

{

  //  translin=(lincc.m)*lin*(lincc.mInv)
   transLinearPart(lin,lincc,lincc,translin);
}



LinCC* insertcoord(const LinCC *coordch,int sel_coord,interval sel_coord_value)
// this function creates a coordinate transformation in the dimension
//  coordch->dim + 1, by inserting sel_coord, and seting origin for
//  this coordinate to <sel_coord_value>
{

  int ldim=coordch->dim+1;
  interval IP[MAX_DIM+1];
  intervalmatrix1 *IM;
  IM=new intervalmatrix1[ldim+1];

  intervalmatrix1 *InvIM;
  InvIM=new intervalmatrix1[ldim+1];

  insertcoord(coordch->intv_origin_shift,sel_coord,IP,sel_coord_value,ldim);
  insertcoord(coordch->intv_m,sel_coord,IM,ldim);
  IM[sel_coord][sel_coord]=1.0;

  insertcoord(coordch->intv_mInv,sel_coord,InvIM,ldim);
  InvIM[sel_coord][sel_coord]=1.0;


  LinCC *ncc=new LinCC(ldim,IP,IM,InvIM);

  delete [] IM;
  delete [] InvIM;

  return ncc;
}



LinCC* LinCCincreasedim(LinCC *coordch,int newdim)
// this function creates a coordinate transformation in dimension
//   newdim - by adding identity on coordinates i=coordch->dim+1,\dots,newdim
{
  if(newdim <= coordch->dim){
    cerr << "LinCCincreasedim - new dimension is too small"  << endl;
    #if defined(WIN95) && defined(KRAK)
      waitBt();
    #endif
    exit(EXIT_FAILURE);
  }

  int ldim=newdim;
  interval *IP;
  IP=new interval[ldim+1];
  intervalmatrix1 *IM;
  IM=new intervalmatrix1[ldim+1];

  intervalmatrix1 *InvIM;
  InvIM=new intervalmatrix1[ldim+1];

  zero(IP,ldim);
  identity(IM,ldim);
  identity(InvIM,ldim);
  for(int i=1;i<=coordch->dim;i++)
    IP[i]=coordch->intv_origin_shift[i];


  for(int i=1;i<=coordch->dim;i++){
    for(int j=1;j<=coordch->dim;j++){
      IM[i][j]=coordch->intv_m[i][j];
      InvIM[i][j]=coordch->intv_mInv[i][j];
    }
  }

  LinCC *ncc=new LinCC(ldim,IP,IM,InvIM);

  delete [] IM;
  delete [] InvIM;
  delete [] IP;
  return ncc;
}


LinCC *genLinCCfd(int pdim,double M[][MAX_DIM+1],int thisIsInvM)
// creates a new LinCC - with P=0; and rigorously computed M^{-1}
//   if thisIsInvM, then the parameter M - is the inverse
//   of the transition from old to new coordinates
{
  LinCC *ret=NULL;
  intervalmatrix1 *IM=new intervalmatrix1[pdim+1];
  intervalmatrix1 *invIM=new intervalmatrix1[pdim+1];
  tointerval(M,IM,pdim);

  int ok=inverse(IM,invIM,pdim);
  if(ok){
    if(thisIsInvM)
      ret=new LinCC(pdim,invIM,IM);
    else
      ret=new LinCC(pdim,IM,invIM);
  }

  delete [] invIM;
  delete [] IM;

  return ret;
}




LinCC* composeLinCC(const LinCC *c1,const LinCC *c2)
// the function creates a new change of coordinates c1 \circ c2
{


  if(c1->dim!=c2->dim){
    cerr << "composeLinCC - dimensions don't match" << endl;
    #if defined(WIN95) && defined(KRAK)
      waitBt();
    #endif
    exit(EXIT_FAILURE);
  }

  int ldim=c1->dim;


  interval *IP;
  IP=new interval[ldim+1];
  intervalmatrix1 *IM;
  IM=new intervalmatrix1[ldim+1];

  intervalmatrix1 *InvIM;
  InvIM=new intervalmatrix1[ldim+1];



  // computing the new shift: observe that for an affine transformation y=M(x-P)
  // x=M^{-1}y + P, hence P=(inverse transformation)(0)
  //// in our case the transformation is c1 \circ c2, hence the inverse is
  //     c2^{-1} \circ c1^{-1}
  for(int i=1;i<=ldim;i++)
    IP[i]=0.0;
  c1->coordinvtran(IP,IP); // IP=c1^{-1}(0)
  c2->coordinvtran(IP,IP);  // IP=c2^{-1}\circ c1^{-1}(0)


  ArrArr(c1->intv_m,c2->intv_m,IM,ldim);
  ArrArr(c2->intv_mInv,c1->intv_mInv,InvIM,ldim);


  LinCC *ncc=new LinCC(ldim,IP,IM,InvIM);

  delete [] IM;
  delete [] InvIM;
  delete [] IP;

  return ncc;
}


int LinCC_thesame(const LinCC *pcc1,const LinCC *pcc2,int  whattocheck)
// 1 - is returned when the coordinate changes are the same
// 0 - otherwise
//   if whattocheck==0 - we check doubles
//                1 - we check intervals
//                2 - we check both doubles and intervals
{
  int thesame=(pcc1->dim==pcc2->dim);
  if(thesame){
    if((whattocheck==0) || (whattocheck==2)){ // we check doubles
      for(int i=1;i<=pcc1->dim;i++){
        thesame=thesame && (pcc1->origin_shift[i]==pcc2->origin_shift[i]);
      }
      for(int i=1;i<=pcc1->dim;i++){
        for(int j=1;j<=pcc1->dim;j++){
          thesame=thesame && (pcc1->m[i][j]==pcc2->m[i][j]);
          thesame=thesame && (pcc1->mInv[i][j]==pcc2->mInv[i][j]);
        }
      }
    }

    if((whattocheck==1) || (whattocheck==2)){ // we check interval
      for(int i=1;i<=pcc1->dim;i++){
        thesame=thesame && (pcc1->intv_origin_shift[i]==pcc2->intv_origin_shift[i]);
      }
      for(int i=1;i<=pcc1->dim;i++){
        for(int j=1;j<=pcc1->dim;j++){
          thesame=thesame && (pcc1->intv_m[i][j]==pcc2->intv_m[i][j]);
          thesame=thesame && (pcc1->intv_mInv[i][j]==pcc2->intv_mInv[i][j]);
        }
      }
    }

  }
  return thesame;
}
