#include<iostream>
#include<fstream>
#include<cstdlib>
using  namespace std;

#include"defo.h"
#include"coordc.h"
#include"diffincl.h"
#include"zdynsys.h"
#include"rigpoinc.h"
#include"multiint.h"
#include"lohner.h"
#include"zset.h"
#include"section.h"
#include"inclpoin.h"
#include"fnames.h"
#include"zgfiles.h"


int incl_OneStepLohner(InclOde &inclode,const InclOneStepParamType &onesteppar,
       interval h,Set &x,interval xd[], SetWithTail & y, Set &xn, interval xdn[],
       SetWithTail &yn,interval xencl[],SetWithTail &yencl, Set &xs,
        interval xds[],interval yc[],interval w1_encl[],interval rem[])
// this function realizes one time step of the system
//  x' \in P_m F(x,y)
//  y - is submitted to separate evolution
///////////////
//  for perturbations of ODE we have x' \in F(x) + y
//            where y is a set and is not changing with the time
//////////////////
//  for KS with evolving tail
//   y - is the structure containing, the Galerkin errors i.e. a table[1...dim]
//    and the tail  T which we also alow to evolve
// for the main variables: we solve x' \in P_m F(x) + galerrors (Tail)
//    but we evolve the Tail (pointed by y.ptail)
/////////////////
//  h - time step
// on INPUT :  xd=interval enclosure of (x)
// on OUTPUT :
//     \varphi(h,(x,y)) \subset xn \times yn  \subset xdn \times yn
//    \varphi([0,h],(x,y)) \subset xencl \times yencl
//    \Pi_x \overline{\varphi}([0,h],x,yc)  \subset w1_encl  - this is
//             for the deterministic part with a fixed y=yc
//    \varphi(h,(x,yc)) \subset xs \subset xds - i.e. this an image for
//              'deterministic eq'
//    yc - the value of y_c used during this time step; it is later
//         used, when in the section region we estimate the difference
//          between the segment joining x and xn the the orbit
//
// rem; // remainder term in Taylor method, before the multiplication
//         by h^(order+1)
{
  // interval prev_yc[MAX_DIM+1];
  interval* prev_yc;
  prev_yc=new interval[inclode.dim+1];
  inclode.current_yc(prev_yc);


  int do_not_sent_bad_info=0;



  // 1. enclosure for x and y
  int ok=inclode.incl_enclosure(xd,y,h,xencl,yencl,do_not_sent_bad_info);
          // w2=xencl,   wy=yencl
          // xencl \times yencl - the enclosure for (x,y)[0,h]
  if(ok){

    // computation of a new <yn>
    ok=inclode.y_move(xencl,yencl,h,y,yn);

    // we choose y_c
    inclode.set_ycenter(yencl.y);
    inclode.current_yc(yc);  // yc=inclode.current_yc
  }
  if(ok){
    //  integration of x'=f(x)+y_c;    with fixed y=y_c
    if(onesteppar.onestepmethod==dfincl_C0simpleLo){
      ok=OneStepLohner(inclode,onesteppar,h,x,xd,xn,xdn,w1_encl,rem);
    }else{
      if(onesteppar.onestepmethod==dfincl_C0thruC1Lo){
        ok=C0throughC1OneStepLohner(inclode,onesteppar,h,x,xd,xn,xdn,w1_encl,rem);
      }else{
        cerr << "unknown one step method in [incl_OneStepLohner] " << endl;
        #if defined(KRAK) && defined(WIN95)
        waitBt();
        #endif
        exit(EXIT_FAILURE);
      }
    }
    if(ok){
      xs=xn;
      xs.intvencl(xds);  // xds=interval enclosure of xs

      // part 4, computation of Delta
     // interval Ldelta[MAX_DIM+1];
      interval *Ldelta;
      Ldelta=new interval[inclode.dim+1];
      ok=inclode.incl_compute_Ldelta(h,w1_encl,xencl,yencl,Ldelta);

      if(ok){
        //  the rearrangement - the inclusion of LDelta
        xn.add(Ldelta);
        xn.intvencl(xdn);  // xdn=interval enclosure of xdn

        if(inclode.dissipative){
          interval *xstep;
          xstep=new interval[inclode.dim+1];
          inclode.incl_dissOneStep(xd,h,xencl,xstep);  // \pi_x \varphi(h,x,y) \subset xstep
             // but the bounds are obtained for dissipative coordinates from
             //  linear inequalites


          if(onesteppar.doC0intersection){
            ok=xn.intersect_interval(xstep);    // if xn.sintersect_rep() then it also
                      // does if  necessary  a rearrangment, which will guarante
                      // that <xn.xc> is   in the interval enclosure  of <xn>
            xn.intvencl(xdn);
          }else{
            if(xn.sintersect_rep()){
              ok=commonpart(xstep,xn.enclo_intv,xn.enclo_intv,inclode.dim);
              if(ok){
                copyvector(xn.enclo_intv,xdn,inclode.dim);
                if(!xn.test_xc_in_encl(false)){ // we do the intersection of "Lohner set"
                                           // because xc in not in <xn.enclo>
                  ok=xn.intersect_interval(xn.enclo_intv);
                  xn.intvencl(xdn);
                }
              }
            }
          }
          if(!ok){
            char comment[]="an empty intersection of  computed bound for x(h) and rough enclosure for x(h) ";
            cerr <<  "incl_OneStepLohner - serious error " << endl;
            cerr << comment << endl;
            info_err_commonpart(xdn,xstep,inclode.dim,comment);
            #if defined(WIN95) && defined(KRAK)
            waitBt();
            #endif
         //   exit(EXIT_FAILURE);
          }
          /*  // these are tests lines
          if(!xn.test_xc_in_encl(true)){
            cerr <<  "incl_OneStepLohner - warning: xc is not in enclosure " << endl;
            #if defined(WIN95) && defined(KRAK)
            waitBt();
            #endif
          }
          */

          delete [] xstep;

        }


      }
      delete [] Ldelta;
      Ldelta=NULL;


    }

  }

  inclode.set_yc(prev_yc);
  delete [] prev_yc;


  return ok;
}



void  incl_c0updates_in_sec_reg(InclOde &ode,interval h,Set &x,interval xd[],
    SetWithTail & y,SetWithTail & yn,interval w2_encl[],
    SetWithTail & yencl,  Set &xs, interval xds[],interval yc[],
    interval w1_encl[], int usediffcoord,LinCC *coordch,interval lmarrmax[],
    interval PoincVal[],SetWithTail & Py,interval CPoincVal[])
// The function updates PoincVal,Py and CPoincVal , which are enclosures of
//  the intersection of the current trajectory with current section
// it is called in the 'section region', (not present in parameters here)
// Input data:
//     ode - diff. eq. - defining the flow \varphi
//      h - time step
//      xd=x,  xdn=xn;   are such that \varphi(h,(x,y)) \subset xn \times yn
//    but since we don't really need xn, xdn - in this procedure they
//               are not in the parameters, but they still are listed
//               in this commentary,
//      and \varphi([0,h],(x,y)) \subset w2_encl \times yencl
//       \pi_x varphi([0,h],(x,y_c)) \subset w1_encl
//        \pi_x varphi(h,(x,y_c)) \subset xs \subset xds
//    if <usediffcoord> then we are interested in in the intersection expressed
//      in other coordinates <coordch>
//     PoincVal, Py, CPoincVal - current values of the intersection
//  Output data:
//     PoincVal, Py and CPoincVal - will contain all [0,h]-trajectories emanating
//           from x \times y, we estimate them more effectively (on x-variables)
//           than by taking just  xencl   \times yencl, in fact on y - we just take hull
//                   of yencl
//     lmarrmax - a control variable used to monitor the size of the section
//           error due to difference between actual trajctory and the segment
//           joining x and \varphi(h,x)
{

//  interval prev_yc[MAX_DIM+1];
  interval *prev_yc;
  prev_yc=new interval[ode.dim+1];
  ode.current_yc(prev_yc);

  ode.set_ycenter(yc);

//  interval lmarr[MAX_DIM+1];
//  interval xdnLM[MAX_DIM+1],xdLM[MAX_DIM+1];
  interval *lmarr,*xdnLM, *xdLM;
  lmarr=new interval[ode.dim+1];
  xdnLM=new interval[ode.dim+1];
  xdLM=new interval[ode.dim+1];

  ode.incl_fromsegdist(w1_encl,w2_encl,yencl,lmarr,h);
  for(int i=1;i<=ode.dim;i++)
    lmarr[i]=Ihull(lmarr[i],-lmarr[i]);

  for(int i=1;i<=ode.dim;i++){
    xdnLM[i]=xds[i]+lmarr[i];
    xdLM[i]=xd[i]+lmarr[i];  // here in fact I'm adding 'Delta' twice -
                             // at the begining and at the end of each step
  }

  if(!usediffcoord)
    MultiIhull(lmarrmax,lmarr,lmarrmax,ode.dim);

  MultiIhull(xdLM,PoincVal,PoincVal,ode.dim);
  MultiIhull(xdnLM,PoincVal,PoincVal,ode.dim);

  //  SwT_hull(yencl,Py,Py);
  SwT_hull(y,Py,Py);
  SwT_hull(yn,Py,Py); // hence we added tp Py the hull of y and yn
                      // this correct for the tail evolution
                      // and obviously for constant pertubatation

  if(usediffcoord){
   // we compute the Poincare map in other coordinates
    // interval CxdnLM[MAX_DIM+1],CxdLM[MAX_DIM+1];
    interval *CxdnLM,*CxdLM;
    CxdnLM=new interval[ode.dim+1];
    CxdLM=new interval[ode.dim+1];
    ode.incl_fromsegdist(w1_encl,w2_encl,yencl,lmarr,h,coordch->intv_m);
    for(int i=1;i<=ode.dim;i++)
      lmarr[i]=Ihull(lmarr[i],-lmarr[i]);

    xs.intvencl(CxdnLM,usediffcoord,coordch);
    x.intvencl(CxdLM,usediffcoord,coordch);
    for(int i=1;i<=ode.dim;i++){
      CxdnLM[i]=CxdnLM[i]+lmarr[i];
      CxdLM[i]=CxdLM[i]+lmarr[i];
    }

    MultiIhull(CxdLM,CPoincVal,CPoincVal,ode.dim);
    MultiIhull(CxdnLM,CPoincVal,CPoincVal,ode.dim);
    MultiIhull(lmarrmax,lmarr,lmarrmax,ode.dim);

    delete [] CxdLM;
    CxdLM=NULL;
    delete [] CxdnLM;
    CxdnLM=NULL;
  }

  ode.set_yc(prev_yc);

  delete [] xdLM;
  delete [] xdnLM;
  delete [] lmarr;
  delete [] prev_yc;

}






int incl_in_sec_region(int secdir,interval secnew,int *startPoinc,InclOde &ode,
    interval h,Set &x,interval xd[],SetWithTail &y,Set & /*xn*/,
    interval /*xdn */[],SetWithTail & yn, interval xencl[],SetWithTail & yencl,
    Set &xs, interval xds[],interval yc[],interval w1_encl[],
    int usediffcoord,LinCC *coordch, interval lmarrmax[],
    interval PoincVal[], SetWithTail &Py,interval CPoincVal[])
//  the function handles the section region for Incl_Lo_Poincare
//  It is called close to the section (in 'Section region')
//  returns 1 - if we passed the section (i.e. we got on the other
//               'desirable' side of the section), in this case we don't
//             update PoincVal, Py and CPoincVal
//  otherwise returns 0
////////////
//  secdir - UP or DOWN
//  secnew = section(new set)
//  *startPoinc - tells if we started to save the image of the Poincare map
//    in <PoincVal> and <CPoincVal>
//  h - time step
//  xencl \times yencl =enclosure for \varphi([0,h],(x,y))
//  xd - interval enclosure of x
//  xdn - interval enclosure of xn
//  \varphi(h,(x,y)) \subset xn \times yn
//  ON OUTPUT:
//      we update (when 0 is returned) *startPoinc,lmarrmax, PoincVal,
//           Py, CPoincVal
{

  int poinc_finished=0;  // this is the returned variable
  int oldstartPoinc=*startPoinc;


  if (!(*startPoinc)){ // we did'nt started yet to build the Poincare image
    if ((secdir == DOWN)? (secnew.left_b() <= 0.0): (secnew.right_b() >= 0.0)){
       // here  we start to build Poincare image
       //  we have to initalize PoincVal,Px,CBr,CDx,Clmarr
      *startPoinc=1;
    }
  } // endif startPoinc

  if ((secdir == DOWN)? (secnew < 0): (secnew > 0)){
    poinc_finished=1;
  }


  // Making various updates
  if(*startPoinc){
    if(!oldstartPoinc){  // we have to initialize PoincVal, Py and CPoincVal
      copyvector(xd,PoincVal,ode.dim);
      Py=y;
      if(usediffcoord)
        x.intvencl(CPoincVal,usediffcoord,coordch);
    }
    if(!poinc_finished)
      incl_c0updates_in_sec_reg(ode,h,x,xd,y,yn,xencl,yencl,xs,xds,
            yc,w1_encl,usediffcoord,coordch,lmarrmax,PoincVal,Py,CPoincVal);
  }

  return poinc_finished;
}



int incl_get_past_section(const C1Func &section,int secdir,InclOde &ode,
     InclOneStepParamType &onesteppar,interval h,Set &x,interval xd[],SetWithTail &y,
     Set &xn,interval xdn[], SetWithTail & yn,interval xencl[], SetWithTail &yencl,
     Set &xs, interval xds[], interval yc[], interval w1_encl[],interval rem[],
     interval *h_used)
// We assume that on input:
//   x \subset xd=enclosure(x)
//   \varphi(h,(x,y)) \subset (xn \times yn) \subset xdn=enclosure(xn) \times yn
//        is on the other side of the section  <section>
//    \pi_x \varphi(h,(x,y_c)) \subset xs \subset xds for fixed y=y_c
//    \pi_x \varphi([0,h],(x,y_c)) \subset w1_encl for fixed y=y_c
//
//  On output:
//     if 1 - is returned, then
//    \varphi(h_used,(x,y)) \subset xn \times yn \subset (xdn=enclosure(xn)) \times yn
//    \varphi([0,h_used],(x,y)) \subset xencl \times yencl
//    \pi_x \varphi(h_used,(x,y_c)) \subset xs \subset xds for fixed y=y_c
//    \pi_x \varphi([0,h_used],(x,y_c)) \subset w1_encl for fixed y=y_c
//    h_used - was the time step found to get close to section
//            but but on the other side
//   if 0 - is returned, then xn,xs, are undefined
/////////////////
///  Idea:
//     We can estimate the crossing time (i.e. how long does it takes for
//        x \times y to flow past the section) by estimating
//        ds/dt=d section / dt over <x+xn> \subset xd + xdn
//     Assuming that the size of <x> or <xn> does not change much
//     during the passing of <x> through the section, we can estimate
//     the time needed to flow past section but not further than
//        diam(sec(x))*c; where c - is some
//                     small number c=0.01 or c=0.1
{
  interval sec_on_xn=section.eval(xn);
  if(sec_on_xn.contains(0)){
    cerr << "incl_get_past_section - bad xn - intersects section" << endl;
     #if defined(WIN95) && defined(KRAK)
      cerr << "press any key " << endl;
      waitBt();
    #endif
    exit(EXIT_FAILURE);
  }



  double c=0.01;
  int ok=1;



  interval xxn[MAX_DIM+1];
  MultiIhull(xd,xdn,xxn,ode.dim);  //  x+xn \subset xxn

  interval valvfield[MAX_DIM+1];
  ode.incl_field(xxn,yencl,valvfield);

  interval grad_of_sec_func[MAX_DIM+1];
  section.gradient(xxn,grad_of_sec_func);

  interval dsec_dt=scprod(grad_of_sec_func,valvfield,ode.dim); // dsec_dt = dsec/dt
  if(dsec_dt.contains(0)){
    cerr << "incl_get_past_section - Bad direction of vect. f. in section region" << endl;
     #if defined(WIN95) && defined(KRAK)
      cerr << "press any key " << endl;
      waitBt();
    #endif
    exit(1);
  }
  dsec_dt=abs(dsec_dt);
  double ds_dt=mid(dsec_dt).left_b();
  interval sec_on_x=section.eval(x);

  double diamSx=diameter(sec_on_x).right_b();

  double dist_x;  // this tells how far is  the 'rear' end of <x> before the section
  if(secdir==DOWN) dist_x=sec_on_x.right_b();
  else dist_x=-sec_on_x.left_b();

  double ha=(dist_x + c*diamSx)/ds_dt; // assuming constant speed, this is
        // an estimate how long it takes to fly through section

  double minh=0;
  double maxh=h.left_b(); // the interval (minh,maxh) - gives an reasonable
               // interval for test ; minh - is bad, maxh - is ok. but too big
  ha=newtimestep(ha,minh,maxh,0.1);

  int done;

  if((secdir==UP) && (sec_on_xn >0))
    done=(sec_on_xn.left_b() < c*diamSx);
  if((secdir==DOWN) && (sec_on_xn <0))
    done=(sec_on_xn.right_b() > -c*diamSx);


  Set cxn=xn;
  interval cxdn[MAX_DIM+1];
  interval cxencl[MAX_DIM+1];
  interval crem[MAX_DIM+1],  cw1_encl[MAX_DIM+1];
  SetWithTail cyn(ode.dim), cyencl(ode.dim);
  Set cxs=xs;
  interval cxds[MAX_DIM+1], cyc[MAX_DIM+1];
  //  the variables 'c' + name - are working variables for a corresponding
  //     <name>-variable


  int improvement=0;

  interval h_impr;
  Set xn_impr=xn;
  interval xdn_impr[MAX_DIM+1];
  interval xencl_impr[MAX_DIM+1];
  interval rem_impr[MAX_DIM+1];
  Set xs_impr=xs;
  interval xds_impr[MAX_DIM+1];
  SetWithTail  yencl_impr(yencl);
  SetWithTail  yn_impr(yn);
  interval w1_encl_impr[MAX_DIM+1];
  interval yc_impr[MAX_DIM+1];



  int itno=0;
  int maxitnum=10;
  while(!done && ok  && (itno < maxitnum)){
    itno++;
    ok=incl_OneStepLohner(ode,onesteppar,ha,x,xd,y,cxn,cxdn,cyn,cxencl,
           cyencl,cxs,cxds,cyc,cw1_encl,crem);
    if(ok){
      int saveimpr=0;
      interval sec_on_xn=section.eval(cxn);
      if((secdir==UP) && (sec_on_xn >0)){
        done=(sec_on_xn.left_b() < c*diamSx);
        saveimpr=1;
      }
      if((secdir==DOWN) && (sec_on_xn <0)){
        saveimpr=1;
        done=(sec_on_xn.right_b() > -c*diamSx);
      }

      // saving improved values
      if(saveimpr){
        improvement=1;
        h_impr=ha;
        xn_impr=cxn;
        copyvector(cxdn,xdn_impr,ode.dim);
        copyvector(cxencl,xencl_impr,ode.dim);
        copyvector(crem,rem_impr,ode.dim);
        yn_impr=cyn;
        yencl_impr=cyencl;
        xs_impr=cxs;
        copyvector(cxds,xds_impr,ode.dim);
        copyvector(cyc,yc_impr,ode.dim);
        copyvector(cw1_encl,w1_encl_impr,ode.dim);
      } // endif saveimpr
      if(!done  && (itno < maxitnum)){
        double dha;
        if(((secdir==UP) && (sec_on_xn > c*diamSx))
          || ((secdir==DOWN)  && sec_on_xn < -c*diamSx)  ){ // I overshot
          double toofar;
          if(secdir==UP) toofar=sec_on_xn.left_b() - c*diamSx;
          if(secdir==DOWN) toofar=fabs(sec_on_xn.right_b()) - c*diamSx;
          dha=-toofar/ds_dt;

          if(maxh > ha) maxh=ha; // we know that with maxh - we are
                  // on the correct side but we overshot

        } else{ // we need a bigger time step
          double adddist;
          if(secdir==UP) adddist=c*diamSx - sec_on_xn.left_b() ;
          if(secdir==DOWN) adddist=sec_on_xn.right_b() + c*diamSx;
          dha=adddist/ds_dt;


          if(ha > minh) minh=ha;

        }
        ha+=dha;
        ha=newtimestep(ha,minh,maxh,0.1);
      } // endif !done

    } // endif ok
  } // end while (!done ...

  if(improvement){
    ok=1;
    *h_used=h_impr;
    xn=xn_impr;
    copyvector(xdn_impr,xdn,ode.dim);
    copyvector(xencl_impr,xencl,ode.dim);
    copyvector(rem_impr,rem,ode.dim);
    yn=yn_impr;
    yencl=yencl_impr;
    xs=xs_impr;
    copyvector(xds_impr,xds,ode.dim);
    copyvector(yc_impr,yc,ode.dim);
    copyvector(w1_encl_impr,w1_encl,ode.dim);
  } else{
    if(done){   // this is possible only if our initial data were satisfactory
      ok=1;
      *h_used=h;
    } else{    // we must accept the xn - even if we think it is not optimal
      ok=1;
      *h_used=h;
    }
  }
  if(!done && ok){
    char comment[]="get_past_section - inclusion - problems";
    near_section_problems_report(comment,ode.dim,maxitnum,xd,xdn,sec_on_x,
        section.eval(xn),c,dsec_dt,ds_dt,h.right_b(),minh,maxh,ha);
  }

  return(ok);
}


int incl_get_close_to_section(const C1Func &section,int secdir,InclOde &ode,
     InclOneStepParamType &onesteppar,interval h,Set &x,interval xd[], SetWithTail &y,
     Set &xn,interval xdn[], SetWithTail &yn, interval xencl[],
     SetWithTail &yencl,Set &xs,interval xds[],interval yc[],
     interval w1_encl[],interval rem[],interval *h_used,interval *new_h)
// We assume that on input:
//   x \subset xd - xd is an enclosure for x
//   \varphi(h,x) \subset xn \subset xdn=enclosure(xn)  intersects the
//        the section <section>
//    \pi_x \varphi(h,(x,y_c)) \subset xs \subset xds for fixed y=y_c
//    \pi_x \varphi([0,h],(x,y_c)) \subset w1_encl for fixed y=y_c
//  On output:
//    x,xd - are unchanged
//    h_used - is the time step found to get close to section
//            but to be on the same side as <x>
//   if 1 is returned, then:
//    *h_used - is computed so that we are still before the section
//    and if *h_used==0, then xn,xencl,xs,xds- are not changed
//    if *h_used > 0, then xn,xecnl,xds,.. - is changed so that
//    \varphi(h_used,x) \subset xn \subset xdn=enlosure(xn)
//    \varphi([0,h_used],x) \subset xencl
//    \pi_x \varphi(h_used,(x,y_c)) \subset xs \subset xds for fixed y=y_c
//    \pi_x \varphi([0,h_used],(x,y_c)) \subset w1_encl for fixed y=y_c
//    new_h  - a time step suggested for the computation of crossing
// if 0 is returned - then xn,xdn,xs,.. - are undefined, (may be changed)
/////////////////
///  Idea:
//     new_h - shall be related to the section crossing time of the set
//                    <x>
//     We can estimate this crossing time by estimating
//        ds/dt=d section / dt over <x \cup xn> \subset xd \cup xdn
//     Assuming that the size of <x> or <xn> does not change much
//     during the passing of <x> through the section, we can estimate
//       the passing time by  t_passsing= diam(sec(x))/|ds/dt|
//     new_h - will be defined then in terms of t_passing;
//          new_h=t_passing/(k > 2)
//     We compute heuristically <h_used> so that
//       |sec(\varphi(h_used,x))|_{min} > diam(sec(x))*c; where c - is some
//                     small number c=0.01 or c=0.1
//     Due to these choices we can expect the computed passing time to be
//      only slightly bigger than heuristic estimates
{
  interval sec_on_x=section.eval(x);
  if(sec_on_x.contains(0)){
    cerr << "incl_get_close_to_section - bad initial condition - already intersects section" << endl;
     #if defined(WIN95) && defined(KRAK)
      cerr << "press any key " << endl;
      waitBt();
    #endif
    exit(EXIT_FAILURE);
  }


  double c=0.01;
  int ok=1;

  interval xxn[MAX_DIM+1];
  MultiIhull(xd,xdn,xxn,ode.dim);  //  x \cup xn \subset xxn

  interval valvfield[MAX_DIM+1];
  ode.incl_field(xxn,yencl,valvfield);

  interval grad_of_sec_func[MAX_DIM+1];
  section.gradient(xxn,grad_of_sec_func);

  interval dsec_dt=scprod(grad_of_sec_func,valvfield,ode.dim); // dsec_dt = dsec/dt
  if(dsec_dt.contains(0)){
    cerr << "incl_get_close_to_section - Bad direction of vect. f. in section region" << endl;
     #if defined(WIN95) && defined(KRAK)
      cerr << "press any key " << endl;
      waitBt();
    #endif
    exit(1);
  }
  dsec_dt=abs(dsec_dt);
  double ds_dt=mid(dsec_dt).left_b();


  double diamSx=diameter(sec_on_x).right_b();
  double t_passing=diamSx/ds_dt;  // this is an estimate for the crossing time

  double sec_min_on_x=abs(sec_on_x).left_b(); // this gives me a minimal value of the
         // section function on x
  //  to estimate the time to section we use  sec_min_on_x=ds/dt * h
  //  but to leave some margin for wrapping (from interval comp. ) we want
  //        sec_min_x  - c*diamSx = ds/dt *h
  double ha=(sec_min_on_x - c*diamSx)/ds_dt;


  double minh=0;
  double maxh=h.left_b(); //  minh - should be ok, but too far before the section
               // maxh - allways to big, we hit the section
  ha=newtimestep(ha,minh,maxh,0.1);
  int done=0;

  if(secdir==UP)
    done=sec_on_x.contains(-c*diamSx);
  if(secdir==DOWN)
    done=sec_on_x.contains(c*diamSx);


  Set cxn=xn;  // to initialize dimension
  interval cxdn[MAX_DIM+1];
  interval cxencl[MAX_DIM+1];
  interval crem[MAX_DIM+1],  cw1_encl[MAX_DIM+1];
  SetWithTail cyn(ode.dim), cyencl(ode.dim);
  Set cxs=xs; // to iniatialize dim
  interval cxds[MAX_DIM+1], cyc[MAX_DIM+1];
  //  the variables 'c' + name - are working variables for a corresponding
  //     <name>-variable



  int improvement=0;
  interval h_impr;
  Set xn_impr=xn;
  interval xdn_impr[MAX_DIM+1];
  interval xencl_impr[MAX_DIM+1];
  interval rem_impr[MAX_DIM+1];
  Set xs_impr=xs;
  interval xds_impr[MAX_DIM+1];
  SetWithTail  yencl_impr(yencl);
  SetWithTail  yn_impr(yn);
  interval w1_encl_impr[MAX_DIM+1];
  interval yc_impr[MAX_DIM+1];


  int itno=0;
  int maxitnum=10;
  while(!done && ok  && (itno < maxitnum)){
    itno++;
    ok=incl_OneStepLohner(ode,onesteppar,ha,x,xd,y,cxn,cxdn,cyn,cxencl,cyencl,
           cxs,cxds,cyc,cw1_encl,crem);
    if(ok){
      int saveimpr=0;
      interval sec_on_xn=section.eval(cxn);
      if(secdir==UP){
        if(sec_on_xn<0){
          saveimpr=1;
          if(sec_on_xn.right_b() > -c*diamSx) done=1;
        }
      } else{
        if(sec_on_xn>0){
          saveimpr=1;
          if(sec_on_xn.left_b() < c*diamSx)  done=1;
        }
      }
            // saving improved values
      if(saveimpr){
        improvement=1;
        h_impr=ha;
        xn_impr=cxn;
        copyvector(cxdn,xdn_impr,ode.dim);
        copyvector(cxencl,xencl_impr,ode.dim);
        copyvector(crem,rem_impr,ode.dim);
        yn_impr=cyn;
        yencl_impr=cyencl;
        xs_impr=cxs;
        copyvector(cxds,xds_impr,ode.dim);
        copyvector(cyc,yc_impr,ode.dim);
        copyvector(cw1_encl,w1_encl_impr,ode.dim);
      } // endif saveimpr

      if(!done && (itno < maxitnum)){
        double dha;
        int overshot;
        if(secdir==UP) overshot=(sec_on_xn.right_b() > 0);
        else overshot=(sec_on_xn.left_b() <0);
        if(overshot){  // I overshot, I need a better guess
          maxh=ha;
          double sec_min_xn=(secdir==UP)?sec_on_xn.right_b():fabs(sec_on_xn.left_b());
          dha=-(sec_min_xn+c*diamSx)/ds_dt;

        }else{ //  I'm  before section
          minh=ha;
          double sec_min_xn=abs(sec_on_xn).left_b();
          dha=(sec_min_xn-c*diamSx)/ds_dt;

        } // endif overshot
        ha+=dha;
        ha=newtimestep(ha,minh,maxh,0.1);
      } // endif !done
    } //endif ok
  } // end while

  double divfact=4;
  if(ode.divsecreg > divfact) divfact=ode.divsecreg;
  *new_h=Left(t_passing/divfact);
  if(*new_h > h) *new_h=h;

  if(improvement){
    ok=1;
    *h_used=h_impr;
    xn=xn_impr;
    copyvector(xdn_impr,xdn,ode.dim);
    copyvector(xencl_impr,xencl,ode.dim);
    copyvector(rem_impr,rem,ode.dim);
    yn=yn_impr;
    yencl=yencl_impr;
    xs=xs_impr;
    copyvector(xds_impr,xds,ode.dim);
    copyvector(yc_impr,yc,ode.dim);
    copyvector(w1_encl_impr,w1_encl,ode.dim);
  }else{
    if(done){ // this happens only if our initial data was satisfatory
      ok=1;
      *h_used=0;
    }else{    // this means that the function above failed to produce
      // something closer to section, but we can still continue computations
      *h_used=0;
    }
  }

  if(!done && ok){
    char comment[]="Problem with the inclusion get_close_to_section";
    near_section_problems_report(comment,ode.dim,maxitnum,xd,xdn,
       sec_on_x,section.eval(xn),c,dsec_dt,ds_dt,
                                   h.right_b(),minh,maxh,ha);
  }

  return(ok);
}


void infoInclOneStepFailed(Set &x,interval xd[],SetWithTail &y,
        int dim,interval t,interval h,const char *funcname)
{
  ofstream errorlog(zglerrorlogname,ios::app);
  entriesinerrorlog=1;
  errorlog <<"----------------------------------" << endl;
  writeCompInfo(errorlog);
  writedatetime(errorlog);

  cerr << funcname << " - OneStep function failed " << endl;
  cerr << "t=" << t << "   h=" << h <<  endl;

  errorlog << funcname << " - OneStep function failed " << endl;
  errorlog << "t=" << t << "   h=" << h <<  endl;
  errorlog << "the set causing problem: " << endl;
//   saveAllCoord(errorlog,xd,dim,3,(dim > 4),4);
  x.savetxt(errorlog);
  errorlog << "y-part - perturbation : " << endl;
  saveAllCoord(errorlog,y.y,y.dim,3,(y.dim>4),4);
  if(y.tailexists)
    y.ptail->save(errorlog,3,"Tail of y:");
}


void OkGen_InclPoincare(int *ok,Set &x,interval xd[],SetWithTail &y,interval xdn[],
        int dim,interval t,interval h,const char *funcname)
// this function is a  test for errors and sends error messages
// corresponding to function <funcname>
 // we test whether the image stabilized (which we also
// consider an error - because we will be not able to compute Poincare map)
{

   if(!(*ok)){
     infoInclOneStepFailed(x,xd,y,dim,t,h,funcname);
   }
   if(*ok){
     *ok=!MultiIn(xd,xdn,0,dim);
     if(!(*ok)){
       #if defined(WIN95) && defined(KRAK)
       if(dim >3)   rootFrame.Clear();
       #endif

       ofstream errorlog(zglerrorlogname,ios::app);
       entriesinerrorlog=1;
       errorlog <<"----------------------------------" << endl;
       writeCompInfo(errorlog);
       writedatetime(errorlog);


       cerr << funcname << "- image stabilized " << endl;
       cerr << "t=" << (t+h) << endl;
       cerr << "xd[i]    xdn[i] " << endl;

       errorlog << funcname << "- image stabilized " << endl;
       errorlog << "t=" << (t+h) << endl;
       errorlog << "xd[i]    xdn[i] " << endl;

       for(int i=1;i<=dim;i++){
         cerr << xd[i] << "   " << xdn[i] << endl;
         errorlog << xd[i] << "   " << xdn[i] << endl;
       }
     }
   }
   if(!(*ok)){
      #if defined(WIN95) && defined(KRAK)
//       cerr << "press any key " << endl;
//       waitBt();
     #endif
   }
}





int incl_Lo_Poincare(InclOde& ode, InclPoincareParType &param, Set &x,
           SetWithTail & y0,const C1Func &section, int secdir,estimates  *estm,
           interval Px[],SetWithTail & Py,interval dp[][MAX_DIM+1],
           int usediffcoord,LinCC *coordch)
//  computes rigorously the Poincare map of x using the Lohner
//     algorithm   for a differential inclusion <ode> x' \in ode.f(x) + y
//   where y0 is an initial value for y - for true differential inclusions
//    this parameter can be ignored, but it is important for PDEs with
//     the  evolving tail

//  1 if returned, when the Poincare map can be computed and the return time
//          is <= tmax,
//     then P(set(numLohner,x0,C0,r0)) \subset Px  (see below for the description
//     of the representation)
//    Py - on output is the value for the y-Poincare map, usually can be ignored
//              , but it is important for PDEs with th tail  evolving
//  equation for section is given by  section.eval(x)=0
//  dp - the variational matrix just after passing the section - Lohner matrix
//         C - for controlling a Lipschitz part of error, used elsewhere
//             for diagnosing properties, sooner or latter this type of information
//         shall become part of the estm - class, which will collect what is
//         needed for a particular computations
//

//  usediffcoord - tells if you different coordinates in the section region
//  coordch  - coordinates in which we express the result <Px>
//     (these are new coordinates), dp - is always expressed in OLD coordinates

// on OUTPUT :   on output Px  \times Py - is an enclosure of the image
//                usually Py - is  just perturbation of an ODE and is not interesting for us
//                 but for PDE with an evolving tail - here is the info about tail


{


// int method=param.method;
// int order=param.order;
 interval h=param.h;  // this is a value of time step used from the begining,
      // but close to the section the time-step is chosen to
      // assure an reasonable section error
 double tmax=param.tmax;

 interval rem[MAX_DIM+1];
 interval h0=h;
 double remMax=0;

 interval xencl[MAX_DIM+1];  // used for enclosures of \varphi([0,h],x)

 interval lmarrmax[MAX_DIM+1];    // to deal with the section
                                //errors  componentwise
 zero(lmarrmax,ode.dim);
 interval CPoincVal[MAX_DIM+1]; // used if <usediffcoord> - to compute
                // the Poincare map in other coordinates


 Set xn;
 interval xd[MAX_DIM+1],xdn[MAX_DIM+1]; // x \subset xd
                                      // xn \subset xdn
 x.intvencl(xd);
 SetWithTail  y(y0);
 SetWithTail yn(ode.dim),yencl(ode.dim);



 if(param.onesteppar.method==RUNGEKUTTA)  param.onesteppar.order=4;

 initestm(estm,ode.dim);
 interval secnew=section.eval(x);
 interval secold=secnew;

 int insecregion=0;
 int startPoinc=0;
 interval PoincVal[MAX_DIM+1];

 Set xs;
 interval xds[MAX_DIM+1], yc[MAX_DIM+1], w1_encl[MAX_DIM+1];

 interval tPoinc;

 int num_time_steps_in_sec_reg=0;
 int ok=1;
 int firststep=1;
 interval t=0;
 interval new_h=h;
 int poinc_finished=0;
 while((t<=tmax) && (!poinc_finished) && ok){
   //   cout << t << endl;
   int dosubst=1;
   h=new_h;

   // here is the place to try several different values for  <h> if there are
   //                                 some problems

   ok=incl_OneStepLohner(ode,param.onesteppar,h,x,xd,y,xn,xdn,yn,xencl,yencl,
           xs,xds,yc,w1_encl,rem);
     // xencl \times yencl =enclosure for \varphi([0,h],x,y)
     //   xn \times yn = enclosure for \varphi(h,x,y)
     // xd=enclosure for x = eval(x)
     // xdn=eval(xn) enclosure for  xn
     //   xds=enclosure(xs),  xs = \varphi(h,x,y==y_c) - the deterministic part

   OkGen_InclPoincare(&ok,x,xd,y,xdn,ode.dim,t,h,"Incl_Lo_Poincare");
      // a test for errors and eventually whether the image stabilized (which we also
        // consider an error - because we will be not able to compute Poincare map)


   //
   if(!ok) continue;


   secnew=section.eval(xn);
   if (!firststep){
     if (insecregion){
       num_time_steps_in_sec_reg++;
       int oldstartPoinc=startPoinc;
       poinc_finished=incl_in_sec_region(secdir,secnew,&startPoinc,ode,h,
          x,xd,y,xn,xdn,yn,xencl,yencl,xs,xds,yc,w1_encl,usediffcoord,
              coordch,lmarrmax,PoincVal,Py,CPoincVal);
       if(poinc_finished){
         // we try to make the time-step as small as possible and
         //  still pass through section:
         interval h_used;
         if(incl_get_past_section(section,secdir,ode,param.onesteppar,h,x,xd,y,
                  xn,xdn,yn,xencl,yencl,xs,xds,yc,w1_encl,rem,&h_used)){
           h=h_used;
           incl_c0updates_in_sec_reg(ode,h,x,xd,y,yn,xencl,yencl,
                          xs,xds,yc,w1_encl,usediffcoord,coordch,
                              lmarrmax,PoincVal,Py,CPoincVal);

         } else{
           cerr << "Incl_Lo_Poincare - get_past... returned error" << endl;
            #if defined(WIN95) && defined(KRAK)
             cerr << "press any key" << endl;
             waitBt();
           #endif
           ok=0;
           poinc_finished=0;
           continue;
         }
       }


       // keeping track of the return time
       if((!oldstartPoinc) && startPoinc)
         tPoinc=t;
       if(startPoinc) tPoinc=Ihull(tPoinc,t+h);

     } // endif insecregion
     if (!insecregion){
       if ((secdir == DOWN ) ? ((secnew.left_b() <= 0.0) && (secold.left_b() > 0.0) ):
          ((secnew.right_b() >= 0.0) && (secold.right_b() < 0.0) )){
              // we have reached the section
         insecregion=1;
         dosubst=0;   // if !dosubst and the end oth loop, then
                      // we will repeat the calculations for this time 't'
                       // with smaller time step h

         interval h_used;
         if(incl_get_close_to_section(section,secdir,ode,param.onesteppar,h,x,xd,y,
             xn,xdn,yn,xencl,yencl,xs,xds,yc,w1_encl,rem,&h_used,&new_h)){
           if(new_h < 0.01*h0)
             new_h=0.01*h0; // there is no point is using too small time step
              // because  it happens (in case of small sets) the
                   // proposed time step is many orders of magnitude smaller
                   // than h0
           if(h_used >0){  // because it may happen that <x> was already
                           // close enough, we have then h_used=0;
                           // and we are interested only in new_h
                           // and don't do any substitution x=xn for the next
                           // time step
             h=h_used;
             dosubst=1;
             secnew=section.eval(xn);
           } // endif (h_used >0)
         } // get_close_...
         else{
           cerr << "Incl_Lo_Poincare - get_close... returned error" << endl;;
            #if defined(WIN95) && defined(KRAK)
             cerr << "press any key" << endl;
             waitBt();
           #endif
           ok=0;
           continue;
         }
       }
     }
   }  // endif !firststep

   if (dosubst){

     //  These things are less important - keeping track of various estimates
     // we calculate various quantities along trajectory
     UpdateBounds(xencl,estm,firststep);
     if(arrfabs(rem,ode.dim) > remMax) remMax=arrfabs(rem,ode.dim);
     //  END OF - These things are less important - keeping track of various
     //                          estimates

     t+=h;
     secold=secnew;
     x=xn;
     y=yn;
     copyvector(xdn,xd,ode.dim);
     if(x.rearrangement(xd)) estm->num_of_rep_switches= estm->num_of_rep_switches +interval(1.0);

     firststep=0;
   } // endif dosubst


 } // end while



 if(poinc_finished){
   interval *valvfield;
   valvfield=new interval[ode.dim+1];
   ode.incl_field(PoincVal,Py,valvfield);

   interval *grad_of_sec_func;
   grad_of_sec_func=new interval[ode.dim+1];
   section.gradient(PoincVal,grad_of_sec_func);

   interval scprod_f_grad_s=scprod(grad_of_sec_func,valvfield,ode.dim);
   delete [] grad_of_sec_func;
   grad_of_sec_func=NULL;
   delete [] valvfield;
   valvfield=NULL;

   poinc_finished=((secdir == DOWN)?scprod_f_grad_s<0:scprod_f_grad_s>0 );
   if (!poinc_finished){
     cerr << "Incl_Lo_Poincare - Bad direction of vect. f. in section region" << endl;
      #if defined(WIN95) && defined(KRAK)
       cerr << "press any key " << endl;
       waitBt();
     #endif
   }

   if(poinc_finished){
     if(!usediffcoord)  copyvector(PoincVal,Px,ode.dim);
     else   copyvector(CPoincVal,Px,ode.dim);

     xn.get_Cmatrix(dp);

     poincUpdateEstm(estm,ode,section,param.onesteppar.order,h0,x,PoincVal,tPoinc,
                 remMax,lmarrmax,usediffcoord,coordch,CPoincVal);
     estm->sec_timestepnumber=num_time_steps_in_sec_reg;

   }
 } // endif poinc_finished


 return(poinc_finished);
}




void show_info_incl_orbit(double h,double tmax, double firstbias,
      double biasstep,double trwrbstep,int savetraj, int saveestm,
       const char report[],const char repestm[])
// info for the screen for <incl_Lo_orbit>
{
   #if defined(WIN95) && defined(KRAK)
     rootFrame.Clear();
  #endif
  cout << "Function computes the trajectory the differential inclusion/ PDE"
        << endl
     << " for t in (0," << tmax <<") . Time step h=" << h << endl ;

  if(savetraj){
    cout  << "The computed trajectory is stored in the file " << endl;
    cout << " [" << report <<
     "] with the time step=" << trwrbstep  << endl;
  }
  else cout << "the trajectory is not written to file " << endl;



  cout << " Output on the screen is produced for the first time for t="
          << firstbias << endl
       << "and then after  every " << biasstep << "  time units" << endl;
  cout << "<norm> is the usual euclidean norm "  << endl;


  if(saveestm){
     cout << "Any time when we have output on screen - the current values "
     << endl;
     cout << "of estimates are stored in the file ["<< repestm <<  "] " <<
       endl  ;
  }

   #if defined(WIN95) && defined(KRAK)
    rootFrame << "Press any key \n" ;
    waitBt();
  #endif
}



void incl_Lo_orbit(InclOde& ode, InclPoincareParType &param,long num_time_steps, Set &x,
           SetWithTail & y0,const char comment[], double firstbias,
           bool savetraj, const char *report,bool saveestm,const char *repestm,bool savediam,
            ostream &outdiam)
//  computes rigorously the trajectory of x \oplus y0 using the Lohner
//       algorithm   for a differential inclusion <ode>
//       where y0 is an initial value for y - for true differential inclusions
//    this parameter can be ignored, but it is important for PDEs with
//       the evolving tail
//       num_time_steps - the number of time steps ; ignored if 0,
//       report - the name of file where the trajectory is saved
//       repestm - the name of file where the estimates along the trajectory
//                      are saved
//       outdiam - the file where we write the diameters
{

 bool showinfo=true;

// int method=param.method;
// int order=param.order;
 interval h=param.h;  // this is a value of time step used in computations
 double tmax=param.tmax;
 if(num_time_steps <= 0)
    num_time_steps=1+ long(tmax/param.h);;

 interval rem[MAX_DIM+1];
 double remMax=0;

 interval xencl[MAX_DIM+1];  // used for enclosures of \varphi([0,h],x)


 Set xn;
 interval xd[MAX_DIM+1],xdn[MAX_DIM+1]; // x \subset xd
                                      // xn \subset xdn
 x.intvencl(xd);
 SetWithTail  y(y0);
 SetWithTail yn(ode.dim),yencl(ode.dim);

 Set xs;  // this is for deterministic equation x'=f(x) + y_c
 interval xds[MAX_DIM+1], yc[MAX_DIM+1], w1_encl[MAX_DIM+1];




 estimates estm;
 initestm(&estm,ode.dim);

 double nextbias, biasstep;
 double trwrbstep,trwnextb;

 // biases: used for output
 nextbias=firstbias;
 biasstep=0.2;

 // biases: used for the trajectory writting to a file
 trwrbstep=0.01;
 trwnextb=firstbias;



 int RInitEstm=1;

 ofstream rhandle;
 if(savetraj){
   rhandle.open(report,ios::out);
   rhandle << ode.dim << " - the dimension "  << comment << endl;
 }

 if(showinfo)
   show_info_incl_orbit(h.right_b(),tmax,firstbias, biasstep,trwrbstep,
   savetraj,saveestm,report,repestm);


 int ok=1;
 interval t=0;

 long step_num=0;

 while((step_num < num_time_steps ) && ok){
   step_num++;
   t+=h;
   //   cout << t << endl;


   ok=incl_OneStepLohner(ode,param.onesteppar,h,x,xd,y,xn,xdn,yn,xencl,yencl,
           xs,xds,yc,w1_encl,rem);
     // xencl \times yencl =enclosure for \varphi([0,h],x,y)
     //   xn \times yn = enclosure for \varphi(h,x,y)
     // xd=enclosure for x = eval(x)
     // xdn=eval(xn) enclosure for  xn


   OkGen_InclPoincare(&ok,x,xd,y,xdn,ode.dim,t,h,"Incl_Lo_orbit");
                                          // a test for errors
   if(!ok) continue;

   if (t>= firstbias){
     UpdateBounds(xencl,&estm,RInitEstm);
     RInitEstm=0;
   }

   x=xn;
   y=yn;
   copyvector(xdn,xd,ode.dim);
   if(x.rearrangement(xd)){
          estm.num_of_rep_switches= estm.num_of_rep_switches +interval(1.0);
     if(savediam){
       outdiam << "t=" << t.right_b() << "  time step no=" << step_num << "   rearrangement done " << endl;
     }
   }


   if( t.right_b()>=trwnextb){
     if(savetraj) {
       rhandle  << t.right_b() << "   ";
       saveAllCoord(rhandle,xd,ode.dim);
     }
     trwnextb+=trwrbstep;
   }

   if (t.right_b() >= nextbias){
      #if defined(WIN95) && defined(KRAK)
       rootFrame.Clear();
     #endif
     cout << " t="<< t.right_b() <<"  t_max=" << tmax  <<
        "  norm=" << norm(xd,ode.dim,NORM_EUCL).right_b() << "  diam=" <<
       MultiIDiam(xd,ode.dim,NORM_MAX) << endl;

     if(saveestm) writeestm(repestm,&estm);

     if(savediam){
       outdiam << " t="<< t.right_b() << "  time step no=" << step_num  << "  norm=" << norm(xd,ode.dim,NORM_EUCL).right_b() << "  diam=" <<
          MultiIDiam(xd,ode.dim,NORM_MAX) << endl;
       writediam(outdiam,xd,ode.dim);
     }

     nextbias+=biasstep;
   }

 } // end while

 if(savediam){
   outdiam << " t="<< t.right_b() << "  time step no=" << step_num <<  "  norm=" << norm(xd,ode.dim,NORM_EUCL).right_b() << "  diam=" <<
           MultiIDiam(xd,ode.dim,NORM_MAX) << endl;
   writediam(outdiam,xd,ode.dim);
   outdiam << "the number of representation switches = " << int((estm.num_of_rep_switches).right_b()) << endl;
 }
}



bool incl_Lo_comp_traj(InclOde& ode,const InclPoincareParType &param, Set &x,
           SetWithTail & y0, InclEventManAlongTrajType & trajHandler )
//  computes rigorously the trajectory of x \oplus y0 using the Lohner
//       algorithm   for a differential inclusion <ode>
//       where y0 is an initial value for y - for true differential inclusions
//    this parameter can be ignored, but it is important for PDEs with
//       the evolving tail
//  Output:
//    returns true - when the condition from trajHandler   is satisfied
//              when the computation of the trajectory ended
//            false - otherwise


{
  interval h=param.h;  // this is a value of time step used in computations
  double tmax=param.tmax;

  interval rem[MAX_DIM+1];
  double remMax=0;

  interval xencl[MAX_DIM+1];  // used for enclosures of \varphi([0,h],x)


  Set xn;
  interval xd[MAX_DIM+1],xdn[MAX_DIM+1]; // x \subset xd
                                      // xn \subset xdn
  x.intvencl(xd);
  SetWithTail  y(y0);
  SetWithTail yn(ode.dim),yencl(ode.dim);

  Set xs;  // this is for deterministic equation x'=f(x) + y_c
  interval xds[MAX_DIM+1], yc[MAX_DIM+1], w1_encl[MAX_DIM+1];

  double nextshowinfo= trajHandler.nextShowInfo(0);
  double nextcondcheck=trajHandler.nextCheckCondition(0);

  interval t=0;
  long step_num=0;
  bool ok=true;
  bool finished=false;
  bool conditionSatisfied=false;

  while(!finished && ok){
    step_num++;
    t+=h;
    //   cout << t << endl;


    ok=incl_OneStepLohner(ode,param.onesteppar,h,x,xd,y,xn,xdn,yn,xencl,yencl,
           xs,xds,yc,w1_encl,rem);
     // xencl \times yencl =enclosure for \varphi([0,h],x,y)
     //   xn \times yn = enclosure for \varphi(h,x,y)
     // xd=enclosure for x = eval(x)
     // xdn=eval(xn) enclosure for  xn



    if(!ok) {
      infoInclOneStepFailed(x,xd,y,ode.dim,t,h, "incl_Lo_comp_traj");
      continue;
    }

    x=xn;
    y=yn;
    copyvector(xdn,xd,ode.dim);
    x.rearrangement(xd);


    if (t.right_b() >= nextshowinfo){
         trajHandler.showInfo(t,ode,h,x,xd,y,xn,xdn,yn,xencl,yencl,
               xs,xds,yc,w1_encl,rem);
         nextshowinfo=trajHandler.nextShowInfo(t.right_b());
    }

    if( t.right_b()>=nextcondcheck){
      conditionSatisfied=trajHandler.sucessCondition(t,ode,h,x,xd,y,xn,xdn,
         yn,xencl,yencl,xs,xds,yc,w1_encl,rem);

      if(conditionSatisfied) finished=true;
      else   nextcondcheck=trajHandler.nextCheckCondition(t.right_b());
    }


    if(t.right_b()>=tmax) finished=true;
  } // end while

  return conditionSatisfied;

}






/// the functions for the Poincare map for differential inclusions - BEGIN


InclPoincareMap::InclPoincareMap()
// default constructor
{
  destroy_Insection=0;
  pInsection=NULL;
  destroy_Outsection=0;
  pOutsection=NULL;
}


InclPoincareMap::InclPoincareMap(InclOde *ppode,GeneralLinearSection *ppsection,int psecdir,
        const InclPoincareParType &pparam,int pdestroy_section /*=0*/)
{
  dim=ppode->dim-1;
  pode=ppode;
  pInsection=ppsection;
  destroy_Insection=pdestroy_section;

  pOutsection=ppsection;  // the same section as pInsection
  destroy_Outsection=0;

  secdir=psecdir;

  param=pparam;
}


InclPoincareMap::InclPoincareMap(InclOde *ppode,GeneralLinearSection *ppInsection,
                   GeneralLinearSection *ppOutsection,int psecdir,
                   const InclPoincareParType &pparam,
                   int pdestroy_Insection/*=0 */,int  pdestroy_Outsection /*=0 */)
{
  dim=ppode->dim-1;
  pode=ppode;
  pInsection=ppInsection;
  destroy_Insection=pdestroy_Insection;

  pOutsection=ppOutsection;
  destroy_Outsection=pdestroy_Outsection;

  if(destroy_Outsection){
    if(pInsection==pOutsection)
      destroy_Outsection=!destroy_Insection;
  }


  secdir=psecdir;
  param=pparam;
}



InclPoincareMap::~InclPoincareMap()
{
  if(destroy_Outsection)
    destroy_Outsection=(pInsection!=pOutsection);

  if(destroy_Insection)
    delete pInsection;

  if(destroy_Outsection)
    delete pOutsection;
}



int InclPoincareMap::imap( Set &x,SetWithTail &y, interval fx[],SetWithTail &fy,
              estimates *estm)
       // on output fx \times fy =f(x \times y)
{
  return imapCCE(x,y,fx,fy,0,(LinCC *) NULL,0,(LinCC *) NULL,estm);
}


int InclPoincareMap::imap(Set &x, SetWithTail &y, interval fx[],SetWithTail & fy)
       // on output fx \times fy =f(x \times y)
{
  estimates estm;
  return  imap(x,y,fx,fy,&estm);
}


int  InclPoincareMap::imapCCE(Set &x, SetWithTail & y,interval fx[],SetWithTail & fy,
               int coordOnInput,LinCC* inputcoordch,int coordOnOutput,
               LinCC* outputcoordch, estimates *estm)
       //   on output: f((x + C*r0) \times y) = fx \times fy
       // usually this is Poincare map
       //  It makes sense sometimes to play with different coordinate
       //  systems
       //   if (coordOnInput) then the output (x-variable) is expressed
       //   using the other coordinate  (linear) system  given by inputcoordch
       //
       //   if (coordOnOutput) then fx is given in coordinates <outputcoorch>
       // this function  assumes that the coordinates on section z[1]=0
       //  are (z[2],...) - where z - are the coordinates in this->psection->plincc

{

  if(dim!=x.sdim()){
    wrongdiminfo(dim,x.sdim(),"InclPoincareMap::imapCCE");
    exit(EXIT_FAILURE);
  }

  Set *xfull=NULL;
  xfull=new Set;

  transOnSec_to_full(pInsection,coordOnInput,inputcoordch,x,*xfull); // transforms our set
                // on section to the ambient space coordinates

  interval *fxfull;
  fxfull=new interval[pode->dim+1];

  int delpccOut=0;
  LinCC* pccOut=(this->pOutsection)->plincc;
  if(coordOnOutput){
    // we have to extend pcoordch - to (dim+1) - dimensions
    // by inserting a new coordinate
    LinCC *extcoordch=insertcoord(outputcoordch,1,0.0);

    // now we compose extcoordch and pcc
    pccOut=composeLinCC(extcoordch,pccOut);

    delpccOut=1;
    delete extcoordch;
  }

  intervalmatrix1 *dxtrash;
  dxtrash=new intervalmatrix1[pode->dim+1];

  int retval=incl_Lo_Poincare(*pode,param,*xfull,y,*pOutsection,
                  secdir,estm,fxfull,fy,dxtrash, 1,pccOut);
  delete [] dxtrash;
  dxtrash=NULL;

  // we have to do the projection on section coordinates
  if(retval)
    dropcoord(fxfull,1,fx,dim+1);

  if(delpccOut)
    delete pccOut;


  delete [] fxfull;
  delete xfull;

  return retval;
}

int  InclPoincareMap::imapCC(Set &x, SetWithTail & y,interval fx[],SetWithTail & fy,
               int coordOnInput,LinCC* inputcoordch,int coordOnOutput,
               LinCC* outputcoordch)
{
 estimates estm;
 return imapCCE(x,y,fx,fy,coordOnInput,inputcoordch,coordOnOutput,
          outputcoordch,&estm);
}




double InclPoincareMap::isettimestep(double ph)
  //  returns  previous for interval computations timestep, and sets timesptep=ph
{
  double ret=param.h;
  param.h=ph;
  return ret;
}


double InclPoincareMap::igettimestep()
{
  return param.h;
}


double InclPoincareMap::settmax(double ptmax)
{
  double ret=param.tmax;
  param.tmax=ptmax;
  return ret;
}


int InclPoincareMap::isetorder(int porder)
  // returns previous value of order; and set order=porder and orderdf=order
  // interval computations
{
  int ret=param.onesteppar.order;
  param.onesteppar.order=porder;
  return(ret);
}


int InclPoincareMap::igetorder()
{
  return param.onesteppar.order;
}


double InclPoincareMap::gettmax()
{
  return param.tmax;
}

//  END of Poincare map for diff inclusions

int inclPoinMap_map_divide(InclPoincareMap& inclPMap,int divisiontable[],
       int divdim, interval ixb[],interval C[][MAX_DIM+1],interval r0[],int numLohner,
       SetWithTail & y,interval fx[],SetWithTail & fy,int coordOnInput,
       LinCC* inputcoordch,int coordOnOutput,LinCC* outputcoordch,estimates *estm)
// the function computes pmap.map(ixb + C*r0)
//  for this computation   the set ixb + C*r0  is first divided
//     into \Pi_{1<= i <=divdim}  divisiontable[i] parts
 //  the i-th direction in <r0> is divided into divisiontabl[i] equal parts
{
  int ok=1;
  int the_end=0;

  interval pr0[MAX_DIM+1];
  interval pixb[MAX_DIM+1];
  interval pfx[MAX_DIM+1];
  SetWithTail  pfy(y.dim);
  Set xset;

  int seq[MAX_DIM+1];
  for(int i=1;i<=divdim;i++)
    seq[i]=0;
  int firsttime=1;
  while(ok && !the_end){
    // we generate the new inital condition; we have to divide
    //  the set ixb + C*r0 into \Pi_{1<= i <=divdim}  divisiontable[i] parts
    //  the i-th direction in <r0> is divided into divisiontabl[i] equal parts
    divide_set(inclPMap.dim,seq,divisiontable,divdim,ixb,C,r0,pixb,pr0);

    xset.newvalues(inclPMap.dim,numLohner,pixb,C,pr0);
    ok=inclPMap.imapCCE(xset,y,pfx,pfy,coordOnInput,inputcoordch,coordOnOutput,
                 outputcoordch,estm);
    if(ok){
      // update the results:
      if(firsttime){
        copyvector(pfx,fx,inclPMap.dim);
        fy=pfy;
      } else{
        MultiIhull(fx,pfx,fx,inclPMap.dim);
        SwT_hull(fy,pfy,fy);
      }

      the_end=!gennewseqi(seq,divisiontable,1,divdim);
    }
    firsttime=0;
  }
  return ok;
}



////

