#ifndef _lohner_h
#define _lohner_h
#include"defo.h"
#include"zset.h"
#include"zdynsys.h"



int OneStepLohner(Ode &ode,const OneStepParamType &paronestep, interval h,
       Set &x,interval xd[],Set &xn, interval xdn[],interval xencl[],
       interval rem[]);
//  h - time step
// on INPUT : x  \subset xd
// on OUTPUT :
//       \varphi(h,x) \subset xn  \subset xdn
//    \varphi([0,h],x) \subset xencl
// rem; // remainder term in Taylor method, before multiplication
//         by h^(order+1)



int C1OneStepLohner(Ode &ode,const OneStepParamType &param, interval h,
       Set &x,interval xd[],Set & xn, interval xdn[],interval xencl[],
       interval rem[], C1Set &V, interval /*Vd*/[][MAX_DIM+1],
       C1Set &Vn,interval Vdn[][MAX_DIM+1], interval Vencl[][MAX_DIM+1],
       interval Vrem[][MAX_DIM+1]       );

// one step of C1-Lohner algorithm
//  h - a time step
// on INPUT : x  \subset xd, V \subset Vd
// on OUTPUT :
//       \varphi(h,x) \subset xn  \subset xdn
//    \varphi([0,h],x) \subset xencl
//      rem - a  remainder term in Taylor method, before the multiplication
//         by h^(order+1)

//  Variational data:
//  are treated analogously:
//  param.orderdf - an order of numerical method used for variational matrix
//  On OUTPUT:
//    d/dx \varphi(t+h,x0)=V(t+h,x0) \subset Vn \subset Vdn
//    d/dx \varphi(t+[0,h],x0)=V(t+[0,h],x0) \subset Vencl=[W4] - "C^1-Lohner alg."
// IT IS NOT OBVIOUS THAT THIS IS DESIRABLE - we may want W3
//  Vrem - remainder term for the variational matrix V(h,x), before
//           the multiplication   by h^{orderV+1} (includes already
//           the multiplication  by [W3] - an enclosure for V([0,h],x))
/////////////////////////////////////
//  currently only C^1_1 - algorithm is implemented, it does not
//           coincide with C^0-Lohner algorithm on x-variables

//// when numLohner==LOHNER1 - (Lohner evaluation 1)
////// then r0,C,Cn,B,Bn - are not used , in this case
// on INPUT : x + r \subset xd
//     in fact  xd= x + r (realised in computer arithemetic)
// on OUTPUT :
//       \varphi(h,x + r) \subset xn + rn \subset xdn
//     in fact  xdn= xn + rn (realised in computer arithmetic)
//    \varphi([0,h],x + r) \subset xecnl

///// numLohner== cuboid set - (QR-decomposition)
////// then r0,C,Cn - are not used , in this case
// on INPUT : x +  B*r \subset xd
//     in fact  xd= x +B*r (realised in computer arithmetic)
// on OUTPUT :
//       \varphi(h,x + B*r) \subset xn +  Bn*rn \subset xdn
//     in fact  xdn= xn +  Bn*rn (realized in computer arithmetic)
//    \varphi([0,h],x + B*r) \subset xecnl


//// numLohner== doubleton representation
// on INPUT : x + C*r0 + B*r \subset xd
//     in fact  xd= x + C*r0 + B*r (realised in computer arithmetic)
// on OUTPUT :
//       \varphi(h,x + C*r0 + B*rn) \subset xn + Cn*r0 + Bn*rn \subset xdn
//     in fact  xdn= xn + Cn*r0 + Bn*rn (realised in computer arithmetic)
//    \varphi([0,h],x + C*r0 + B*rn) \subset xecnl

// rem[]; // remainder term in Taylor method, before multiplication
//         per h^(order+1)

//  Variational data:
//  are treated analogously:
//   numLohnerV - LOHNER1, LOHNER3, LOHNER5
//  orderV - order of numerical method used for variational matrix
//  On INPUT:
//    d/dx \varphi(t,x0) \subset V \subset Vd
//  On OUTPUT
//    d/dx \varphi(t+h,x0)=V(t+h,x0) \subset Vn \subset Vdn
//    d/dx \varphi(t+[0,h],x0)=V(t+[0,h],x0) \subset Vencl=[W4] - "C^1-Lohner alg."
// IT IS NOT OBVIOUS THAT THIS IS DESIRABLE - we may want W3
//  Vrem - remainder term for the variational matrix V(h,x), before
//           the multiplication   by h^{orderV+1} (includes already
//           the multiplication  by [W3] - an enclosure for V([0,h],x))
//////  the use and meaning of other variables:
//  we represent variational matrix as
//   V_{k}  =VC_{k} + VE_{k}*V_0 + VB_k*VR_k
// then the update rule is:
//  A_{k+1,k}=A + Da; A - is a point matrix
//   V_{k+1}=A_{k+1,k}*V_k  = A*C_{k} + A*VE_k * V_0 + A_{k+1,k}*VB_k * VR_k
//                   + Da*(VC_k + VE_k*V_0)
//  if numLohnerV==LOHNER1 - interval evaluation
//     interval arithmetic evaluation of: V_{k+1}=A_{k+1,k}*V_k
//     Vc, Vcn - the only variables used
//    other variables are not used
//  numLohnerV==LOHNER3 - QR-decomposition
//    Ve,Ven,V0 - are not used
//    On input:
//       Vc + Vb*Vr \subset V
//    On output:
//       \varphi(h,Vc + Vb*Vr) \subset Vcn + Vbn*Vrn \subset Vn
//  numLohnerV==LOHNER5 - inner enclosure  with QR-decomposition
//    On input:
//      Vc + Ve*V0 + Vb*Vr \subset V
//    On output:
//       \varphi(h,Vc + Ve*V0 + Vb*Vr) \subset Vcn + Ven*V0 + Vbn*Vrn \subset Vn



int C0throughC1OneStepLohner(Ode &ode,const OneStepParamType & param, interval h,
       Set &x,interval xd[],Set &xn, interval xdn[],interval xencl[],
       interval rem[]);
// uses the C1-Lohner algorithm (C^1_1-algorithm) to make C0-time step, it should be
// useful in the context of dissipative ODEs
//  It does not coincide on x with the C^0-algorithm
// on INPUT : x  \subset xd
// on OUTPUT :
//       \varphi(h,x) \subset xn  \subset xdn
//    \varphi([0,h],x) \subset xencl
// rem; // remainder term in Taylor method, before multiplication
//         by h^(param.order+1)

#endif



