
#include<iostream>
using namespace std;

#include<cstdlib>

#include"defo.h"
#include"zdynsys.h"
#include"multiint.h"
#include"metnum.h"

#define GL_TAYLOR_ARRAYS  // if defined then we use file-wide defined arrays
                           // otherwise they are created at each call to taylor or dertaylor


long int newtontab[MAX_ORDER+2][MAX_ORDER+2];

int newtonsymbdone=0;
int maxorderinnewtontab=0;




#ifdef GL_TAYLOR_ARRAYS

int taylor_iarrays_done=0;
int dim_taylor_arrays=0;

// der[i][s] = x_i^{(s)}
intervalmatrix1maxorder *der=NULL;

// Dfij[i,j]=df_i/dx_j(x)
intervalmatrix1 *Dfij=NULL;


//  D2fijk[i,j,k]=d^2f_i/dxjxk(x)
intervalmatrix2 *D2fijk=NULL;

 // sder[j][k] = \Sigma_{s=0}^{r-3}newton(r-2,s)x_k^{(r-2-s)}x_j^{(1+s)}
intervalmatrix1 *sder=NULL;


 // Dlder[l,i,s]=\frac{\partial x_i^{(s)}}{\partial x_l}
intervalmatrix2maxorder *Dlder=NULL;


// Dlsder[l][j][k]= d/dx_l (sder[j][k])
intervalmatrix2 *Dlsder=NULL;

#endif // GL_TAYLOR_ARRAYS


void inittabnewton(int nmax)
{
  if(nmax > MAX_ORDER+1){
    cerr << "inittabnewton - order too large: " << nmax << endl;
    cerr << "it should be <= " << MAX_ORDER << endl;
     #if defined(KRAK) && defined(WIN95)
                   waitBt();
    #endif
    exit(EXIT_FAILURE);
  }
  for(int n=1;n<=nmax;n++)
    for(int k=0;k<=nmax;k++)
      newtontab[n][k]=0;

  newtontab[1][0]=1;
  newtontab[1][1]=1;

  for(int n=2;n<=nmax;n++){
    newtontab[n][0]=1;
    for(int k=1;k<=n;k++)
      newtontab[n][k]= newtontab[n-1][k-1] +  newtontab[n-1][k];
  }
  newtonsymbdone=1;
  maxorderinnewtontab=nmax;

}


int newton(int n,int k)
{
  return(newtontab[n][k]);
}


void remove_taylor_arrays()
{
  #ifdef GL_TAYLOR_ARRAYS
  if(taylor_iarrays_done){
    delete [] der;
    der=NULL;

    delete [] Dfij;
    Dfij=NULL;

    delete [] D2fijk;
    D2fijk=NULL;

    delete [] sder;
    sder=NULL;

    delete [] Dlder;
    Dlder=NULL;

    delete [] Dlsder;
    Dlsder=NULL;

    taylor_iarrays_done=0;
  }

  #endif
}

#ifdef GL_TAYLOR_ARRAYS
void init_taylor_arrays(int dim)
{
  if(dim > MAX_DIM){
    cerr << "init_taylor_arrays - order too large: " << dim << endl;
    cerr << "it should be <= " << MAX_DIM << endl;
     #if defined(KRAK) && defined(WIN95)
                   waitBt();
    #endif
    exit(EXIT_FAILURE);
  }
  if(dim > dim_taylor_arrays){
    remove_taylor_arrays();
  }

  der=new  intervalmatrix1maxorder[dim+1];

  Dfij=new  intervalmatrix1[dim+1];

  D2fijk=new intervalmatrix2[dim+1];

  sder=new intervalmatrix1[dim+1];

  Dlder=new intervalmatrix2maxorder[dim+1];

  Dlsder=new intervalmatrix2[dim+1];

  taylor_iarrays_done=1;
  dim_taylor_arrays=dim;
}
#endif



void taylor_init(int order,int dim)
{
  int callnewtoninit=0;
  if(!newtonsymbdone)
    callnewtoninit=1;
  else{
    if(maxorderinnewtontab < order +1){
      callnewtoninit=1;
    }
  }
  if(callnewtoninit) inittabnewton(order+1);


  #ifdef GL_TAYLOR_ARRAYS
  int callarraysint=0;
  if(!taylor_iarrays_done)
    callarraysint=1;
  else{
    if(dim > dim_taylor_arrays)
      callarraysint=1;
  }
  if(callarraysint)
    init_taylor_arrays(dim);
  #endif
}


int taylor(Ode &ode,int order, interval h, interval x[],interval y[],
           int calcrem, interval rem[])
// we assume that derivatives of the vector field higher than of 3-rd order vanish
 // order - order of the numerical method
 // h - time step
 // x - point at which moves under flow
 // y - its image after time h (computed only when !calcrem)
 //  rem - remainder from Taylor formula (before multiplication by h^{order +1}
 //       to have this computed correctly we need to use as <x> a rough enclosure.
 //        <rem> is  computed only when <calcrem>=1

 //
 //   to move forward by time step <h> we call taylor
 //   twice : 1) for x and calcrem=0
 //           2) for x=enclosure and calcrem=1


{

 int i,j,k,r;
 int max_order;
 interval hp;

 interval temp; // auxilary variable

 taylor_init(order,ode.dim);

 #ifndef GL_TAYLOR_ARRAYS
 // der[i][s] = x_i^{(s)}
 intervalmatrix1maxorder *der;
 der=new  intervalmatrix1maxorder[ode.dim+1];

 // Dfij[i,j]=df_i/dx_j(x)
 intervalmatrix1 *Dfij;
 Dfij=new  intervalmatrix1[ode.dim+1];

//  D2fijk[i,j,k]=d^2f_i/dxjxk(x)
 intervalmatrix2 *D2fijk;
 D2fijk=new intervalmatrix2[ode.dim+1];

 #endif

 int calckj[MAX_DIM+1][MAX_DIM+1];  // logical: c[j[k]] is 1 if
        //     D2fijk[i,j,k] is nonzero for some i

 for(k=1;k<=ode.dim;k++)
   for(j=1;j<=ode.dim;j++) calckj[k][j]=0;


  /* filling-in tables of partial derivates of vec. field  */
 for(i=1;i<=ode.dim;i++)
   for(j=1;j<=ode.dim; j++) Dfij[i][j]=ode.dvfield(i,j,x);




 for(i=1;i<=ode.dim;i++){
   for(j=1;j<=ode.dim; j++){
     for(k=1;k <= j; k++) {
       temp=ode.d2vfield(i,j,k,x);
       D2fijk[i][j][k]=temp;
       D2fijk[i][k][j]=temp;
       if(temp.left_b()!=0.0){ // this is enough because
                         // the function D2fijk[i][j][k] is a constant,
         calckj[k][j]=1;
         calckj[j][k]=1;
       }
     }
   }
 }



 /* calculate first derivatives of x */
 for(i=1; i<= ode.dim; i++) der[i][1]=ode.vfield(i,x);

 /* calculate second derivatives */
 for(i=1; i<= ode.dim; i++){
   temp=0;
   for(j=1;j <= ode.dim; j++) temp+=Dfij[i][j]*der[j][1];
   der[i][2]= temp;
 }


 if (calcrem) max_order=order + 1;
 else max_order=order;

 /* regular formula for higher derivates */

 #ifndef GL_TAYLOR_ARRAYS
 // interval sder[MAX_DIM + 1][MAX_DIM + 1];
 intervalmatrix1 *sder;
 sder=new intervalmatrix1[ode.dim+1];
 #endif

 interval tt;
 for(r=3; r <= max_order; r++){
   tt=0;

   // computation of: sder[k][j]=
   //           \Sigma_{s=0}^{r-2} newton(r-2,s) x_k^{(r-2-s)} x_j^{(1+s)}
   for(j=1;j<=ode.dim;j++){
     for(k=1;k<=ode.dim;k++){
       if(calckj[k][j]){     // this means that one D2fijk's is non zero
                              // for some i
         int s;
         tt=0.0;
         for (s=0;s<=r-3;s++)
           tt+=newton(r-2,s)*der[k][r-2-s]*der[j][1+s];
         sder[k][j]=tt;
       }
       else sder[k][j]=0.0;
     } // end for k
   } // end for j

   for(i=1;i <= ode.dim; i++){
     temp=0;
     for(j=1; j<= ode.dim; j++) temp+=Dfij[i][j]*der[j][r-1];
         /*temp = df_i/dx_j *x_j^{r-1}  */

     tt=0;
     for(j=1;j<=ode.dim; j++){
       for(k=1;k <= ode.dim; k++){
         if(D2fijk[i][j][k].left_b()!=0.0)
           tt+=D2fijk[i][j][k]*sder[k][j];
       } // end for k
     } // end for j
     der[i][r]=temp + tt;
   } // end for i
 }  // endfor r
 #ifndef GL_TAYLOR_ARRAYS
 delete[] sder;
 sder=NULL;
 #endif


 if (!calcrem){
   /* calculation of y */
   hp=1.0;

   for(i=1;i<=ode.dim;i++)
     y[i]=x[i];

   for(r=1;r <= order; r++){
     hp*=h;
     hp/=r;    /* hp=h^r/(r!) */
     for(i=1; i <= ode.dim; i++)
         y[i]+=hp*der[i][r];

   }
 }

 if (calcrem){
   /* an estimation of remainder term */
   for(i=1;i<= ode.dim; i++)
     rem[i]=der[i][order+1];

   for(r=2;r<=order+1;r++)
     for(i=1;i<=ode.dim; i++)  rem[i]/=r;
 }

 #ifndef GL_TAYLOR_ARRAYS
 delete [] D2fijk;
 delete [] Dfij;
 delete [] der;
 #endif
 return(1);
}




int dertaylor(Ode &ode,int order, interval h, interval x[],
       interval Vencl[][MAX_DIM+1],interval y[], interval Dx[][MAX_DIM+1],
       int calcrem, interval DxRem[][MAX_DIM+1])
 // order - an order of the numerical method
 // h - a time step
 // x - an interval vector on which we calculate the derivative of Taylor method
 //      when <calcrem>=0 - then this is our value of [x_k], otherwise
 //      it is [W_1] - enclosure for \varphi([0,h],[x_k])
 // Vencl - an interval matrix , used only when <calcrem> -
 //          this an enclosure of \frac{\partial \varphi}{\partial x}([0,h],x)=
 //               V([0,h],x); this a matrix W3 in 'C1-Lohner algorithm'
 //        otherwise is ignored (which is equivalent to taking Vencl=Id)
 //  calcrem - 1 - we compute remainder term - dertalor of order <order+1>
// ON OUTPUT:
 // y - an enclosure for  TaylorExpansion(h,x) up to <order>
 // Dx - a derivative of the Taylor method  : D TaylorExpansion(h,x) up to <order>
 // DxRem - the remainder term, before a multiplication by h^{order+1}:
  //  d^{order}/dt^{order}( \frac{\partial f}{\partial x} V)([x],id)*Vecnl/(order+1)!
 //  y,Dx are computed only when <!calcrem>
 //  DxRem is computed only when <calcrem>

 // we suppose that higher derivates (>2) of vector field vanish

//  for C^0-Lohner - calcrem=0
//  to use this function in C1-Lohner we need to call this function twice
//   1) - for x (our starting set) calcrem=0 - we compute Taylor expansion
//               for variational matrix
//   2) for x=enclosure of \varphi([0,h],x),
//    Vencl=W3=enclosure of
//         \frac{\partial \varphi}{\partial x}([0,h],x)=V([0,h],x)
//          and calcrem=1
//////////////
// this is an implementation of Thm. 10 and Thm. 11 from "C^1-Lohner algorithm"
// by Piotr Zgliczynski, Foundations of Computational Mathematics (2002) 2:429--465
{

 taylor_init(order,ode.dim);

 int max_order;
 interval temp;

 #ifndef GL_TAYLOR_ARRAYS
 // interval der[MAX_DIM+1][MAX_ORDER+2]; /* der[i][s] = x_i^{(s)}*/
 intervalmatrix1maxorder *der;
 der=new intervalmatrix1maxorder[ode.dim+1];


 // interval Dfij[MAX_DIM+1][MAX_DIM+1]; /* Dfij[i,j]=df_i/dx_j(x) */
 intervalmatrix1 *Dfij;
 Dfij=new intervalmatrix1[ode.dim+1];


 // interval D2fijk[MAX_DIM+1][MAX_DIM+1][MAX_DIM+1] ; /* D2fijk[i,j,k]=d^2f_i/dxixj (x)*/
 intervalmatrix2 *D2fijk;
 D2fijk=new intervalmatrix2[ode.dim+1];


//  interval Dlder[MAX_DIM+1][MAX_DIM+1][MAX_ORDER+2];
//  Dlder[l,i,s]=\frac{\partial x_i^{(s)}}{\partial x_l}
 intervalmatrix2maxorder *Dlder;
 Dlder=new intervalmatrix2maxorder[ode.dim+1];
 #endif

 int calckj[MAX_DIM+1][MAX_DIM+1];  // logical: tells if  D2fijk[i,j,k] is nonzero
           // for some i

 for(int k=1;k<=ode.dim;k++)
   for(int j=1;j<=ode.dim;j++) calckj[k][j]=0;



  // filling-in tables of partial derivates of vec. field
 for(int i=1;i<=ode.dim;i++)
    for(int j=1;j<=ode.dim; j++) Dfij[i][j]=ode.dvfield(i,j,x);




 for(int i=1;i<=ode.dim;i++)
   for(int j=1;j<=ode.dim; j++)
     for(int k=1;k <= j; k++){
       temp=ode.d2vfield(i,j,k,x);
       D2fijk[i][j][k]=temp;
       D2fijk[i][k][j]=temp;
       if(temp.left_b()!=0.0){ // this is enough because
                         // the function D2fijk[i][j][k] is a constant,
          calckj[k][j]=1;
          calckj[j][k]=1;
       }
     }




 // calculation of first derivatives of x
 for(int i=1; i<= ode.dim; i++) der[i][1]=ode.vfield(i,x);

 // calculation of partial derivatives of first derivatives of x
 for(int l=1; l<= ode.dim; l++){
   for(int i=1;i<=ode.dim; i++){
     Dlder[l][i][1]=Dfij[i][l];
   }
 }


 // we calculate second derivatives
 for(int i=1; i<= ode.dim; i++){
   temp=0;
   for(int j=1;j <= ode.dim; j++) temp+=Dfij[i][j]*der[j][1];
   der[i][2]= temp;
 }

 // we calculate partial derivatives of second derivatives
 for(int l=1;l<=ode.dim; l++){
   for(int i=1; i<= ode.dim; i++){  // d x^{(2)}_i/ dx_l
     temp=0;
     for(int j=1; j<= ode.dim; j++){ // remember that: D_lx_k^{(0)}=\delta_{l,k}
       temp+=D2fijk[i][j][l]*der[j][1] + Dfij[i][j]*Dlder[l][j][1];
     }
     Dlder[l][i][2]=temp;
   }
 }

 max_order=order;
 if(calcrem)
   max_order=order+1;

 #ifndef GL_TAYLOR_ARRAYS
// interval sder[MAX_DIM + 1][MAX_DIM + 1];
   // sder[j][k] = \Sigma_{s=0}^{r-3}newton(r-2,s)x_k^{(r-2-s)}x_j^{(1+s)}
 intervalmatrix1 *sder;
 sder=new intervalmatrix1[ode.dim+1];

 //  interval Dlsder[MAX_DIM+1][MAX_DIM+1][MAX_DIM+1];
   // Dlsder[l][j][k]= d/dx_l (sder[j][k])
 intervalmatrix2 *Dlsder;
 Dlsder=new intervalmatrix2[ode.dim+1];
 #endif
 // a regular formula for higher derivates
 for(int r=3; r <= max_order; r++){

   // a calculation of sder
   // sder[j][k] = \Sigma_{s=0}^{r-3}newton(r-2,s)x_k^{(r-2-s)}x_j^{(1+s)}
   for(int j=1;j<=ode.dim;j++){
     for(int k=1;k<=ode.dim;k++){
       if(calckj[k][j]){      // this means that one D2fijk's is non zero
                              // for some i; and a_{rkj} will effectively enter
                              //  into computations
         interval tt=0.0;
         for (int s=0;s<=r-3;s++)
           tt+=newton(r-2,s)*der[k][r-2-s]*der[j][1+s];
         sder[k][j]=tt;
       } else sder[k][j]=0.0;
     }
   }
   // end of sder calculation

   // calculation of d/dx_l sder = Dlsder
   //    Dlsder[l][j][k]=\Sum_{s=0}^{r-2} newton(r-2,s)(d/d_xl x_k^{(r-2-s)}*
   //     x_j^{(1+s)} + x_k^{(r-2-s)}d/dx_l x_j^{(1+s)})
   for(int l=1;l<=ode.dim; l++){
     for(int j=1;j<=ode.dim;j++){
       for(int k=1;k<=ode.dim;k++){
         if(calckj[k][j]){     // this means that one D2fijk's is non zero
                              // for some i; and D a_{rkj} will effectively enter
                              //  into computations
           interval tt=0.0;
           for (int s=0;s<=r-3;s++)
              tt+=newton(r-2,s)*(Dlder[l][k][r-2-s]*der[j][1+s] +
                    der[k][r-2-s]*Dlder[l][j][1+s]);
           Dlsder[l][k][j]=tt;
         } else Dlsder[l][k][j]=0.0;
       }
     } // end for j
   }
   // end of Dlsder calculation


   // calculation of der[i][r] = x_i^{(r)}
   for(int i=1;i <= ode.dim; i++){
     temp=0;
     for(int j=1; j<= ode.dim; j++) temp+=Dfij[i][j]*der[j][r-1];
         /*temp = df_i/dx_j *x_j^{r-1}  */

     interval tt=0;
     for(int j=1;j<=ode.dim; j++)
       for(int k=1;k <= ode.dim; k++)
         tt+=D2fijk[i][j][k]*sder[k][j];

     der[i][r]=temp + tt;
   }
   // end of calculation of der[i][r]


   // calculation of Dlder[]= d/dx_l (x_i^{(r)})
   for(int l=1; l<= ode.dim; l++){
     for(int i=1;i <= ode.dim; i++){
       temp=0;
       for(int j=1; j<= ode.dim; j++)  // remember that: D_lx_k^{(0)}=\delta_{l,k}
         temp+=D2fijk[i][j][l]*der[j][r-1];

       for(int j=1; j<= ode.dim; j++)
         temp+= Dfij[i][j]*Dlder[l][j][r-1];


       for(int j=1;j<=ode.dim; j++)
         for(int k=1;k <= ode.dim; k++)
           temp+=D2fijk[i][j][k]*Dlsder[l][k][j];

       Dlder[l][i][r]=temp;
     }
   }


 } // end for r
 #ifndef GL_TAYLOR_ARRAYS
 delete [] Dlsder;
 Dlsder=NULL;
 delete [] sder;
 sder=NULL;
 #endif


 if(!calcrem){
   // calculation of Dx
   interval hp=1.0;

   // 0-th order term is Id:
   identity(Dx,ode.dim);

   for(int r=1;r <= order; r++){
     hp*=h;
     hp/=r;    // hp=h^r/(r!)
     for(int i=1; i <= ode.dim; i++)
       for(int l=1;l <= ode.dim; l++)
         Dx[i][l]+=hp*Dlder[l][i][r];

   }



   /* calculation of y */

   for(int i=1;i<=ode.dim;i++)
     y[i]=x[i];

   hp=1.0;
   for(int r=1;r <= order; r++){
     hp*=h;
     hp/=r;    /* hp=h^r/(r!) */
     for(int i=1; i <= ode.dim; i++)
       y[i]+=hp*der[i][r];
   }
   // end of y - calculation
 }

 if(calcrem){
   interval invf=1.0;
   for(int i=2;i<=order+1;i++)
     invf/=i;
   //  invf=1/(order+1)!

   for(int i=1;i<=ode.dim;i++)
     for(int l=1;l<=ode.dim;l++)
       DxRem[i][l]=Dlder[l][i][order+1]*invf;

   ArrArr(DxRem,Vencl,DxRem,ode.dim); //  DxRem=DxRem*Vencl
 }

 #ifndef GL_TAYLOR_ARRAYS
 delete [] Dlder;
 delete [] Dfij;
 delete [] der;
 delete [] D2fijk;
 #endif

 return(1);
}



//*******************************

//*** double version *****


int taylor(Ode &ode,int order, double h, double x[],double y[],int calcrem,
           double rem[])
 /* order - order of the numerical method */
 /* h - time step */
 /* x - point at which moves under flow */
 /* y - its image after time h */
  // we suppose that derivatives of the vector field higher than 3 order vanish
 /* we suppose that higher derivates of vector field vanish */
 /*  rem - remainder from Taylor formula (before multiplication by h^{order +1}*/
{

 int callnewtoninit=0;
 if(!newtonsymbdone)
   callnewtoninit=1;
 else{
   if(maxorderinnewtontab < order +1){
     callnewtoninit=1;
   }
 }
 if(callnewtoninit) inittabnewton(order+1);


 int i,j,k,r;
 int max_order;
 double hp;

 double der[MAX_DIM+1][MAX_ORDER+2]; /* der[i][s] = x_i^{(s)}*/
 double Dfij[MAX_DIM+1][MAX_DIM+1]; /* Dfij[i,j]=df_i/dx_j(x) */
 double D2fijk[MAX_DIM+1][MAX_DIM+1][MAX_DIM+1] ;  /* D2fijk[i,j,k]=d^2f_i/dxixj (x)*/



  /* filling-in tables of partial derivates of vec. field  */
 for(i=1;i<=ode.dim;i++)
   { for(j=1;j<=ode.dim; j++) Dfij[i][j]=ode.dvfield(i,j,x); }




 for(i=1;i<=ode.dim;i++)
   for(j=1;j<=ode.dim; j++)
     for(k=1;k <= j; k++){
       double t= ode.d2vfield(i,j,k,x);
       D2fijk[i][j][k]=t;
       D2fijk[i][k][j]=t;
     }



 /* calculate first derivatives of x */
 for(i=1; i<= ode.dim; i++) der[i][1]=ode.vfield(i,x);

 /* calculate second derivatives */
 for(i=1; i<= ode.dim; i++)
  {
    double temp=0;
    for(j=1;j <= ode.dim; j++) temp+=Dfij[i][j]*der[j][1];
    der[i][2]= temp;
  }


 if (calcrem) max_order=order + 1;
 else max_order=order;

 /* regular formula for higher derivates */
 double sder[MAX_DIM + 1][MAX_DIM + 1];
 double tt;
 for(r=3; r <= max_order; r++)
   {
     for(j=1;j<=ode.dim;j++)
       for(k=1;k<=ode.dim;k++)
         {
           int s;
           tt=0.0;
           for (s=0;s<=r-3;s++)
              tt+=newton(r-2,s)*der[k][r-2-s]*der[j][1+s];
           sder[k][j]=tt;
          }

     for(i=1;i <= ode.dim; i++)
       {
         double temp=0;
         for(j=1; j<= ode.dim; j++) temp+=Dfij[i][j]*der[j][r-1];
         /*temp = df_i/dx_j *x_j^{r-1}  */

         tt=0;
         for(j=1;j<=ode.dim; j++)
            for(k=1;k <= ode.dim; k++)
              tt+=D2fijk[i][j][k]*sder[k][j];

         der[i][r]=temp + tt;
       }
   }

 if (!calcrem){
   /* calculation of y */
   hp=1.0;

   for(i=1;i<=ode.dim;i++)
     y[i]=x[i];

   for(r=1;r <= order; r++)
     {
       hp*=h;
       hp/=r;    /* hp=h^r/(r!) */
       for(i=1; i <= ode.dim; i++)
         y[i]+=hp*der[i][r];

     }
 }

 if (calcrem){
   /* estimation of remainder term */
   for(i=1;i<= ode.dim; i++)
     rem[i]=der[i][order+1];

   for(r=2;r<=order+1;r++)
     for(i=1;i<=ode.dim; i++)  rem[i]/=r;
 }

 return(1);
}



int dertaylor(Ode &ode,int order, double h, double x[],
       double Vencl[][MAX_DIM+1],double y[], double Dx[][MAX_DIM+1],
       int calcrem, double DxRem[][MAX_DIM+1])
 // order - an order of the numerical method
 // h - a time step
 // x - an double vector on which we calculate the derivative of Taylor method
 //      when <calcrem>=0 - then this is our value of [x_k], otherwise
 //      it is [W_1] - enclosure for \varphi([0,h],[x_k])
 // Vencl - an double matrix , used only when <calcrem> -
 //          this an enclosure of \frac{\partial \varphi}{\partial x}([0,h],x)=
 //               V([0,h],x); this a matrix W3 in 'C1-Lohner algorithm'
 //        otherwise is ignored (which is equivalent to taking Vencl=Id)
 //  calcrem - 1 - we compute remainder term - dertalor of order <order+1>
// ON OUTPUT:
 // y - an enclosure for  TaylorExpansion(h,x) up to <order>
 // Dx - a derivative of the Taylor method : D TaylorExpansion(h,x) up to <order>
 // DxRem - the remainder term, before a multiplication by h^{order+1}:
  //  d^{order}/dt^{order}( \frac{\partial f}{\partial x} V)([x],id)*Vecnl/(order+1)!
 //  y,Dx are computed only when <!calcrem>
 //  DxRem is computed only when <calcrem>

 // we suppose that higher derivates (>2) of vector field vanish

//  for C^0-Lohner - calcrem=0
//  to use this function in C1-Lohner we need to call this function twice
//   1) - for x (our starting set) calcrem=0 - we compute Taylor expansion
//               for variational matrix
//   2) for x=enclosure of \varphi([0,h],x),
//    Vencl=W3=enclosure of
//         \frac{\partial \varphi}{\partial x}([0,h],x)=V([0,h],x)
//          and calcrem=1
//////////////
// this is an implementation of Thm. 10 and Thm. 11 from "C^1-Lohner algorithm"
// by Piotr Zgliczynski, Foundations of Computational Mathematics (2002) 2:429--465
{

 taylor_init(order,ode.dim);

 int max_order;
 double temp;

// #ifndef GL_TAYLOR_ARRAYS
 // double der[MAX_DIM+1][MAX_ORDER+2]; /* der[i][s] = x_i^{(s)}*/
 doublematrix1maxorder *der;
 der=new doublematrix1maxorder[ode.dim+1];


 // double Dfij[MAX_DIM+1][MAX_DIM+1]; /* Dfij[i,j]=df_i/dx_j(x) */
 doublematrix1 *Dfij;
 Dfij=new doublematrix1[ode.dim+1];


 // double D2fijk[MAX_DIM+1][MAX_DIM+1][MAX_DIM+1] ; /* D2fijk[i,j,k]=d^2f_i/dxixj (x)*/
 doublematrix2 *D2fijk;
 D2fijk=new doublematrix2[ode.dim+1];


//  double Dlder[MAX_DIM+1][MAX_DIM+1][MAX_ORDER+2];
//  Dlder[l,i,s]=\frac{\partial x_i^{(s)}}{\partial x_l}
 doublematrix2maxorder *Dlder;
 Dlder=new doublematrix2maxorder[ode.dim+1];
// #endif

 int calckj[MAX_DIM+1][MAX_DIM+1];  // logical: tells if  D2fijk[i,j,k] is nonzero
           // for some i

 for(int k=1;k<=ode.dim;k++)
   for(int j=1;j<=ode.dim;j++) calckj[k][j]=0;



  // filling-in tables of partial derivates of vec. field
 for(int i=1;i<=ode.dim;i++)
    for(int j=1;j<=ode.dim; j++) Dfij[i][j]=ode.dvfield(i,j,x);




 for(int i=1;i<=ode.dim;i++)
   for(int j=1;j<=ode.dim; j++)
     for(int k=1;k <= j; k++){
       temp=ode.d2vfield(i,j,k,x);
       D2fijk[i][j][k]=temp;
       D2fijk[i][k][j]=temp;
       if(temp!=0.0){ // this is enough because
                         // the function D2fijk[i][j][k] is a constant,
          calckj[k][j]=1;
          calckj[j][k]=1;
       }
     }




 // calculation of first derivatives of x
 for(int i=1; i<= ode.dim; i++) der[i][1]=ode.vfield(i,x);

 // calculation of partial derivatives of first derivatives of x
 for(int l=1; l<= ode.dim; l++){
   for(int i=1;i<=ode.dim; i++){
     Dlder[l][i][1]=Dfij[i][l];
   }
 }


 // we calculate second derivatives
 for(int i=1; i<= ode.dim; i++){
   temp=0;
   for(int j=1;j <= ode.dim; j++) temp+=Dfij[i][j]*der[j][1];
   der[i][2]= temp;
 }

 // we calculate partial derivatives of second derivatives
 for(int l=1;l<=ode.dim; l++){
   for(int i=1; i<= ode.dim; i++){  // d x^{(2)}_i/ dx_l
     temp=0;
     for(int j=1; j<= ode.dim; j++){ // remember that: D_lx_k^{(0)}=\delta_{l,k}
       temp+=D2fijk[i][j][l]*der[j][1] + Dfij[i][j]*Dlder[l][j][1];
     }
     Dlder[l][i][2]=temp;
   }
 }

 max_order=order;
 if(calcrem)
   max_order=order+1;

// #ifndef GL_TAYLOR_ARRAYS
// double sder[MAX_DIM + 1][MAX_DIM + 1];
   // sder[j][k] = \Sigma_{s=0}^{r-3}newton(r-2,s)x_k^{(r-2-s)}x_j^{(1+s)}
 doublematrix1 *sder;
 sder=new doublematrix1[ode.dim+1];

 //  double Dlsder[MAX_DIM+1][MAX_DIM+1][MAX_DIM+1];
   // Dlsder[l][j][k]= d/dx_l (sder[j][k])
 doublematrix2 *Dlsder;
 Dlsder=new doublematrix2[ode.dim+1];
// #endif
 // a regular formula for higher derivates
 for(int r=3; r <= max_order; r++){

   // a calculation of sder
   // sder[j][k] = \Sigma_{s=0}^{r-3}newton(r-2,s)x_k^{(r-2-s)}x_j^{(1+s)}
   for(int j=1;j<=ode.dim;j++){
     for(int k=1;k<=ode.dim;k++){
       if(calckj[k][j]){      // this means that one D2fijk's is non zero
                              // for some i; and a_{rkj} will effectively enter
                              //  into computations
         double tt=0.0;
         for (int s=0;s<=r-3;s++)
           tt+=newton(r-2,s)*der[k][r-2-s]*der[j][1+s];
         sder[k][j]=tt;
       } else sder[k][j]=0.0;
     }
   }
   // end of sder calculation

   // calculation of d/dx_l sder = Dlsder
   //    Dlsder[l][j][k]=\Sum_{s=0}^{r-2} newton(r-2,s)(d/d_xl x_k^{(r-2-s)}*
   //     x_j^{(1+s)} + x_k^{(r-2-s)}d/dx_l x_j^{(1+s)})
   for(int l=1;l<=ode.dim; l++){
     for(int j=1;j<=ode.dim;j++){
       for(int k=1;k<=ode.dim;k++){
         if(calckj[k][j]){     // this means that one D2fijk's is non zero
                              // for some i; and D a_{rkj} will effectively enter
                              //  into computations
           double tt=0.0;
           for (int s=0;s<=r-3;s++)
              tt+=newton(r-2,s)*(Dlder[l][k][r-2-s]*der[j][1+s] +
                    der[k][r-2-s]*Dlder[l][j][1+s]);
           Dlsder[l][k][j]=tt;
         } else Dlsder[l][k][j]=0.0;
       }
     } // end for j
   }
   // end of Dlsder calculation


   // calculation of der[i][r] = x_i^{(r)}
   for(int i=1;i <= ode.dim; i++){
     temp=0;
     for(int j=1; j<= ode.dim; j++) temp+=Dfij[i][j]*der[j][r-1];
         /*temp = df_i/dx_j *x_j^{r-1}  */

     double tt=0;
     for(int j=1;j<=ode.dim; j++)
       for(int k=1;k <= ode.dim; k++)
         tt+=D2fijk[i][j][k]*sder[k][j];

     der[i][r]=temp + tt;
   }
   // end of calculation of der[i][r]


   // calculation of Dlder[]= d/dx_l (x_i^{(r)})
   for(int l=1; l<= ode.dim; l++){
     for(int i=1;i <= ode.dim; i++){
       temp=0;
       for(int j=1; j<= ode.dim; j++)  // remember that: D_lx_k^{(0)}=\delta_{l,k}
         temp+=D2fijk[i][j][l]*der[j][r-1];

       for(int j=1; j<= ode.dim; j++)
         temp+= Dfij[i][j]*Dlder[l][j][r-1];


       for(int j=1;j<=ode.dim; j++)
         for(int k=1;k <= ode.dim; k++)
           temp+=D2fijk[i][j][k]*Dlsder[l][k][j];

       Dlder[l][i][r]=temp;
     }
   }


 } // end for r
// #ifndef GL_TAYLOR_ARRAYS
 delete [] Dlsder;
 Dlsder=NULL;
 delete [] sder;
 sder=NULL;
// #endif


 if(!calcrem){
   // calculation of Dx
   double hp=1.0;

   // 0-th order term is Id:
   identity(Dx,ode.dim);

   for(int r=1;r <= order; r++){
     hp*=h;
     hp/=r;    // hp=h^r/(r!)
     for(int i=1; i <= ode.dim; i++)
       for(int l=1;l <= ode.dim; l++)
         Dx[i][l]+=hp*Dlder[l][i][r];

   }



   /* calculation of y */

   for(int i=1;i<=ode.dim;i++)
     y[i]=x[i];

   hp=1.0;
   for(int r=1;r <= order; r++){
     hp*=h;
     hp/=r;    /* hp=h^r/(r!) */
     for(int i=1; i <= ode.dim; i++)
       y[i]+=hp*der[i][r];
   }
   // end of y - calculation
 }

 if(calcrem){
   double invf=1.0;
   for(int i=2;i<=order+1;i++)
     invf/=i;
   //  invf=1/(order+1)!

   for(int i=1;i<=ode.dim;i++)
     for(int l=1;l<=ode.dim;l++)
       DxRem[i][l]=Dlder[l][i][order+1]*invf;

   ArrArr(DxRem,Vencl,DxRem,ode.dim); //  DxRem=DxRem*Vencl
 }

// #ifndef GL_TAYLOR_ARRAYS
 delete [] Dlder;
 delete [] Dfij;
 delete [] der;
 delete [] D2fijk;
// #endif

 return(1);
}




////////// RUNGE - KUTTA 4th order

void rk4(Ode &ode,double h, double x[],double y[])

 /* h - time step */
 /* x - point at which moves under flow */
 /* y - its image after time h */
     /*
        k1=f(x)
        k2=f(x+h/2*k1)
        k3=f(x+h/2*k2)
        k4=f(x+h*k3)
        xn=x + h*(k1+2k2+2k3 + k4)/6
      */

{
  double k1[MAX_DIM +1];
  double k2[MAX_DIM +1];
  double k3[MAX_DIM +1];
  double k4[MAX_DIM +1];
  double t[MAX_DIM+1];

  int i;
  for(i=1;i<= ode.dim; i++)
    k1[i]=ode.vfield(i,x);

  for(i=1;i<= ode.dim; i++)
    t[i]=x[i]+h*k1[i]/2;

  for(i=1;i<= ode.dim; i++)
    k2[i]=ode.vfield(i,t);


 for(i=1;i<= ode.dim; i++)
    t[i]=x[i]+h*k2[i]/2;


  for(i=1;i<= ode.dim; i++)
    k3[i]=ode.vfield(i,t);

 for(i=1;i<= ode.dim; i++)
    t[i]=x[i]+h*k3[i];


  for(i=1;i<= ode.dim; i++)
    k4[i]=ode.vfield(i,t);

  for(i=1; i<= ode.dim; i++)
    y[i]=x[i]+h*(k1[i]+2*k2[i]+2*k3[i] + k4[i])/6;

}



void rk4(Ode &ode,interval h, interval x[],interval y[])

 /* h - time step */
 /* x - point at which moves under flow */
 /* y - its image after time h */
     /*
        k1=f(x)
        k2=f(x+h/2*k1)
        k3=f(x+h/2*k2)
        k4=f(x+h*k3)
        xn=x + h*(k1+2k2+2k3 + k4)/6
      */

{
  interval k1[MAX_DIM +1];
  interval k2[MAX_DIM +1];
  interval k3[MAX_DIM +1];
  interval k4[MAX_DIM +1];
  interval t[MAX_DIM+1];

  int i;
  for(i=1;i<= ode.dim; i++)
    k1[i]=ode.vfield(i,x);

  for(i=1;i<= ode.dim; i++)
    t[i]=x[i]+h*k1[i]/2;

  for(i=1;i<= ode.dim; i++)
    k2[i]=ode.vfield(i,t);


 for(i=1;i<= ode.dim; i++)
    t[i]=x[i]+h*k2[i]/2;


  for(i=1;i<= ode.dim; i++)
    k3[i]=ode.vfield(i,t);

 for(i=1;i<= ode.dim; i++)
    t[i]=x[i]+h*k3[i];


  for(i=1;i<= ode.dim; i++)
    k4[i]=ode.vfield(i,t);

  for(i=1; i<= ode.dim; i++)
    y[i]=x[i]+h*(k1[i]+2*k2[i]+2*k3[i] + k4[i])/6;

}



void OneStepVar(Ode &ode,double h,double x[], double Dx[][MAX_DIM+1],
     double Dxn[][MAX_DIM+1], int dimVar,int order)
// performs one step for Jacobian matrix of the flow
//  using Taylor method
// INPUT values:
//  x - current position
//  Dx - current value of Jacobian matrix
//  dimVar - we will do Galerkin proj. onto first <dimVar> -variables
// OUTPUT VALUE:
//  Dxn - value of Jacobian matrix after time +<h>

{
  int prevdim=ode.dim;
  ode.dim=dimVar;
  double y[MAX_DIM+1];
  int calcrem=0;
  double Dxh[MAX_DIM+1][MAX_DIM+1];
  dertaylor(ode,order,h,x,NULL,y,Dxh,calcrem,NULL);
  ArrArr(Dxh,Dx,Dxn,ode.dim);
  ode.dim=prevdim;
}



