
#include<ostream>
#include<fstream>

#include<sstream>
using namespace std;

#include<cstdlib>

#include"defo.h"
#include"multiint.h"
#include"zgfiles.h"
#include"zgmatrix.h"
#include"qrdecmp.h"


#define TESTY
// 6.VIII.2008 - to check how the clustering process of eigenvalues goes


void normalizetransf(double p[][MAX_DIM+1], double pinv[][MAX_DIM+1],int coord,
                  int pdim,double n)
// function modifies p,pinv - describing a coordinate change
//  the <coord>-column of <p> will be divided by <n>
//  the <coord>-row of pinv will be multiplied by <n>
//  the condition: pinv*p = id is preserved
//  <pdim> - the dimension of matrices
{
  for(int i=1;i<=pdim;i++){
    p[i][coord]/=n;
    pinv[coord][i]*=n;
  }
}


int normalizetransf(double p[][MAX_DIM+1],
  double pinv[][MAX_DIM+1],const double eigenval[][3],int pdim,
   int normfunc)
// function modifies p,pinv - describing coordinate change to make them
//  more ballanced, i.e the column of <p>, will have a norm equal to 1
//  for single eigenvalues; and if we have a pair then the longer vector
//  will have this norm
//  the condition: pinv*p = id is preserved
//  <eigenval> - serve only to decide if we have pair or single eigenvector
//  <pdim> - the dimension of matrices
//  <normfunc> - the type of norm: possible values: NORM_EUCL, NORM_MAX, NORM_SUM
{
  double x[MAX_DIM+1];
  int ok=1;
  int coord=1;
  while((coord <=pdim) && ok){
    if(eigenval[coord][2]==0){ // simple eigenvector
      for(int i=1;i<=pdim;i++)
        x[i]=p[i][coord];
      double n=norm(x,pdim,normfunc);
      if(n==0){
        ok=0;
        cerr << "normalizetransf: one eigenvector has  zero norm " << endl;
        #ifdef WIN95
           cerr << "press any key " << endl;
        #endif
        break;
      }
      normalizetransf(p,pinv,coord,pdim,n);

      coord++;
    } else {  // 2x2 cell
      for(int i=1;i<=pdim;i++)
        x[i]=p[i][coord];
      double n=norm(x,pdim,normfunc);
      for(int i=1;i<=pdim;i++)
        x[i]=p[i][coord+1];
      double n2=norm(x,pdim,normfunc);

      if(n2>n) n=n2;
      if(n==0){
        ok=0;
        cerr << "normalizetransf: one eigenvector has  zero norm " << endl;
        #ifdef WIN95
           cerr << "press any key " << endl;
        #endif
        break;
      }
      normalizetransf(p,pinv,coord,pdim,n);
      normalizetransf(p,pinv,coord+1,pdim,n);

      coord+=2;
    }
  }
  return(ok);
}

void normalizetransf(interval p[][MAX_DIM+1],interval pinv[][MAX_DIM+1],
                    int coord,int pdim,interval n)
// function modifies p,pinv - describing coordinate change
//  the <coord>-column of <p> will be divided by <n>
//  the <coord>-row of pinv will be multiplied by <n>
//  the condition: pinv*p = id is preserved
//  <pdim> - the dimension of matrices
{
  for(int i=1;i<=pdim;i++){
    p[i][coord]/=n;
    pinv[coord][i]*=n;
  }
}


int normalizetransf(interval p[][MAX_DIM+1],
  interval pinv[][MAX_DIM+1],const double eigenval[][3],int pdim,
   int normfunc)
// function modifies p,pinv - describing coordinate change to make them
//  more ballanced, i.e the column of <p>, will have a norm equal to 1
//  for single eigenvalues; and if we have a pair then the longer vector
//  will have this norm
//  the condition: pinv*p = id is preserved
//  <eigenval> - serve only to decide if we have pair or single eigenvector
//  <pdim> - the dimension of matrices
//  <normfunc> - the type of norm: possible values: NORM_EUCL, NORM_MAX, NORM_SUM
{
  interval x[MAX_DIM+1];
  int ok=1;
  int coord=1;
  while((coord <=pdim) && ok){
    if(eigenval[coord][2]==0){ // simple eigenvector
      for(int i=1;i<=pdim;i++)
        x[i]=p[i][coord];
      interval n=norm(x,pdim,normfunc);
      if(!(n>0)){
        ok=0;
        cerr << "normalizetransf: one eigenvector has  zero norm " << endl;
        #ifdef WIN95
           cerr << "press any key " << endl;
        #endif
        break;
      }
      normalizetransf(p,pinv,coord,pdim,Right(n));

      coord++;
    } else {  // 2x2 cell
      for(int i=1;i<=pdim;i++)
        x[i]=p[i][coord];
      interval n=Right(norm(x,pdim,normfunc));
      for(int i=1;i<=pdim;i++)
        x[i]=p[i][coord+1];
      interval n2=Right(norm(x,pdim,normfunc));

      if(n2>n) n=n2;
      if(!(n > 0)){
        ok=0;
        cerr << "normalizetransf: one eigenvector has  zero norm " << endl;
        #ifdef WIN95
           cerr << "press any key " << endl;
        #endif
        break;
      }
      normalizetransf(p,pinv,coord,pdim,n);
      normalizetransf(p,pinv,coord+1,pdim,n);

      coord+=2;
    }
  }
  return(ok);
}



int gennewvectnorm(double u[][MAX_DIM+1],int vectnum,int pdim)
//  a function generates a new normal vector which is orthogonal
//   to the vectors u[i] for i=1..vectnum
//   this vector is added to the table u as u[vectnum+1]
{
  int ok=0;
  int cno=1;

  double smallprj=0.5/sqrt(double(pdim));

  double w[MAX_DIM+1];
  while(!ok && (cno <=pdim)){
    for(int i=1;i<=pdim;i++)
      w[i]=0.0;

    w[cno]=1.0;

    for(int vno=1;vno<=vectnum;vno++){
      orthogonalcomponent(w,u[vno],pdim);
    }
    double c=normEucl(w,pdim);
    if(c > smallprj){
      ok=1;
      vect_normalize(w,pdim,NORM_EUCL);
      for(int j=1;j<=pdim;j++)
         u[vectnum+1][j]=w[j];
    }
    cno++;
  }
  return ok;
}


int igennewvectnorm(interval u[][MAX_DIM+1],int vectnum,int pdim)
//  a function generates a new normal vector which is orthogonal
//   to the vectors u[i] for i=1..vectnum
//   this vector is added to the table u as u[vectnum+1]
{
  int ok=0;
  int cno=1;

  double smallprj=0.5/sqrt(double(pdim));

  interval w[MAX_DIM+1];
  while(!ok && (cno <=pdim)){
    for(int i=1;i<=pdim;i++)
      w[i]=0.0;

    w[cno]=1.0;

    for(int vno=1;vno<=vectnum;vno++){
      orthogonalcomponent(w,u[vno],pdim);
    }
    interval c=normEucl(w,pdim);
    if(c > smallprj){
      ok=1;
      vect_normalize(w,pdim,NORM_EUCL);
      for(int j=1;j<=pdim;j++)
         u[vectnum+1][j]=w[j];
    }
    cno++;
  }
  return ok;
}



void SortTab(const double t[],int Perm[],int pdim,int numunchanged)
//  a function sorts an input table TabNorm[1...pdim] in descending order
// OUTPUT:
//     table Perm[1..pdim]  satisfies:
//       Perm[i]=i for i<=numunchanged
//       t[Perm[i]] >= t[Perm[i+1]]  for i=numunchanged+1,...,pdim-1
{
  for(int i=1;i<=pdim;i++)
    Perm[i]=i;

  for(int i=numunchanged+1;i<=pdim;i++)
    for(int j=pdim-1;i<=j;j--){
      if(t[Perm[j+1]] > t[Perm[j]]){
        int k=Perm[j];
        Perm[j]=Perm[j+1];
        Perm[j+1]=k;
      }
    }
}



int iQRDecomp(const interval A[][MAX_DIM+1],interval Q[][MAX_DIM+1],
                int pdim)
// on output Q is an orthogonal factor in QR-decomposition for A
// if the procedure fails, then Q=id
// this a special QR-decomposition - required for Lohner algorthm
{
  // interval  U[MAX_DIM+1][MAX_DIM+1]; // an auxliary table at the end Q=U^T
  intervalmatrix1 *U;
  U=new intervalmatrix1[pdim+1];
  double TabNorm[MAX_DIM+1];
  int perm[MAX_DIM+1];

  int permute_inQR=0;     //   tells if we  permute before next step
          //vectors, before the generation of a new coordinate axis
          // it turns out that it is better (often much better) to
          //  keep original ordering - I don't understand fully reasons
          //  for this phenomenon
  int ok=1;

  // to look nice in debugger program, we define also U[i][0], but this is
                   // not used in the computations
  for(int i=1;i<=pdim;i++)
    U[i][0]=0.0;

  for(int i=1;i<=pdim; i++)
    for(int j=1;j<=pdim; j++)
      U[i][j]=A[j][i];

  // U[i] - i-th column of A


  for(int j=1;j<=pdim; j++)
    TabNorm[j]=normEucl(U[j],pdim).right_b();

  SortTab(TabNorm,perm,pdim,0);


  for(int i=1;((i<=pdim) && ok); i++){
    if (!vect_normalize(U[perm[i]],pdim,NORM_EUCL)){
      if(i==1){
        zero(U[perm[i]],pdim);
        U[perm[i]][1]=1.0;
      }else{
        intervalmatrix1 *uord;
        uord=new intervalmatrix1[i+1];
        for(int k=1;k<=i-1;k++){
          copyvector(U[perm[k]],uord[k],pdim);
        }
        if(igennewvectnorm(uord,i-1,pdim)){
          copyvector(uord[i],U[perm[i]],pdim);
        }else{
          char fname[80];
          strcpy(fname,catwork);
          strcat(fname,"qrprobl.txt");
          saveArray(fname,A,pdim);
          cerr << "iQRDecomp - problem " << endl;
          cerr << "see file [" << fname << "]" << endl;
          #if defined(WIN95) && defined(KRAK)
            waitBt();
          #endif
          ok=0;
        }
        delete [] uord;
      }
    }
    if(ok && (i<pdim)){
      for(int j=i+1;j<=pdim; j++){
        orthogonalcomponent(U[perm[j]],U[perm[i]],pdim); // from U([perm[j]]) we
                                    // we subtract its projection onto U[perm[i]]
      }
      if(permute_inQR){
        for(int j=i+1;j<=pdim; j++)
          TabNorm[j]=normEucl(U[perm[j]],pdim).right_b();

        int perm2[MAX_DIM+1],tperm[MAX_DIM+1];
        SortTab(TabNorm,perm2,pdim,i);

        // we need now to compose  permutations <Perm> and <perm2>
        //   perm \circ perm2

        for(int j=1;j<=pdim;j++)
          tperm[j]=perm[perm2[j]];

        for(int j=1;j<=pdim;j++)
          perm[j]=tperm[j];

      }
    }
  }


  for(int i=1;i<=pdim; i++)
    for(int j=1;j<=pdim; j++){
      if(ok) Q[i][j]=U[j][i];
      else{
        Q[i][j]=0.0;
        if(i==j) Q[i][j]=1.0;
      }
    }
  delete [] U;
  return(ok);
}


// QR - decomposition and QR algortithm


int qrdecomp(const double a[][MAX_DIM+1],double q[][MAX_DIM+1],
              double r[][MAX_DIM+1],  int pdim)
// on output Q  and R define a QR-decomposition for A
// if procedure fails Q=id
{

  double smalldouble=1e-12;

  double u[MAX_DIM+1][MAX_DIM+1]; // matrix of 'pseudoeigenvectors'
              //   for QR decomposition - the basis for new orthogonal
              //     coordinates q=u^t; u[i] - i-th vector

  int ok=1;
  zero(q,pdim);
  zero(r,pdim);
  zero(u,pdim);



  double w[MAX_DIM+1];
  for(int i=1;i<=pdim;i++){

     for(int j=1;j<=pdim;j++)
        w[j]=a[j][i];

     // w - is an i-th column of a
     for(int j=1;j<i;j++){
       r[j][i]=orthogonalcomponent(w,u[j],pdim);
     }
     double c=normEucl(w,pdim);

     if(c > smalldouble){
       r[i][i]=c;
       vect_normalize(w,pdim,NORM_EUCL);
       for(int j=1;j<=pdim;j++)
         u[i][j]=w[j];
     }else{
        r[i][i]=0.0;
        gennewvectnorm(u,i-1,pdim);
     }
  }

  if(ok){
    transpose(u,q,pdim);     // q=u^t
  }

  return(ok);
}





void  cleanmatrix(double q[][MAX_DIM+1],int rownum,int colnum,double bias)
{
  for(int i=1;i<=rownum;i++)
    for(int j=1;j<=colnum;j++){
      if(fabs(q[i][j]) < bias){
        q[i][j]=0.0;
      }
    }
}

void  icleanmatrix(interval q[][MAX_DIM+1],int rownum,int colnum,double bias)
{
  for(int i=1;i<=rownum;i++)
    for(int j=1;j<=colnum;j++){
      if(abs(q[i][j]) < bias){
        q[i][j]=0.0;
      }
    }
}


int qrimprove(const double a[][MAX_DIM+1],double q[][MAX_DIM+1],
       double a1[][MAX_DIM+1],int pdim)
//  the function 'orthogonalizes' - an almost orthogonal matrix <q>
//   and uses it as the coordinate change a1=q^t * a * q
{
  double t[MAX_DIM+1][MAX_DIM+1],qt[MAX_DIM+1][MAX_DIM+1];

  transpose(q,qt,pdim);
  int ok=orthogonalize(qt,pdim,pdim);
  transpose(qt,q,pdim);

  if(ok){
    ArrArr(qt,a,t,pdim);  // t=q^t a
    ArrArr(t,q,a1,pdim);  // a1=q^t a q
  }
  else{
    #if defined(WIN95) && defined(KRAK)
      rootFrame.Clear();
    #endif
    cerr << "qrimprove error - unable to orthogonalize " << endl;
    #if defined(WIN95) && defined(KRAK)
      cerr << "press any key " << endl;
      waitBt();
    #endif
    exit(EXIT_FAILURE);
  }

  return (ok);
}



int testbl2x2(const double a[][MAX_DIM+1],int /*pdim */,int cno,double *abslambda)
//   the procedure tests if the columns and rows <cno>, <cno+1> in <a>
// represent 2x2 block, with complex eigenvalues - if yes then 1 is returned
//     and in this case on output *abslambda is equal to absolute value
//     of the complex eigenvalue

{

  // we will test now if we have complex eigenvalue
  //  the characteristic polynomial  is: x^2 - t x + w
  //   w=det, t=trace
  double t=a[cno][cno] + a[cno+1][cno+1];
  double w=a[cno][cno]*a[cno+1][cno+1] - a[cno][cno+1]*a[cno+1][cno];
  double delta=t*t - 4*w;
  int ret=(delta < 0);

  if(ret)
    *abslambda=sqrt(w); // this is an absolute value of the eigenvalue

  return ret;
}




int testuptriangular(int doclear,  double ac[][MAX_DIM+1],int pdim,
    double bias,double relbias)
//  the function test if the <pdim x pdim> - matrix <ac> is in an upper
//   block 2x2 triangular form ; the entry in the matrix (below diagnonal)
//     is treated as
//    zero if its absolute value is less then <bias>  or
//     when  divided by the diagonal element is less relbias
//   - if this is the case then  it is reset to zero (if doclear)
{

  int ret=1;
  double a[MAX_DIM+1][MAX_DIM+1];
  copymatrix(ac,a,pdim);


  // in the first run we basically try to check elements a[j][i] with
  //  j - i >=2 that they are smal, and for j=i+1 we allow for something

  ///// then in the second run we look closer at elements a[i+1][i]

  int cno=1;
  while(ret && (cno < pdim)){
    int i=cno+1;
    while(ret && (i<=pdim)){
      int iok=0;
      if(fabs(a[i][cno]) < bias){
        a[i][cno]=0;
        iok=1;
      }
      if(fabs(a[i][cno]) < relbias*fabs(a[cno][cno])){
        a[i][cno]=0;
        iok=1;
      }
      if(!iok && (i > cno +1)){ // we check with possible blocks (cno-1,cno)
 //  (cno,cno+1)
        double abslambda;
        if(cno > 1){
          if(testbl2x2(a,pdim,cno-1,&abslambda)){
            if(fabs(a[i][cno]) < relbias*abslambda){
              iok=1;
              a[i][cno]=0;
            }
          }
 }
        if(!iok && (cno < pdim)){
          if(testbl2x2(a,pdim,cno,&abslambda)){
            if(fabs(a[i][cno]) < relbias*abslambda){
              iok=1;
              a[i][cno]=0;
            }
          }
        }
      }
      if(!iok && (i > cno+1)) ret=0;

      i++;
    }
    cno++;
  }

  // second run with full tests of 2x2 blocks, but only elements a[i+1][i]
  cno=1;
  while(ret && (cno < pdim)){
    if(fabs(a[cno+1][cno]) > 0){ // then it must block 2x2 and no blocks
                                 // for cno -1
      double abslambda;
      if(testbl2x2(a,pdim,cno,&abslambda)){
        if(cno > 1) ret=!testbl2x2(a,pdim,cno-1,&abslambda);
      } else ret=0;
    }
    cno++;
  }

  if(doclear && ret)
    copymatrix(a,ac,pdim);
  return ret;
}






int trypermute(double a1[][MAX_DIM+1],double q[][MAX_DIM+1],int pdim,double bias)
//  a1 - is a matrix similar to some matrix <a> via an orthogonal change of coordinates
//        a1=q^t a q, a - is not a parameter here
//  a1 - is obtained usually via QR algorithm, when we didn't reach upper
//       triangular form, basically due to existence of invariant subspaces
//       from complex eigenvalue or subspaces with the same eigenvalue
//    this procedure tries to order the basis so that vectors from the same
//    invariant subspace will be consequtive
//   when such permutation is found we update <q>
{
  int ret=1;
  for(int cno=1;cno<pdim;cno++){
    int jmax=0;
    double amax=0.0;
    int j=cno+1; // this means that we only look below the diagonal for
    // 'large' entries
    while(j<=pdim){
      if(fabs(a1[j][cno]) > amax){
        amax=fabs(a1[j][cno]);
        jmax=j;
      }
      j++;
    }
    if((jmax > 0) && (amax > bias)){
      int permut[MAX_DIM+1];    // permutation from new to old
      for(int i=1;i<=pdim;i++){
        if(i<=cno) permut[i]=i;
        if(i==jmax) permut[i]=cno+1;
        if((i>cno) && (i<jmax)) permut[i]=i+1;
        if(i>jmax) permut[i]=i;
      }

      // new basis will be e(permut^{-1}[i])
      // hence the transition matrix from new to old cooridinates
      //  should have '1' for q[i][permut[i]]
      double qper[MAX_DIM+1][MAX_DIM+1],qpert[MAX_DIM+1][MAX_DIM+1];
      zero(qper,pdim);
      zero(qpert,pdim);
      for(int i=1;i<=pdim;i++){
        qper[i][permut[i]]=1.0;
        qpert[permut[i]][i]=1.0;
      }

      ArrArr(qpert,a1,a1,pdim);  // a1=qpert * a1
      ArrArr(a1,qper,a1,pdim);   //  a1=qpert * a1 * qper

      ArrArr(q,qper,q,pdim);    // q=q*qper  ; hence a1= q^t * a * q
    }
  }
  return(ret);
}

int qralg(const double a[][MAX_DIM+1],double q[][MAX_DIM+1],int pdim,
       long maxiterno,double bias,double relbias)
// the procedure performs the QR-algorithm to find an orthogonal matrix
//  q such that
//   q^t a q - is almost upper triangular (with eventually 2x2 blocks
//       on the diagonal)
{


  int stepimpr=10; // every <stepimpr>-iterations of QR-decomposition
                   //                  we orthogonalize the matrix q

  double q1[MAX_DIM+1][MAX_DIM+1],r[MAX_DIM+1][MAX_DIM+1],
            a1[MAX_DIM+1][MAX_DIM+1];
  double preva1[MAX_DIM+1][MAX_DIM+1];
  double qt[MAX_DIM+1][MAX_DIM+1];

  //   q=id
  identity(q,pdim);

  copymatrix(a,a1,pdim);  // a1=a

  int tryno=1;
  int maxtryno=3;
  int ok=1;
  int done=0;
  long itno;

  while(ok && (!done) && (tryno<=maxtryno)){

    itno=1;
    if((tryno <= maxtryno-1) && (tryno >1)){
      trypermute(a1,q,pdim,bias);
    }
    if(tryno==maxtryno){ // this is out last attempt
      // we start from the begining with a slighty randomized matrix
      //  to kill some degeneracies
      identity(q,pdim);

      copymatrix(a,a1,pdim);  // a1=a
      for(int i=1;i<=pdim;i++)
        for(int j=1;j<=pdim;j++){
          a1[i][j]+=zglrandom(-1e-2,1e-2)*bias;
        }
    }
    while((itno <=maxiterno) && ok && (!done) ){

      copymatrix(a1,preva1,pdim); // preva1=a1

      ok=qrdecomp(a1,q1,r,pdim);  // a1=q1 r
      if(ok){
        ArrArr(q,q1,q,pdim);    //  q = q q1
              //  we shall have now q^t a q = a1

        // we compute a1=q^t a q
        transpose(q,qt,pdim);
        ArrArr(qt,a,a1,pdim);
        ArrArr(a1,q,a1,pdim);

        if(itno % stepimpr ==0)
           qrimprove(a,q,a1,pdim);  // orthogonalization of q and
                 // recomputation of a1, to avoid a cumulation of small errors


        done=testuptriangular(0,a1,pdim,bias,relbias);
      }
      if(!done){
        double diff=distNormMax(a1,preva1,pdim);
        if(diff < bias)
          trypermute(a1,q,pdim,bias);
      }


      itno++;
    }
    tryno++;
  }

  return done;
}


int qralg_with_rescaling(const double a[][MAX_DIM+1],double q[][MAX_DIM+1],int pdim,
       long maxiterno,double bias,double relbias,double *scok)
// the procedure performs the QR-algorithm to find an orthogonal matrix
//  q such that
//   q^t a q - is almost upper triangular (with eventually 2x2 blocks
//       on the diagonal)
//  the matrix a is left unchanged
//// it tries qralg for a, 0.1a 0.01a; this somehow amounts to increasing <bias>
  //                  but this is not the same
//  in scok - on output it returns 'x' such that qralg applied to x*a
//             succeds
{
  double scalefactor=0.1;
  double ap[MAX_DIM+1][MAX_DIM+1];

  int sciterno=0;
  int scitermax=2;
  int ok=0;
  *scok=1.0;
  while((sciterno <=scitermax) && !ok){
    copymatrix(a,ap,pdim);
    if(sciterno>0){
      double t=pow(scalefactor,sciterno);
      for(int i=1;i<=pdim;i++)
        for(int j=1;j<=pdim;j++)
          ap[i][j]*=t;
      *scok=t;
    }
    ok= qralg(ap,q,pdim,maxiterno,bias,relbias);
    sciterno++;
  }
  return(ok);
}


void notclear_message(const double notupptr[][MAX_DIM+1],const double a[][MAX_DIM+1],
  int pdim,double biasclear,int maxiterno, double biasqr,double relbiasqr)
// the error-handling procedure for <realqrdiag>
//   <notupptr> - is a matrix obtained from <a>, via QR-diagonalization,
//    but it turns out not to be upper-triangular within the <biasclear>
{
  #if defined(WIN95) && defined(KRAK)
    rootFrame.Clear();
  #endif
  ofstream report(zglerrorlogname,ios::app);
  entriesinerrorlog=1;
  report <<"************************" << endl;
  writeCompInfo(report);
  writedatetime(report);



  std::ostringstream com;

  com << "realqrdiag - unable to clear the matrix   bias=" << biasclear
           << std::ends;

  wrmatrix(notupptr,pdim,report,(com.str()).c_str());



  std::ostringstream com2;

  com2 << "from this matrix <qralg> returned the previous one " <<
    " with bias=" << biasqr << "  relbias=" << relbiasqr  <<
              "  maxiterno=" << maxiterno << std::ends;

  wrmatrix(a,pdim,report,(com2.str()).c_str());

  cerr << "realqrdiag - unable to clear the matrix " << endl;
  cerr << "the matrix causing problems is in the report file ";
  cerr << zglerrorlogname ;
  cerr << endl;

  #if defined(WIN95) && defined(KRAK)
    cerr << "press any key " << endl;
    waitBt();
  #endif
}

void qr_alg_failure_message(const double a[][MAX_DIM+1],int pdim,int maxiterno,
             double bias,double relbias,const char *procname)
// an error message function for <procname>
//   <a> - the matrix to diagonalize
//    pdim - dim(a)
//   maxiterno, bias, relbias - the parameters of qralg
{
  #if defined(WIN95) && defined(KRAK)
    rootFrame.Clear();
  #endif
  ofstream report(zglerrorlogname,ios::app);
  entriesinerrorlog=1;
  report <<"************************" << endl;
  writeCompInfo(report);
  writedatetime(report);



  std::ostringstream com;

  com << procname << " - unable to bring to upper triangular form:  bias="
      << bias  << "  relbias=" << relbias << "  iter=" << maxiterno << std::ends;


  wrmatrix(a,pdim,report,(com.str()).c_str());

  cerr <<  procname <<  " - unable to bring to upper triangular form " << endl;

  cerr << "the matrix causing problems is in the report file "
       << zglerrorlogname  << endl;

  #if defined(WIN95) && defined(KRAK)
     cerr << "press any key " << endl;
     waitBt();
  #endif
}


int realqrdiag(const double a[][MAX_DIM+1],double q[][MAX_DIM+1],
     double p[][MAX_DIM+1],double a1[][MAX_DIM+1],double eigenval[][3],
     int pdim,int maxiterno,double bias,double relbias)
// the real part for  functions <qrdiagonalize> (both the real and the
//                    interval versions)
//        - common initial part for both
//  On output:
//     a1=q^t*a*q - is allmost upper triangular
//        q-orthogonal
//        p - collumns are eigenvectors for <a1>
//        eigenval - eigenvalues for a (and a1)
{


  double scfactor;
  if(!qralg_with_rescaling(a,q,pdim,maxiterno,bias,relbias,&scfactor)){  // q^t a q - is practically
                  //   upper (with eventually some 2x2 blocks)triangular;
                  //           q-is orthogonal

    qr_alg_failure_message(a,pdim,maxiterno,bias,relbias,"realqrdiag");
    return 0;
  }



  double qt[MAX_DIM+1][MAX_DIM+1];
  transpose(q,qt,pdim);
  ArrArr(qt,a,a1,pdim);
  ArrArr(a1,q,a1,pdim);  //  a1=q^t * a * q


  double biasclear=(1 + 1e-4)*bias/scfactor; // it was 1e5*bias

  if(!testuptriangular(1,a1,pdim,biasclear,relbias)){

    notclear_message(a1,a,pdim,biasclear,maxiterno,bias,relbias);
    return 0;
  }


  double leigenval[MAX_DIM+1][MAX_DIM+1];
  int ret=diagupbltrian(a1,pdim,p,leigenval);  // after this routine
                        // the columns of <p>  become eigenvectors

  if(ret){
    // filling-in eigenvalue data
    int i=1;
    while(i<=pdim){
      eigenval[i][1]=leigenval[i][i];
      eigenval[i][2]=0.0;
      if(i<pdim){
        if(fabs(leigenval[i][i+1])>0){
          eigenval[i][2]=leigenval[i][i+1];
          eigenval[i+1][1]=eigenval[i][1];
          eigenval[i+1][2]=-eigenval[i][2];
          i++;
        }
      }
      i++;
    }
 }
 return ret;
}


int qrdecomposition(const double a[][MAX_DIM+1],interval q[][MAX_DIM+1],
     int pdim,int maxiterno,double bias,double relbias)
// the procedure performs QR-algorithm to find an orthogonal matrix
//  q such that
//   q^t a q - is almost upper triangular (with eventually 2x2 blocks
//       on the diagonal)
{

  double qreal[MAX_DIM+1][MAX_DIM+1];

  double scfactor;
  int ret=qralg_with_rescaling(a,qreal,pdim,maxiterno,bias,relbias,&scfactor);

  if(!ret){
    qr_alg_failure_message(a,pdim,maxiterno,bias,relbias,"qrdecomposition");
    return 0;
  }
  if(ret){
    copymatrix(qreal,q,pdim);
    icleanmatrix(q,pdim,pdim,1e-3*bias);
    ret=orthogonalize(q,pdim,pdim);
  }

  return ret;
}




int partialqrdiagonalize(const double a[][MAX_DIM+1],double p[][MAX_DIM+1],
     double pinv[][MAX_DIM+1],double eigenval[][3],
     int pdim,int dimdiag,int maxiterno,double bias,double relbias,
     int normalize /*=1*/)
//   we diagonalize <a> using QR-decomposition
//   on output:
//       pinv = p^{-1}
//     p - is a coordinate change diagonalizing <a> - in the following (partial)
//            sense
//    pinv * a * p - is close to diagonal (complex eigenvalues give rise
//           to 2x2 cell) - in upper dimdiag X dimdiag - corner
//         the remaining part is upper triangular and is obtained from
//                 QR-decomposition
//     eigenvalues are given by eigenval[k][1] + i* eigenval[k][2]
//     we assume that each Jordan cell is simple
//  if <normalize> then we will try to have <p> balanced (if possible
//     the length of each column in <p> will be equal to 1 (for a 2x2 cell the
//     longer vector)), euclidean norm is used for this normalization
{

  if(dimdiag > pdim){
    cerr << "partialqrdiagonalize - problem: dimdiag= "<< dimdiag <<
       " is larger than pdim=" << pdim << endl;
        #if defined(KRAK) && defined(WIN95)
                      waitBt();
    #endif
    exit(EXIT_FAILURE);
  }

  double q[MAX_DIM+1][MAX_DIM+1];
  double a1[MAX_DIM+1][MAX_DIM+1];

  int ret=realqrdiag(a,q,p,a1,eigenval,pdim,maxiterno,bias,relbias);
  //     a1=q^t*a*q - is allmost upper triangular
//        q-orthogonal
//        p - collumns are eigenvectors for <a1>, this is an upper-traingular matrix
//        eigenval - eigenvalues for a (and a1)


  if(ret && (dimdiag < pdim)){
    // we check whether the spliting does not destroy complex eigenvalue
    if(eigenval[dimdiag][2]!=0){
      dimdiag=dimdiag+1;
    }
  }



  if(ret){

    if(dimdiag < pdim){
      for(int j=dimdiag+1;j<=pdim;j++){
        for(int i=1;i<=pdim;i++){
          p[i][j]=0;
          if(i==j) p[i][j]=1.0;
        }
      }
    }


    // computing the coordinate change
    inverse_uptr(p,pinv,pdim);

    cleanmatrix(q,pdim,pdim,1e-3*bias);
    ret=orthogonalize(q,pdim,pdim);

    if(!ret){
      #if defined(WIN95) && defined(KRAK)
        rootFrame.Clear();
      #endif
      cerr << "qrdiagonalize (double) - cannot orthogonalize matrix q" << endl;
      #if defined(WIN95) && defined(KRAK)
        cerr << "press any key " << endl;
        waitBt();
        exit(EXIT_FAILURE);
      #endif
    }

    double qt[MAX_DIM+1][MAX_DIM+1];
    transpose(q,qt,pdim);
    ArrArr(q,p,p,pdim);
    ArrArr(pinv,qt,pinv,pdim);


    if(normalize)
      normalizetransf(p,pinv,eigenval,pdim,NORM_EUCL);

  }

  return ret;

}


int qrdiagonalize(const double a[][MAX_DIM+1],double p[][MAX_DIM+1],
     double pinv[][MAX_DIM+1],double eigenval[][3],
     int pdim,int maxiterno,double bias,double relbias,
     int normalize /*=1*/)
//   we diagonalize <a> using QR-decomposition
//   on output:
//       pinv = p^{-1}
//     p - is a coordinate change diagonalizing <a>.
//    pinv * a * p - is close to diagonal (complex eigenvalues give rise
//           to 2x2 cell)
//     eigenvalues are given by eigenval[k][1] + i* eigenval[k][2]
//     we assume that each Jordan cell is simple
//    columns of <p> are 'eigenvectors' of a
//  if <normalize> then we will try to have <p> balanced (if possible
//     the length of each column in <p> will be equal to 1 (for a 2x2 cell the
//     longer vector)), euclidean norm is used for this normalization
{
  return partialqrdiagonalize(a,p,pinv,eigenval,pdim,pdim,maxiterno,bias,relbias,
     normalize);
}




int ipartialqrdiagonalize(const double a[][MAX_DIM+1],interval p[][MAX_DIM+1],
     interval pinv[][MAX_DIM+1],double eigenval[][3],
     int pdim,int dimdiag,int maxiterno,double bias,double relbias, int normalize /*=1*/)
//   we diagonalize <a> using QR-decomposition
//   on output:
//       pinv = p^{-1}
//     p - is a coordinate change diagonalizing <a> - in the following (partial)
//            sense
//    pinv * a * p - is close to diagonal (complex eigenvalues give rise
//           to 2x2 cell) - in upper dimdiag X dimdiag - corner
//         the remaining part is upper triangular and is obtained from
//                 QR-decomposition
//     eigenvalues are given by eigenval[k][1] + i* eigenval[k][2]
//     we assume that each Jordan cell is simple
//  if <normalize> then we will try to have <p> balanced (if possible
//     the length of each column in <p> will be equal to 1 (for a 2x2 cell the
//     longer vector)), euclidean norm is used for this normalization
{

  if(dimdiag > pdim){
    cerr << "ipartialqrdiagonalize - problem: dimdiag= "<< dimdiag <<
       " is larger than pdim=" << pdim << endl;
        #if defined(KRAK) && defined(WIN95)
                      waitBt();
    #endif
    exit(EXIT_FAILURE);
  }


  double q[MAX_DIM+1][MAX_DIM+1];
  double a1[MAX_DIM+1][MAX_DIM+1];
  double pdouble[MAX_DIM+1][MAX_DIM+1];



  int ret=realqrdiag(a,q,pdouble,a1,eigenval,pdim,maxiterno,bias,relbias);
  //     a1=q^t*a*q - is allmost upper triangular
//        q-orthogonal
//        pdouble - columns are eigenvectors for <a1> - this is an upper
//                   triangular matrix
//        eigenval - eigenvalues for a (and a1)

  if(ret && (dimdiag < pdim)){
    // we check whether the spliting does not destroy complex eigenvalue
    if(eigenval[dimdiag][2]!=0){
      dimdiag=dimdiag+1;
    }
  }


  if(ret){
    // computing the coordinate change
//    interval qint[MAX_DIM+1][MAX_DIM+1],qtint[MAX_DIM+1][MAX_DIM+1];
    intervalmatrix1 *qint;
    qint=new intervalmatrix1[pdim+1];
    intervalmatrix1 *qtint;
    qtint=new intervalmatrix1[pdim+1];

    for(int i=1;i<=pdim;i++)
      for(int j=1;j<=pdim;j++){
        p[i][j]=interval(pdouble[i][j]);
        qint[i][j]=interval(q[i][j]);
      }
    if(dimdiag < pdim){
      for(int j=dimdiag+1;j<=pdim;j++){
        for(int i=1;i<=pdim;i++){
          p[i][j]=0;
          if(i==j) p[i][j]=1.0;
        }
      }
    }

    inverse_uptr(p,pinv,pdim);

    icleanmatrix(qint,pdim,pdim,1e-3*bias);
    ret=orthogonalize(qint,pdim,pdim);
    if(!ret){
      #if defined(WIN95) && defined(KRAK)
        rootFrame.Clear();
      #endif
      cerr << "ipartialqrdiagonalize - cannot orthogonalize matrix q" << endl;
      #if defined(WIN95) && defined(KRAK)
        cerr << "press any key " << endl;
        waitBt();
      #endif
    } else{
      transpose(qint,qtint,pdim);
      ArrArr(qint,p,p,pdim);

      ArrArr(pinv,qtint,pinv,pdim);

      if(normalize){
        normalizetransf(p,pinv,eigenval,pdim,NORM_EUCL);
      }
    }
    delete [] qint;
    qint=NULL;
    delete [] qtint;
    qtint=NULL;
    if(!ret){
      exit(EXIT_FAILURE);
    }
  }
  return ret;
}



int iqrdiagonalize(const double a[][MAX_DIM+1],interval p[][MAX_DIM+1],
     interval pinv[][MAX_DIM+1],double eigenval[][3],
     int pdim,int maxiterno,double bias,double relbias, int normalize /*=1*/)
//   we diagonalize <a> using QR-decomposition
//   on output:
//       pinv = p^{-1}
//     p - is a coordinate change diagonalizing <a>.
//    pinv * a * p - is close to diagonal (complex eigenvalues give rise
//           to 2x2 cell)
//     eigenvalues are given by eigenval[k][1] + i* eigenval[k][2]
//     we assume that each Jordan cell is simple
//  if <normalize> then we will try to have <p> balanced (if possible
//     the length of each column in <p> will be equal to 1 (for a 2x2 cell the
//     longer vector)), euclidean norm is used for this normalization
{
  return ipartialqrdiagonalize(a,p,pinv,eigenval,pdim,pdim,maxiterno,bias,
          relbias,normalize);
}






int calcnounstable(const double df[][MAX_DIM+1],int pdim,int type,int *no_unstabledir,
    double *stabfactor)
//  df - is pdim x pdim matrix - describing linearization for flow or a map
//  type = 1 - flow
//  type = 0 - a map
//  on output *no_ustabledir - is equal to the number of unstable directions
//            *stabfactor - max Real part of eigenvalue when flow
//                        - max absolute value of eigenvalue when flow
// returns 1 - when diagonalization procedure went fine
//         0 - otherwise
{
  int flow=1;
  int map=0;

  int ret;
  double p[MAX_DIM+1][MAX_DIM+1];
  double pinv[MAX_DIM+1][MAX_DIM+1];
  double eigenval[MAX_DIM+1][3];

  int maxiterno=10000;
  double bias=1e-8;
  double relbias=1e-4;

  ret=qrdiagonalize(df,p,pinv,eigenval,pdim,maxiterno,bias,relbias);

  *no_unstabledir =-1;
  if(ret){
    *no_unstabledir=0;
    if(type==flow){
      *stabfactor=eigenval[1][1];
      for(int i=1;i<=pdim;i++){
        if(eigenval[i][1] >=0)    (*no_unstabledir)++;
        if(eigenval[i][1] > *stabfactor) *stabfactor=eigenval[i][1];
      }
    }
    if(type==map){
      double t;
      t=normEucl(eigenval[1],2);
      *stabfactor=t;
      for(int i=1;i<=pdim;i++){
        t=normEucl(eigenval[1],2);
        if(t >=1)
          (*no_unstabledir)++;
        if(t > *stabfactor) *stabfactor=t;
      }
    }
  }

  return(ret);
}


void scalprodeigvect(const double p[][MAX_DIM+1], double g[][MAX_DIM+1],int dim)
// columns of <p> represent eigenvectors
//  on output g[i][j]=scalar product(p[][i],p[][j])
//     this gives some information how singular this change of coordinates
//     is
{
  double ptran[MAX_DIM+1][MAX_DIM+1];
  transpose(p,ptran,dim);
  ArrArr(ptran,p,g,dim);
}

void iscalprodeigvect(const interval p[][MAX_DIM+1], interval g[][MAX_DIM+1],int dim)
// columns of <p> represent eigenvectors
//  on output g[i][j]=scalar product(p[][i],p[][j])
//     this gives some information how singular this change of coordinates
//     is
{
  interval ptran[MAX_DIM+1][MAX_DIM+1];
  transpose(p,ptran,dim);
  ArrArr(ptran,p,g,dim);
}





int qrinverse(const double a[][MAX_DIM+1],double c[][MAX_DIM+1],int dim)
//  the function computes c=a^{-1} using the QR-decomposition
//  1 - is returned if an inversion was possible
//  0 - otherwise
{
  double q[MAX_DIM+1][MAX_DIM+1];
  double r[MAX_DIM+1][MAX_DIM+1];
  int retval=qrdecomp(a,q,r,dim); // a=QR
  if(!retval){
    cerr << "qrinverse - QR-decomposition failed" << endl;
  } else {
    transpose(q,q,dim);
    double rinv[MAX_DIM+1][MAX_DIM+1];
    retval=inverse_uptr(r,rinv,dim);
    if(!retval){
      cerr << "qrinverse - cannot inverse upper traingular part" << endl;
    } else{
      ArrArr(rinv,q,c,dim); // c=rinv*q=R^{-1}*Q^T
    }
  }
  if(!retval){
    #if defined(WIN95) && defined(KRAK)
      cout << "press any key " << endl;
      waitBt();
    #endif
  }
  return retval;
}

int wrdiag(ostream &report,const double df[][MAX_DIM+1],int dim,int dimpartdiag /*=0*/,
    long maxiterno /*=1e4 */,double bias /* = 1e-8*/,double relbias /*=1e-4 */)
// the results of the diagonalization of <df> are sent to the file <report>
//  dim - the dimension of the square matrix <df>
// if dim==2 - the diagonalization is performed using explicit formulas
//  otherwise we use the QR-decomposition algorithm
{
  int ret;
  if(dim==2){
    ret=1;
    wrdiag2x2(report,df);
  }else{
    if(dimpartdiag==0)
      dimpartdiag=dim;

    double p[MAX_DIM+1][MAX_DIM+1];
    double pinv[MAX_DIM+1][MAX_DIM+1];
    double eigenval[MAX_DIM+1][3];

    if(dimpartdiag < dim){
      report << "this is PARTIAL diagonalization with dim_part_diag=" << dimpartdiag << endl;
    }
    report << "diagonalization parameters: ";
    report << "maxiterno= " << maxiterno << "    ";
    report << "bias= " << bias << "    ";
    report << "relbias= " << relbias << endl;

    if(partialqrdiagonalize(df,p,pinv,eigenval,dim,dimpartdiag,maxiterno,bias,relbias)){
      ret=1;

      report << endl << endl;

      double g[MAX_DIM+1][MAX_DIM+1];
      scalprodeigvect(p,g,dim);
      report << endl << "eigenvect_i*eigenvect_j matrix" << endl;
      saveArray(report,g,dim);

      report << endl << endl;

      report <<  "Eigenvalues " << endl ;
      wreigenval(report,eigenval,dim);

      report << endl << endl << "Eigenvectors " << endl;
      for(int i=1;i<=dim;i++){
        for(int j=1;j<=dim;j++){
          report << p[j][i] << endl;
        }
        report << endl << endl;
      }

    } else {
      ret=0;
      report << "problem with diagonalization bias=" << bias <<
 " maxiterno=" << maxiterno  << endl;
    }
  }
  return(ret);
}


int wrdiag(const char fname[],const double df[][MAX_DIM+1],int dim)
// the results of the diagonalisation of <df> are sent to the file <fname>
//  dim - the dimension of the square matrix <df>
// if dim==2 - the diagonalisation is performed using explicit formulas
//  otherwise we use QR-decomposition algorithm
{
  ofstream report(fname,ios::app);
  report << "-----------------------------------" << endl;
  writeCompInfo(report);
  writedatetime(report);

  return  wrdiag(report,df,dim);
}


int iwrdiag(ostream &report,const double df[][MAX_DIM+1],int dim,int dimpartdiag /*=0*/,
    long maxiterno /*=1e4 */,double bias /* = 1e-8*/,double relbias /*=1e-4 */)
// the results of the diagonalization of <df> are sent to the file <report>
//  dim - the dimension of the square matrix <df>
// if dim==2 - the diagonalization is performed using explicit formulas
//  otherwise we use the QR-decomposition algorithm
{
  int ret;
  if(dim==2){
    ret=1;
    wrdiag2x2(report,df);
  }else{
    if(dimpartdiag==0)
      dimpartdiag=dim;

    interval p[MAX_DIM+1][MAX_DIM+1];
    interval pinv[MAX_DIM+1][MAX_DIM+1];
    double eigenval[MAX_DIM+1][3];

    if(dimpartdiag < dim){
      report << "this is PARTIAL diagonalization with dim_part_diag=" << dimpartdiag << endl;
    }
    report << "diagonalization parameters: ";
    report << "maxiterno= " << maxiterno << "    ";
    report << "bias= " << bias << "    ";
    report << "relbias= " << relbias << endl;

    if(ipartialqrdiagonalize(df,p,pinv,eigenval,dim,dimpartdiag,maxiterno,bias,relbias)){
      ret=1;

      report << endl << endl;

      interval g[MAX_DIM+1][MAX_DIM+1];
      iscalprodeigvect(p,g,dim);
      report << endl << "eigenvect_i*eigenvect_j matrix" << endl;
      saveArray(report,g,dim);



      ArrArr(p,pinv,g,dim);
      for(int i=1;i<=dim;i++){
        g[i][i]=g[i][i] - 1.0;
      }
      report <<  endl << endl << "p*p^{-1} - Id" << endl;
      saveArray(report,g,dim);

      report << endl << endl;


      report <<  "Eigenvalues " << endl ;
      wreigenval(report,eigenval,dim);

      report << endl << endl << "Eigenvectors " << endl;
      for(int i=1;i<=dim;i++){
        for(int j=1;j<=dim;j++){
          report << p[j][i] << endl;
        }
        report << endl << endl;
      }


      report << endl << endl << "In new coodinates: A1=P^{-1}*df*P " << endl;
      ArrArr(pinv,df,g,dim);
      ArrArr(g,p,g,dim);
      saveArray(report,g,dim);

      report << endl << endl;

    } else {
      ret=0;
      report << "problem with diagonalization bias=" << bias <<
 " maxiterno=" << maxiterno  << endl;
    }
  }
  return(ret);
}




void QR_rearrange(int dim,const interval M[][MAX_DIM+1],const interval r[],const interval z[],
         interval Bn[][MAX_DIM+1],interval rn[],bool look_at_shape_r)
// our task is to find Bn*rn, such that:  M*x.r + z \subset Bn*rn
// and Bn - is orthogonal
//// it is ok if r,z,rn are pointers to the same array
///     but M and Bn CANNOT be the same pointers
//  if look_at_shape_r==1, then the shape of r  is taken into account

{
  intervalmatrix1 *Mt, *Bnt, *MR;
  Mt=new intervalmatrix1[dim+1];
  MR=new intervalmatrix1[dim+1];
  Bnt=new intervalmatrix1[dim+1];
  interval* zrot;
  zrot=new interval[dim+1];

  interval* rloc;
  rloc=new interval[dim+1];


  // the set to Qr-rearange will have the form MR*rloc + z
  // if look_at_shape_r  then we set MR=M*sizematrix(r) and rloc = (-1,1)^dim
  //                              or 0 on some directions, if r[i]=0

  copyvector(r,rloc,dim);
  copymatrix(M,MR,dim);

  if(look_at_shape_r) {

    intervalmatrix1 *sr;
    sr=new intervalmatrix1[dim+1];
    zero(sr,dim);

    interval ui(-1,1);
    interval t;
    for(int i=1;i<=dim;i++)
    {
      t=abs(rloc[i]).right_b();
      sr[i][i]=t;
      if(t >0) rloc[i]=ui;
      else rloc[i]=0;
    }
    ArrArr(M,sr,MR,dim);
    delete [] sr;
  }

  for(int i=1;i<=dim;i++)
      for(int j=1;j<=dim; j++)
        Mt[i][j]=mid(MR[i][j]);




  if(!iQRDecomp(Mt,Bn,dim)){
    exit(EXIT_FAILURE);
  }


  // rn=(Bn^{-1}(=Bn^T) MR)r + Bn^T z
  transpose(Bn,Bnt,dim);
  ArrArr(Bnt,MR,Mt,dim);   // Mt=Bn^T * MR

  multAV(Bnt,z,zrot,dim);

  multAV(Mt,rloc,rn,dim);   // now rn = (Bn^{-1}MR)r

  for(int i=1;i<=dim; i++)
    rn[i]+=zrot[i];

  delete [] rloc;
  delete [] zrot;
  delete [] Bnt;
  delete [] Mt;
  delete [] MR;
}



int test_orthogonal(const interval b[][MAX_DIM+1],int dim)
// returns 1 - if B^T*B contains identity
//         0 - otherwise
//// 1 on return, this is does not implicate that B contains the orthogonal matrix
{
  intervalmatrix1 *bt;
  bt=new intervalmatrix1[dim+1];
  transpose(b,bt,dim);  // now bt=b^t

  ArrArr(bt,b,bt,dim);  // now bt=b^t * b

  int ret=1;
  for(int i=1;i<=dim;i++){
    for(int j=1;j<=dim;j++){
      if(i==j){
        ret=ret && bt[i][j].contains(1.0);
      }else{
        ret=ret && bt[i][j].contains(0.0);
      }
    }
  }
  delete [] bt;

  return ret;
}


bool clusterOrthogonalize(int dim,const double p[][MAX_DIM+1],double pout[][MAX_DIM+1],
                   int cluster[])
 // we orthogonalize clusters of columns in p,
 // the cluster to which the i-th column  belongs is defined by cluster[i]
 // which means that in the j-th cluster are all collumns, such that
 //  cluster[i]=j
 // On output: pout - contains the 'clusterorthgonalized' basis   as columns
{

   doublematrix1* mat=new doublematrix1[dim+1];
   transpose(p,mat,dim); // mat=p^t

   doublematrix1* tab=new  doublematrix1[dim+1];

   // now we sort mat[i], according to the value of cluster[i]
   double tempwek[MAX_DIM+1];
   for(int i=1;i<=dim;i++){
     for(int j=1;j<=dim-1;j++){
       if(cluster[j] > cluster[j+1]){
         int t=cluster[j];
         copyvector(mat[j],tempwek,dim);

         cluster[j]=cluster[j+1];
         copyvector(mat[j+1],mat[j],dim);   // mat[j]=mat[j+1]
         cluster[j+1]=t;
         copyvector(tempwek,mat[j+1],dim);    // mat[j+1]=tempwek
       }
     }
   }
   // now mat[] is sorted according to cluster[]



   bool ok=true;

   int i=1;
   int curr_cluster;
   while((i<=dim) && ok){
     curr_cluster=cluster[i];
     int vectnum=1;
     bool cluster_finished=false;
     while(!cluster_finished){

       copyvector(mat[i+vectnum-1],tab[vectnum],dim);
       vectnum++;


       if(i+vectnum-1 > dim){
         cluster_finished=true;
       }else{
         cluster_finished=(curr_cluster!=cluster[i+vectnum-1]);
       }
       if(cluster_finished)
         vectnum--;
     }
     if(vectnum > 1){
       ok=orthogonalize(tab,dim,vectnum) ;
       if(!ok){
         cerr << "orthogonalization failure " << endl;
         #if defined(WIN95) && defined(KRAK)
         cerr << "press any key " << endl;
         waitBt();
         #endif
         exit(EXIT_FAILURE);
       }
     }
     if(ok){
       for(int j=1;j<=vectnum;j++){
         copyvector(tab[j],mat[i+j-1],dim);
       }
     }
     i=i+vectnum;
   }

   if(ok)
     transpose(mat,pout,dim);



   delete [] tab;
   delete [] mat;
   return ok;
}


bool clusterOrthogonalize(int dim,const double p[][MAX_DIM+1],interval pout[][MAX_DIM+1],
                   int cluster[])
 // we orthogonalize clusters of columns in p,
 // the cluster to which the i-th column  belongs is defined by cluster[i]
 // which means that in the j-th cluster are all collumns, such that
 //  cluster[i]=j
 // On output: pout - contains the 'clusterorthgonalized' basis   as columns
{

   intervalmatrix1* mat=new intervalmatrix1[dim+1];
   transpose(p,mat,dim); // mat=p^t

   intervalmatrix1* tab=new  intervalmatrix1[dim+1];

   // now we sort mat[i], according to the value of cluster[i]
   interval tempwek[MAX_DIM+1];
   for(int i=1;i<=dim;i++){
     for(int j=1;j<=dim-1;j++){
       if(cluster[j] > cluster[j+1]){
         int t=cluster[j];
         copyvector(mat[j],tempwek,dim);

         cluster[j]=cluster[j+1];
         copyvector(mat[j+1],mat[j],dim);   // mat[j]=mat[j+1]
         cluster[j+1]=t;
         copyvector(tempwek,mat[j+1],dim);    // mat[j+1]=tempwek
       }
     }
   }
   // now mat[] is sorted according to cluster[]



   bool ok=true;

   int i=1;
   int curr_cluster;
   while((i<=dim) && ok){
     curr_cluster=cluster[i];
     int vectnum=1;
     bool cluster_finished=false;
     while(!cluster_finished){

       copyvector(mat[i+vectnum-1],tab[vectnum],dim);
       vectnum++;


       if(i+vectnum-1 > dim){
         cluster_finished=true;
       }else{
         cluster_finished=(curr_cluster!=cluster[i+vectnum-1]);
       }
       if(cluster_finished)
         vectnum--;
     }
     if(vectnum > 1){
       ok=orthogonalize(tab,dim,vectnum) ;
       if(!ok){
         cerr << "orthogonalization failure " << endl;
         #if defined(WIN95) && defined(KRAK)
         cerr << "press any key " << endl;
         waitBt();
         #endif
         exit(EXIT_FAILURE);
       }
     }
     if(ok){
       for(int j=1;j<=vectnum;j++){
         copyvector(tab[j],mat[i+j-1],dim);
       }
     }
     i=i+vectnum;
   }

   if(ok)
     transpose(mat,pout,dim);



   delete [] tab;
   delete [] mat;
   return ok;
}




void EigenVectClusters(int dim,const double p[][MAX_DIM+1],
        const double eigenval[][3],int cluster[])
 // we gather columns in p  according into clusters
 // the cluster is defined as a transitive closure of the relation
 //    abs(p[i]|p[j])/|p[i]|*|p[j]| > 0.9 (or some other constant)
 //  <eigenval> - serves only to decide if we have a pair corresponding to complex
 //             eigenvalue or single eigenvector
 //            this is currently not used, but later it may turn out that it
 //             makes sense to take it into account
//  <pdim> - the dimension of matrices
 // On output: cluster[i] given the cluster identifier for the i-th column in p
{
    const double cluster_bias=0.9;

    for(int i=1;i<=dim;i++)
      cluster[i]=0;

    doublematrix1* sc_prod=new doublematrix1[dim+1];
    double t;
    for(int i=1;i<=dim;i++){
      for(int j=i;j<=dim;j++){
        t=0;
        for(int k=1;k<=dim;k++){
          t+=p[k][i]*p[k][j];
        }
        sc_prod[i][j]=t;
        sc_prod[j][i]=t;
      }
    }
    for(int i=1;i<=dim;i++){
      double normpi=sqrt(sc_prod[i][i]);
      for(int j=1; j<=dim;j++){
        if(j!=i){
          sc_prod[i][j]/=normpi;
          sc_prod[j][i]/=normpi;
        }
      }
    }


    // now try to make clusters
    int clusternum=0;
    for(int i=1;i<=dim;i++){
      if(cluster[i]==0){
        clusternum++;
        cluster[i]=clusternum;
      }
      for(int j=i+1;j<=dim;j++){
        if(abs(sc_prod[i][j]) > cluster_bias){
          cluster[j]=cluster[i];
        }
      }
    }
    delete [] sc_prod;

}


bool EigenVectOrthogonalizeInClusters(int dim,const double p[][MAX_DIM+1],
        const double eigenval[][3],  double pout[][MAX_DIM+1] )
 // we orthogonalize columns in p  according to cluster-decomposition
 // the cluster is defined as a transitive closure of the relation
 //    abs(p[i]|p[j])/|p[i]|*|p[j]| > 0.9 (or some other constant)
 //  <eigenval> - serves only to decide if we have a pair corresponding to complex
 //             eigenvalue or single eigenvector
 //            this is currently not used, but later it may turn out that it
 //             makes sense to take it into account
//  <pdim> - the dimension of matrices
 // On output: pout - contains the 'clusterorthgonalized' basis   as columns
{
    int cluster[MAX_DIM+1];
    EigenVectClusters(dim,p,eigenval,cluster);
    return clusterOrthogonalize(dim,p,pout,cluster);
}


bool EigenVectOrthogonalizeInClusters(int dim,const double p[][MAX_DIM+1],
        const double eigenval[][3],  interval pout[][MAX_DIM+1] )
 // we orthogonalize columns in p  according to cluster-decomposition
 // the cluster is defined as a transitive closure of the relation
 //    abs(p[i]|p[j])/|p[i]|*|p[j]| > 0.9 (or some other constant)
 //  <eigenval> - serves only to decide if we have a pair corresponding to complex
 //             eigenvalue or single eigenvector
 //            this is currently not used, but later it may turn out that it
 //             makes sense to take it into account
//  <pdim> - the dimension of matrices
 // On output: pout - contains the 'clusterorthgonalized' basis   as columns
{
    int cluster[MAX_DIM+1];
    EigenVectClusters(dim,p,eigenval,cluster);
    #if defined(TESTY)
    cout << "cluster: " << endl;
    for(int i=1;i<=dim;i++)
      cout << cluster[i]<< "  " ;
    cout << endl;
    #endif
    return clusterOrthogonalize(dim,p,pout,cluster);
}


