#include<iostream>
#include<fstream>
using namespace std;

#include<cstdlib>

#include"defo.h"
#include"zdynsys.h"
#include"vfkso.h"
#include"multiint.h"
#include"zgfiles.h"

#if defined(DIFF_INCL)
  #include"galks.h"
  #include"rigkserr.h"
#endif






// #define KS_ENCL_ITER_TEST  // this variable enable us to force
// the same number of iterates in the enclosure with tail



// a few interesting values for mu
//  mu=0.02991 from Christansen and all - chaotic
//  mu=0.029924  from Christansen and all - period 3 window
//  mu=0.03 - an attracting period 2 orbit (non-symmetric)
//  mu=0.032 - there exists a symmetric periodic trajectory - bifurcating
//                later by period doubling


// constants defining the type of the rough enclosure function
const int ks_ode_enclosure=0;   // standard ODE
const int ks_ode_iso_enclosure=1; //  standard ODE + (if previous succeds)
                              //  the improvement via isolation computation

const int ks_mixed_enclosure=2; //  the mixture of the standard encl. and
                               // isolation from the begining
                              //  this appears to be slighty worse than ode_iso_encl
                              //  but for 'large' time steps it works were the other  fails

// #define EXP_FROM_TABLE      // it does not help - for m\approx 12 and M\approx 36

#if defined(EXP_FROM_TABLE)
KS_Exp ks_exp(MAX_DIM+1);  // the class used to compute exp(h*lambda(k))
                    // it happens that we quite often call exp() with
                    // the same parameters, hence we put in into a table
#endif

char* ksencl_name(int enclfunctype,char buffer[],int bufflen)
{
  strncpy(buffer,"unknown enclosure function",bufflen-1);
  if(enclfunctype==ks_ode_enclosure)
    strncpy(buffer, "ODE standard enclosure ",bufflen-1 );
  if(enclfunctype==ks_ode_iso_enclosure)
    strncpy(buffer, "ODE + improvement via isolation ",bufflen-1 );

  if(enclfunctype==ks_mixed_enclosure)
    strncpy(buffer, "mixed enclosure ",bufflen-1 );

  return buffer;
}


// functions converting mu to alpha and vice versa
double mutoalpha(double mu)
{
  return(4/mu);
}


double alphatomu(double alpha)
{
  return(4/alpha);
}


interval mutoalpha(interval mu)
{
  return(4/mu);
}


interval alphatomu(interval alpha)
{
  return(4/alpha);
}


void read_dimension(int *dimks,int useinitialval,int maxdim)
{
  if(useinitialval)
    cout << "Press 0 - for dim=" << *dimks << "  or enter the dimension  (<="
              << maxdim << ")";
  else
    cout << "enter the dimension  (<=" << maxdim << ")";

  int tempd;
  cin >> tempd;
  #if defined(WIN95) && defined(KRAK)
  cout << endl;
  #endif
  if(tempd > 0) *dimks=tempd;
  if(tempd > maxdim){

    #if defined(WIN95) && defined(KRAK)
      rootFrame.Clear();
    #endif
    cerr << "Dimension is too large " << endl;
    #if defined(WIN95) && defined(KRAK)
      waitBt();
    #endif
    exit(EXIT_FAILURE);
  }
}

// reading parameters of KS:
void ks_input_par(double *mu,int *dimks,bool useinitialval)
  // reads from the console mu and dim for KuraShi
  // if useinitialval then the current parameter value are proposed
{
  int c1=1;
  if(useinitialval){
    cout << "press 0 - mu=" << *mu << "  1-enter mu " ;
    cin >> c1;
    #if defined(WIN95) && defined(KRAK)
    cout << endl;
    #endif
  }
  if(c1!=0){
    cout << "mu=" ;
    cin >> *mu;
    #if defined(WIN95) && defined(KRAK)
    cout << endl;
    #endif
  }
  read_dimension(dimks,((c1==0) && (useinitialval)),MAX_DIM);
}


void ks_input_par(interval *mu,int *dimks)
  // reads from the console mu and dim for KuraShi
{
  int enterAsRatio=1;
  cout << "Press 1 - to enter mu as ratios   , 0 - otherwise (as doubles) "  ;
  cin >>  enterAsRatio;
  #if defined(WIN95) && defined(KRAK)
  cout << endl;
  #endif


  if(enterAsRatio){
    interval lmu=enterIntervalAsRatio("mu(center)");
    if(lmu <=0){
       cerr << "mu must be positive " << endl;
       #if defined(KRAK) && defined(WIN95)
                   waitBt();
      #endif
      exit(EXIT_FAILURE);
    }
    interval hdiam= enterIntervalAsRatio("half-diameter(mu)");
    if(hdiam.right_b() <0){
       cerr << "diameter must be non-negative " << endl;
       #if defined(KRAK) && defined(WIN95)
                   waitBt();
      #endif
      exit(EXIT_FAILURE);
    }
    double dmu=hdiam.right_b();
    *mu=lmu+interval(-dmu,dmu);
  } else{
    double lmu,dmu;
    cout << "enter  mu (center)=";
    cin >> lmu ;
    #if defined(WIN95) && defined(KRAK)
    cout << endl;
    #endif
    cout << "enter the diameter (=> 0) diam=";
    cin >> dmu;
    #if defined(WIN95) && defined(KRAK)
    cout << endl;
    #endif
    if(lmu <0){
       cerr << "mu must be positive " << endl;
       #if defined(KRAK) && defined(WIN95)
                   waitBt();
      #endif
      exit(EXIT_FAILURE);
    }
    *mu=lmu + interval(-dmu,dmu);
  }

  if(dimks!=NULL)
    read_dimension(dimks,0,MAX_DIM);
}



//////////////////////////////////
KuraShi::KuraShi(int pdim,interval pmu,int prough_encl /*=ks_mixed_enclosure*/)
{
  setdefaultvalues();
  strcpy(infostr,"KURAMOTO");
  dim=pdim;
  dissipative=1;
  rough_encl_func_type=prough_encl;
//  c1rough_encl_func_type=ks_mixed_enclosure;
  c1rough_encl_func_type=ks_ode_iso_enclosure; // the mixed function was'nt any
          // better when compared to this one (applying linear inequalities a
          // after the standard C1-enclosure )



  in_enclmaxiter=0;
  #if defined(KS_ENCL_ITER_TEST)
  in_enclmaxiter=1; // the same number of iterates in the enclosure functions
     //  in_enclmaxiter=0; // the number of iterates in the enclosure functions varies
     // and the iteration is interupted if the validation is achieved
  #endif


// mu=0.029910; // chaotic
// mu=0.029924 // period 3-window

  imu=pmu;
  mu=mid(imu).left_b();

  lognormused=NORM_MAX; // used in C^1 algorithm for enclosure of variational
                        // part and in forward integration  based
                        // on logarithmic norms, euclidean (nonrigorous) norm
                        //  performs by far better


}

interval KuraShi::lambda(int k)    // returns k^2(1-\mu k^2)
{
  interval k2=interval(k)*k;
  return k2*(1 - imu*k2);
}
double KuraShi::dlambda(int k)     // returns k^2(1-\mu k^2)
{
  double k2=double(k)*k;
  return k2*(1 - mu*k2);
}



double KuraShi::timestep(double h)
{

   h=0.5/(dim*dim*(1 - mu*dim*dim));
   h=fabs(h);
   return(h);
}


interval KuraShi::timestep(interval h)
{

   h=0.5/(dim*dim*(1 - mu*dim*dim));
   h=abs(h).right_b();
   return(h);
}


void KuraShi::load_lambda_diss_arr(bool coord_diss[],interval arrlambda[])
   // on output   arrlambda[i]=lambda(i) and (coord_diss[i]=(lambda(i) < 0))
{
  for(int i=1;i<=dim;i++){
    arrlambda[i]=this->lambda(i);
    coord_diss[i]=(arrlambda[i] < 0);
  }
}



interval KuraShi::vfield(int k,const interval x[])
    /* int k; */ /* index of coordinate */
   /*  interval x[]; */ /* argument of the vector field, coordinates start
     form 1 */
{
  interval retval=0.0; /* returned value */
  interval sum=0.0; /* \Sum_{m=-\infty}^{m=\infty} a_m a_{m-k} */
  int m,j1,j2,sign;
  for(m=-dim;m <= dim;m++){
    sign=1;
    j1=m;
    if (j1 < 0) {j1=-j1;sign=-sign;}

    j2=k-m;
    if (j2 < 0) {j2=-j2; sign=-sign;}

    if ((j1 > 0) && (j1 <= dim) && (j2 > 0) && (j2 <= dim)){
      interval xj1,xj2;
      xj1=x[j1];

      xj2=x[j2];

      if(j1!=j2)  sum+= sign*xj1*xj2;
      else sum+= sign*power(xj1,2);
    }
  }
  int k2=k*k;
  retval=k2*(1 - imu*k2)*x[k] - k*sum;
  return(retval);
}


interval KuraShi::vfieldndiagt(int k,const interval x[])
// returns f_k - k^2(1-mu)a_k
    /* int k; */ /* index of coordinate */
   /*  interval x[]; */ /* argument of the vector field, coordinates start
     form 1 */
{
  interval retval=0.0; /* returned value */
  interval sum=0.0; /* \Sum_{m=-\infty}^{m=\infty} a_m a_{m-k} */
  int m,j1,j2,sign;
  for(m=-dim;m <= dim;m++){
    sign=1;
    j1=m;
    if (j1 < 0) {j1=-j1;sign=-sign;}

    j2=k-m;
    if (j2 < 0) {j2=-j2; sign=-sign;}

    if ((j1 > 0) && (j1 <= dim) && (j2 > 0) && (j2 <= dim)){
      interval xj1,xj2;
      xj1=x[j1];

      xj2=x[j2];

      if(j1!=j2)  sum+= sign*xj1*xj2;
      else sum+= sign*power(xj1,2);
    }
  }
  retval=- k*sum;
  return(retval);
}

interval KuraShi::dnonlinearpart(int k, int i,const interval x[])
// calculates DN_k/Dx_i (x)
//  where f_k= \lambda_k x_k + N_k(x)
{
  interval retval=0.0; /* returned value */
  interval sum=0.0; /* D/Dx_i \Sum_{m=-\infty}^{m=\infty} a_m a_{m-k} */
  int m,j1,j2,sign;
  for(m=-dim;m<=dim;m++){
    sign=1;
    j1=m;
    if (j1 < 0) {j1=-j1;sign=-sign;}

    j2=k-m;
    if (j2 < 0) {j2=-j2; sign=-sign;}

    if ((j1 > 0) && (j1 <= dim) && (j2 > 0) && (j2 <= dim)){
      interval xj1,xj2;
      xj1=x[j1];

      xj2=x[j2];

      if (j1==i) sum+=  sign*xj2;
      if (j2==i) sum+=  sign*xj1;
    }
  }
  retval= - k*sum;
  return retval;
}


interval KuraShi::dvfield(int k, int i,const interval x[])
     /* calculates Df_k/Dx_i (x) */
{
  interval retval=0.0; /* returned value */
  interval sum=0.0; /* D/Dx_i \Sum_{m=-\infty}^{m=\infty} a_m a_{m-k} */
  int m,j1,j2,sign;
  for(m=-dim;m<=dim;m++){
    sign=1;
    j1=m;
    if (j1 < 0) {j1=-j1;sign=-sign;}

    j2=k-m;
    if (j2 < 0) {j2=-j2; sign=-sign;}

    if ((j1 > 0) && (j1 <= dim) && (j2 > 0) && (j2 <= dim)){
      interval xj1,xj2;
      xj1=x[j1];

      xj2=x[j2];

      if (j1==i) sum+=  sign*xj2;
      if (j2==i) sum+=  sign*xj1;
    }
  }
  retval= - k*sum;

  if (k==i) {
    int k2=k*k;
    retval=k2*(1 - imu*k2) + retval;
  }
  return(retval);
}

interval KuraShi::d2vfield(int k,int i,int j,const interval[] /*x[]*/)
     /* calculates D^2f_k/Dx_iDx_j (x) */
{
  interval retval=0.0;
  if ((-i+j==k) || (i-j==k)) retval=-2.0;
  if (i+j==k) retval=2.0;

  retval*=-k;
  return(retval);
}

int KuraShi::dissOneStep(const interval x0[],interval h,interval xencl[],
          interval xstep[],int use_perturbation,const interval perturb_vect[])
//  if <use_perturbation> then we add <perturb_vect> to the vector field of KS
  // on input: \varphi([0,h],x0) \subset xencl
  // On output in  <xstep> - the bounds for \varphi(h,x)_i
  //           obtained from the dissipativity are returned
{
  copyvector(xencl,xstep,dim);
  if(!(h > 0)) return 1;


  bool coord_dissipative[MAX_DIM+1];
  interval *lambdai;
  lambdai=new interval[dim+1];
  load_lambda_diss_arr(coord_dissipative,lambdai);


  interval iso,offdiag,xh;
  interval exph0lambda;
  for(int i=1;i<=dim;i++){
    if(coord_dissipative[i]){
      offdiag=vfieldndiagt(i,xencl);
      if(use_perturbation)   offdiag+=perturb_vect[i];

      iso=offdiag/(-lambdai[i]);


      #if defined(EXP_FROM_TABLE)
      exph0lambda=ks_exp.e_to_k_l(h,imu,i);
      #else
      exph0lambda=exp(h*lambdai[i]);
      #endif
      if(!(1 > exph0lambda)){
        cerr << "exp problem in [KuraShi::dissOneStep] " << endl;
        #if defined(WIN95) && defined(KRAK)
          waitBt();
        #endif
        exit(EXIT_FAILURE);
      }
      xh=x0[i]*exph0lambda + iso*(1-exph0lambda);

      if(!commonpart(xencl[i],xh,xstep+i)){
        cerr << "very serious error in [KuraShi::dissOneStep] " << endl;
        #if defined(WIN95) && defined(KRAK)
          waitBt();
        #endif
        exit(EXIT_FAILURE);
      }
    }
  }

  delete [] lambdai;
  return 1;
}


int KuraShi::c1dissOneStep(const interval xencl[],interval h,
               interval Vencl[][MAX_DIM+1],interval Vstep[][MAX_DIM+1])
  // On output: in Vstep  are returned  the bounds for
  //         \frac{\partial \varphi_i}{\partial x}_j(h,x0)
  //           obtained from the dissipativity, x0 is the initial condition
  //            for x, and xencl is such that \varphi([0,h],x0) \subset xencl
{

  copymatrix(Vencl,Vstep,dim);
  if(!(h > 0)) return 1;

  bool coord_dissipative[MAX_DIM+1];
  interval *lambdai;
  lambdai=new interval[dim+1];
  load_lambda_diss_arr(coord_dissipative,lambdai);

  intervalmatrix1 *dN;
  dN=new intervalmatrix1[dim+1];
  for(int i=1;i<=dim;i++){
    if(coord_dissipative[i]){
      for(int j=1;j<=dim;j++)
        dN[i][j]=dnonlinearpart(i,j,xencl);
    }
  }



  for(int i=1;i<=dim;i++){
    if(coord_dissipative[i]){
      interval exphlambda;
      #if defined(EXP_FROM_TABLE)
      exphlambda=ks_exp.e_to_k_l(h,imu,i);
      #else
      exphlambda=exp(h*lambdai[i]);
      #endif

      if(!(1 > exphlambda)){
        cerr << "exp problem in [KuraShi::c1dissOneStep] " << endl;
        #if defined(WIN95) && defined(KRAK)
          waitBt();
        #endif
        exit(EXIT_FAILURE);
      }


      interval offdiag,iso,xh,x0;
      for(int j=1;j<=dim;j++){
        offdiag=0;
        for(int k=1;k<=dim;k++)
          offdiag+=dN[i][k]*Vencl[k][j];

        iso=offdiag/(-lambdai[i]);
        if(i==j) x0=1.0;
        else x0=0.0;
        xh=x0*exphlambda + iso*(1-exphlambda);

        interval cp;

        if(!commonpart(Vencl[i][j],xh,&cp)){
          cerr << "very serious error in [KuraShi::c1dissOneStep] " << endl;
          #if defined(WIN95) && defined(KRAK)
            waitBt();
          #endif
          exit(EXIT_FAILURE);
        }
        Vstep[i][j]=cp;

      }
    }
  }

  delete [] dN;
  delete [] lambdai;

  return 1;
}

void KuraShi::diagBadEnclosure(const interval d[],interval h0,const interval y[],const interval z[],
   int use_perturbation,const interval perturb_vect[])
// the function writes to the file <zglerrorlogname>
// the data from failed enlosure procedure
// it tries to give a reason for this
{
   cerr << "KuraShi - bad enclosure - this diagnostic function needs to be rewritten" << endl;
   cerr << " to take into account dissipativeness " << endl;
   this->Ode::diagBadEnclosure(d,h0,y,z,use_perturbation,perturb_vect);
}



void KuraShi::compute_gi_wi(const int &i,const interval lambda[],
          const interval &xi,const interval &h,const interval &N,interval &gi,
          interval &wi)
   // g_i^\pm=(X_i^\pm - b_i^\pm)e^{h\lambda_i} + b_i^\pm
     // w_i=diam(gi \cup x_i)
{

  interval exp_lambda_h;
  #if defined(EXP_FROM_TABLE)
  exp_lambda_h=ks_exp.e_to_k_l(h,this->imu,i);
  #else
  exp_lambda_h=exp(h*lambda[i]);
  #endif
  gi=linIneq_estm(xi,exp_lambda_h,N,lambda[i]);

  wi= diameter(Ihull(gi,xi));
}

interval KuraShi::compute_gi(const interval &h,const interval lambda[],int i,
     const interval X[],const interval y[],bool useperturbation,
     const interval  perturb_vector[])
  // computes g_i - a bound based on linear inequalites (both for positive and negative \lambda[i]) ;
     //    \varphi_i(h,X) \subset g_i,  provided that y - is an enclosure for \varphi([0,h],X)
{
  interval Ny=this->vfieldndiagt(i,y);
  if(useperturbation)   Ny+=perturb_vector[i];

  interval exp_lambda_h;
  #if defined(EXP_FROM_TABLE)
  exp_lambda_h=ks_exp.e_to_k_l(h,imu,i);
  #else
  exp_lambda_h=exp(h*lambda[i]);
  #endif

  return linIneq_estm(X[i],exp_lambda_h,Ny,lambda[i]);
}

void KuraShi::refineEncl_viaLinUnEq(int maxiter, const interval &h,const interval lambda[],
      const interval X[],interval y[],bool useperturbation,
        const interval perturb_vector[])
// refines the enclosure using linear inequalities, both for \lambda_i >0 and \lambda_i <0
//  On input: x - initial condition
//            y - enclosure for \varphi([0,h],x)
//  On output: y - a new enclosure
//         returns the value of:  volume(y(output))/volume(y(input))
{
  interval t;
  interval h0h=Ihull(interval(0),h);
  double gain=0;
  double goodgain=0.9;
  double d0,d1;


  const double smalleigenvalue=1e-6;
  int count=1;
  while((count <= maxiter)  && (gain <=goodgain)){
    gain=1.0;
    for(int i=1;i<=this->dim;i++){
      d0=diameter(y[i]).right_b();
      if(Right(abs(lambda[i])) < smalleigenvalue){
        t=this->vfield(i,y);
        if(useperturbation) t+=perturb_vector[i];
        t=X[i]+ h0h*t;
      }else{
        t=this->compute_gi(h,lambda,i,X,y,useperturbation,perturb_vector);
        t=Ihull(t,X[i]);
      }
      if(!commonpart(t,y[i],y+i)){
        cerr << "serious error in <refineEncl_viaLinUnEq> " << endl;
        #if defined(WIN95) && defined(KRAK)
        waitBt();
        #endif
        exit(EXIT_FAILURE);
      }
      d1=diameter(y[i]).right_b();
      gain*=(d1/d0);
    }
    count++;
  }
}


#if defined(ENCL_TEST)
int KuraShi::mixodeiso_pert_enclosure(const interval d[],interval h0, interval y[],
      int useperturbation, const interval perturb_vector[],  int nomessages,
      ostream & report)
#else
int KuraShi::mixodeiso_pert_enclosure(const interval d[],interval h0, interval y[],
      int useperturbation,const interval perturb_vector[],  int nomessages)
#endif
  // if true is returned then:
  //    \varphi([0,h0],d) \subset y
  //   \varphi is defined by x' \in field(x) + perturb_vector if useperturb
  //   and    \varphi is defined by x' = field(x)  if !useperturb
//  h0 - cannot  be  negative if <useperturbation>
 // we want the condition:
   //    y \subset interior(Z)
   //  where: y=d + [0,h]*vfield(Z)( + perturb (if necessary)) on
   //          'non-dissipative coordinates' and an 'isolation' for
   //                             the 'dissipative' ones
  // nomessages=0 -  if the function  fails then a message is send
  //             to the screen and files
  // nomessages=1 -  the failure does not result in message to screen
{
  if((Left(h0)<0) && useperturbation){
    cerr << "negative ho - the times step in  'KuraShi::mixodeiso_pert_enclosure' " << endl;
    #if defined(WIN95) && defined(KRAK)
      waitBt();
    #endif
    exit(EXIT_FAILURE);
  }

  bool retval=false;
  if(h0.left_b()<0){
    retval=this->Ode::pert_enclosure(d,h0,y,useperturbation,perturb_vector,
                   nomessages);
    return(retval);
  }

  #if defined(ENCL_TEST)
    report << endl << "begin  of incl_m_var " << endl;

    #if defined(KS_ENCL_ITER_TEST)
      report << "forcing <limit> iterates " << endl;
    #endif

    if(useperturbation){
      report << "perturbation: Galerkin errors " << endl;
      saveAllCoord(report,perturb_vector,dim,2,1,2);
    }else
      report << "no perturbation " << endl;
  #endif



  bool coord_dissipative[MAX_DIM+1];

  interval *lambda;
  lambda=new interval[dim+1];

  this->load_lambda_diss_arr(coord_dissipative,lambda);



  int limit=5;   // tells how many times we iterate the loop of guess and check,
                  // inflate the guess and check again


  if(dim/2 >limit) limit=dim/2;


  const double smalleigenvalue=1e-2; // used to to 'decide' whether the eigenvalue
                              // is to small
  const double cLinDom=0.1;  // used to decide whether linear term dominates
  const double growth_rate=1.1; // c - in my notes
  const double init_growth_rate=1.1; //
  const double diss_growth=0.1; // have to be less than 1; c_d in my notes
 // double diss_growth=0.05;


  interval *z;
  z=new interval[dim+1];
  interval h_0h=interval(0,1)*h0;

  for(int i=1;i<=dim;i++){
    if(!coord_dissipative[i]){
      z[i]=d[i]+h_0h*vfield(i,d);
      if(useperturbation)
        z[i]+=h_0h*perturb_vector[i];
    }
    else{
      z[i]=d[i];
    }
  }


  // we have to be sure that <z> has nonempty interior for nondissipative coordinates
  for(int i=1;i<=dim;i++){
    if(!coord_dissipative[i]){
      if(diameter(z[i])==0.0) z[i]=inf_increase_interval(z[i]);
    }
  }

  interval s; // an auxliary variable used for center + s  - representation
                // of various intervals
  for(int i=1;i<=dim;i++){
    if(!coord_dissipative[i]){
      z[i].split(s);  // after this function z[i](old)=z[i](new) + s, but diam(z[i])=0
      z[i]+=init_growth_rate*s;
    }
  }


  interval bZ,NZ,gi,wi,temp;
  for(int i=1;i<=dim;i++){
    if(coord_dissipative[i]){
      NZ=vfieldndiagt(i,z);    // N(Z) - in my notes
      if(useperturbation)   NZ+=perturb_vector[i];

      bZ=NZ/(-lambda[i]);  //   bZ=b  in my notes
      #if defined(ENCL_TEST)
      report << "i=" << i << "   X_i=" << d[i] << "   b=" << bZ << endl;
      #endif

      this->compute_gi_wi(i,lambda,d[i],h0,NZ,gi,wi);

      if(d[i].right_b() < bZ.right_b()){ // this means that the right end  of d
                       // is below the right end of isolation level and that x[i]
                        // might increase up to  bZ.right_b()


        temp=gi.right_b() + diss_growth*wi;
        #if defined(ENCL_TEST)
          report << "Right end:  gi=" <<  gip.right_b() << "  temp=" << temp.right_b() << endl;
        #endif
        if(temp.right_b() > bZ.right_b()) temp=bZ.right_b();

        z[i]=interval(z[i].left_b(),temp.right_b());
      }
      if(d[i].left_b() > bZ.left_b()){ // this means that the left end  of d is above
                          // the left end of isolation level and x[i] might decrease
                          // down to bZ.left_b()



        temp=gi.left_b() - diss_growth*wi;
        #if defined(ENCL_TEST)
          report << "Left end:  gi=" <<  gim.left_b() << "  temp=" << temp.left_b() << endl;
        #endif
        if(temp.left_b() < bZ.left_b()) temp=bZ.left_b();
        z[i]=interval(temp.left_b(),z[i].right_b());
      }
    }
  }
  // end of building of the initial guess


  #if defined(ENCL_TEST)
  report << "dissipative coordinates: " ;
  for(int i=1;i<=dim;i++){
    if(coord_dissipative[i])
      report << i << "  ";
  }
  report << endl;
  report << "lambda's : " ;
  for(int i=1;i<=dim;i++)
    report << lambda[i].right_b() << "  ";
  report << endl;
  #endif


  // the validation (check) and new guess
  int counter=1;
  while(!retval  && (counter <= limit)){
    for(int i=1; i<=dim;i++){
      if(!coord_dissipative[i]){
        y[i]=d[i]+h_0h*vfield(i,z);
        if(useperturbation)
          y[i]+=h_0h*perturb_vector[i];
      }else{
        y[i]=z[i];
      }
    }


    retval=true;
    for(int i=1;i<=dim; i++){
      if(!coord_dissipative[i]){
        if(!in(y[i],z[i],1)){
          retval=0;
          if(counter < limit){ // we update z[i] for a new attempt
                               //  in case of counter==limit
                               //  we want to leave z unchanged for diagnostic reasons
            z[i]=Ihull(z[i],y[i]);
            z[i].split(s); // z[i](old)=z[i](new) + s, but diam(z[i])=0
            z[i]+=s*growth_rate;
          }
        }
      }else{ // this is dissipative coordinate
        NZ=vfieldndiagt(i,z);
        if(useperturbation)   NZ+=perturb_vector[i];

    bZ=NZ/(-lambda[i]);

    if(!bZ.subset(d[i])){
      this->compute_gi_wi(i,lambda,d[i],h0,NZ,gi,wi);
      if(z[i].right_b() < bZ.right_b()){ // this means that the right end  of z
                       // is below the right end of isolation level
        if(gi.right_b() > bZ.right_b()) gi=interval(gi.left_b(),bZ.right_b()); // this will correct
                     // for  overestimates produced by imperfect arithmetics
                     //   as the result is less than iso.right_b()
          y[i]=interval(y[i].left_b(),gi.right_b());
          if(gi.right_b() > z[i].right_b()){
            retval=false;
            if(counter < limit){
              double rend= gi.right_b() + diss_growth*wi.right_b();
              if(rend > bZ.right_b()) rend=bZ.right_b();
              z[i]=interval(z[i].left_b(),rend );
            }
          }
        }
        if(z[i].left_b() > bZ.left_b()){ // this means that the left end  of z is above
                            // the left end of isolation level
          if(gi.left_b() < bZ.left_b()) gi=bZ.left_b(); // corection of overestimates
            y[i]=interval(gi.left_b(),y[i].right_b());
            if(gi.left_b() < z[i].left_b()){
              retval=false;
              if(counter < limit){
                double lend= gi.left_b() - diss_growth*wi.right_b();
                if(lend < bZ.left_b()) lend=bZ.left_b();
                z[i]=interval(lend,z[i].right_b());
              }
            }
          }
        }
      }
    }

    #if defined(ENCL_TEST)
    report << "iterate=" << counter << " validation = " << retval << endl;
    #endif
    counter++;
  }



  // improvement by iteration of enclosures
  if(retval){
    int limit=dim/2;
    this->refineEncl_viaLinUnEq(limit,h0,lambda,d,y,useperturbation,perturb_vector);
  }


  if((!retval) && (!nomessages)){
    cerr << "bad enclosure " << endl;
    diagBadEnclosure(d,h0,y,z,useperturbation,perturb_vector);
  }



  #if defined(ENCL_TEST)
    report << endl << "end of  of incl_m_var " << endl;
  #endif

  delete [] z;
  delete [] lambda;

  return(retval);
}



int KuraShi::pert_enclosure(const interval d[],interval h0, interval y[],
      int useperturbation, const interval perturb_vector[],  int nomessages /*=0 */)
  // if true is returned then:
  //    \varphi([0,h0],d) \subset y
  //   \varphi is defined by x' \in field(x) + perturb_vector if useperturb
  //   and    \varphi is defined by x' = field(x) if !useperturb
  //  h0 - cannot be  negative if <useperturbation>
  // nomessages=0 -  if the function  fails then a message is send
  //             to a screen and files
  // nomessages=1 -  the failure does not result in message to screen
{
  int retval=0;
  if((rough_encl_func_type==ks_ode_enclosure) ||
                      (rough_encl_func_type==ks_ode_iso_enclosure)){
    retval=this->Ode::pert_enclosure(d,h0,y,useperturbation,perturb_vector,
                 nomessages);

    if(rough_encl_func_type==ks_ode_iso_enclosure){
      if(retval){
        interval* tlambda=new interval[this->dim+1];
        for(int i=1;i<=this->dim;i++)
          tlambda[i]=this->lambda(i);

        this->refineEncl_viaLinUnEq(dim/2,h0,tlambda,d,y,useperturbation,perturb_vector);

        delete [] tlambda;
      }
    }
  }
  if(rough_encl_func_type==ks_mixed_enclosure){
    #if defined(ENCL_TEST)
    char repname[80];
    strcpy(repname,catwork);
    strcat(repname,"ksmixencl");
    strcat(repname,orbext);
    ofstream report(repname,ios::app);
    retval=mixodeiso_pert_enclosure(d,h0,y,useperturbation,perturb_vector,
                 nomessages,report);
    #else
    retval=mixodeiso_pert_enclosure(d,h0,y,useperturbation,perturb_vector,
                 nomessages);
    #endif
  }
 return(retval);
}


int KuraShi::enclosure(const interval d[],interval h0, interval y[],int nomessages/* =0*/)
// if true is returned then:
//    \varphi([0,h0],d) \subset y
//    h0 - can be negative
//
 // nomessages=0 -  if the function  fails then a message is send
  //             to a screen and files
  // nomessages=1 -  the failure does not result in message to screen
{

 int useperturb=0;
 return pert_enclosure(d,h0,y,useperturb,NULL,nomessages);
}

void KuraShi::isoimprvvarenclosure(interval h,const interval xencl[],
          interval Vencl[][MAX_DIM+1])
//  in computation of enclosure for variational part
  //  \frac{\partial \varphi}{\partial x} ([0,h],x)=V([0,h],x)
  // hence the initial condition for variational part is V(0)=Id
  // it takes the already validated enclosure Vencl and tries it improve it
  // by using linear inequalities
{

  const double smalleigenvalue=1e-6;
  const double goodgain=0.9; // it was 0.8

  interval *exphlambda;
  exphlambda=new interval[dim+1];
  interval *lambdai;
  lambdai=new interval[dim+1];


  for(int i=1;i<=dim;i++){
    lambdai[i]=lambda(i);
    exphlambda[i]=Right(exp(h*lambdai[i]));
  }


  intervalmatrix1 *dN;
  dN=new intervalmatrix1[this->dim+1];
  for(int i=1;i<=this->dim;i++)
    for(int j=1;j<=this->dim;j++)
      dN[i][j]=this->dnonlinearpart(i,j,xencl);

  int iterno=0;
  int maxiter=this->dim;
  bool done=false;
  while(!done){
    iterno++;
    bool seriousgain=false;
    for(int i=1;i<=this->dim;i++){
      if( (lambdai[i] > smalleigenvalue) || (lambdai[i]<-smalleigenvalue)){
        for(int j=1;j<=dim;j++){
          interval dNVij=0;
          for(int k=1;k<=dim;k++)
            dNVij+=dN[i][k]*Vencl[k][j];

          interval newbound=dNVij*(exphlambda[i]-1)/lambdai[i];
          if(i!=j){
             newbound=Ihull(interval(0.0),newbound);
          }
          if(i==j){
            newbound+=exphlambda[i];
            newbound=Ihull(interval(1.0),newbound);
          }
          if(!commonpart(Vencl[i][j],newbound,&newbound)){
            #if defined(WIN95) && defined(KRAK)
            rootFrame.Clear();
            #endif
            cerr << "serious error in <KuraShi::isoimprvvarenclosure> - empty  intersection " << endl;
            #if defined(WIN95) && defined(KRAK)
            waitBt();
            #endif
            exit(EXIT_FAILURE);
          }
          if(diameter(newbound) < goodgain*diameter(Vencl[i][j])){
            seriousgain=true;
          }
          Vencl[i][j]=newbound;
        }
      }
    }
    done=!seriousgain;
    done=done || (iterno>=maxiter);
  }

  delete [] dN;
  delete [] lambdai;
  delete [] exphlambda;
}

void KuraShi::diag_bad_varmixed_enclosure(interval h,const interval xencl[],
           bool coord_dissipative[][MAX_DIM+1])
// for now just an indication that something went bad
{
  ofstream report(zglerrorlogname,ios::app);
  entriesinerrorlog=1;
  report << "-------------------------" << endl;
  writeCompInfo(report);
  writedatetime(report);
  this->writeparameters(report);
  report << " [KuraShi::varmixed_enclosure] failed " << endl;
  report << "h=" << h << endl;
  report << "xencl :" << endl;
  saveAllCoord(report,xencl,this->dim,1);
  report << "Lambda's: " << endl;
  for(int i=1;i<=this->dim;i++)
    report << "i=" << i << "   " << this->lambda(i) << endl;

  report << "dissipative coordinates: " << endl;
  for(int i=1;i<=this->dim;i++)
    for(int j=1;j<=this->dim;j++){
      if(coord_dissipative[i][j])
        report << i << "," << j << endl;
    }

  report.flush();

}

bool KuraShi::varmixed_enclosure(interval h,const interval xencl[],
             interval Vencl[][MAX_DIM+1])
    // computation of enclosure for variational part
  //  \frac{\partial \varphi}{\partial x} ([0,h],x)=V([0,h],x)
  //// the function uses dissipativity
{



  if(Left(h)<0){
    cerr << "negative h - the times step in  'KuraShi::varmixed_enclosure' " << endl;
    #if defined(WIN95) && defined(KRAK)
    waitBt();
    #endif
    exit(EXIT_FAILURE);
  }

  bool retval=false;

  bool coord_dissipative[MAX_DIM+1][MAX_DIM+1];
  interval *lambda;
  lambda=new interval[dim+1];

  for(int i=1;i<=dim;i++){
    lambda[i]=this->lambda(i);
    for(int j=1;j<=dim;j++){
      coord_dissipative[i][j]=(lambda[i] < 0);
    }
  }



  // d - is our initial condition
  intervalmatrix1 *d;
  d=new intervalmatrix1[this->dim+1];
  identity(d,this->dim);

  int limit=5;   // tells how many times we iterate the loop of guess and check,
                // inflate the guess and check again


  if(this->dim >limit) limit=this->dim;

  const double smalleigenvalue=1e-2;
  const double growth_rate=1.1;
  const double init_growth_rate=1.1;
  const double diss_growth=0.1; // have to be less than 1
  const double cLinDom=0.1;  // used to decide whether linear term dominates

  intervalmatrix1 *z;
  z=new intervalmatrix1[this->dim+1];
  interval h_0h=interval(0,1)*h;

  intervalmatrix1 *Df;
  Df=new intervalmatrix1[this->dim+1];
  for(int i=1;i<=this->dim;i++){
    for(int j=1;j<=this->dim;j++){
      Df[i][j]=this->dvfield(i,j,xencl);
    }
  }




  // initialization of Z
  //  we first try Z=Id + [0,h]*Df(0)*Id on nondissipative directions
  //   and isolation for the dissipative ones
  for(int i=1;i<=this->dim;i++){
    for(int j=1;j<=this->dim;j++){
      if(!coord_dissipative[i][j]){
        interval t=0;
        for(int k=1;k<=dim;k++){
          t+=Df[i][k]*d[k][j];
        }
        z[i][j]=d[i][j]+h_0h*t;
      }
      else{
        z[i][j]=d[i][j];
      }
    }
  }

  intervalmatrix1 *dN;
  dN=new intervalmatrix1[this->dim+1];

  for(int i=1;i<=this->dim;i++)
    for(int j=1;j<=this->dim;j++)
      dN[i][j]=this->dnonlinearpart(i,j,xencl);



  interval NZ,dummy;


   // we have to be sure that z has nonempty interior for nondissipative coordinates
  for(int i=1;i<=this->dim;i++){
    for(int j=1;j<=this->dim;j++){
      if(!coord_dissipative[i][j]){
        if(diameter(z[i][j])==0.0) z[i][j]=inf_increase_interval(z[i][j]);
      }
    }
  }

  interval s; // an auxliary variable used for center + s  - representation
              // of various intervals

  for(int i=1;i<=this->dim;i++){
    for(int j=1;j<=this->dim;j++){
      if(!coord_dissipative[i][j]){
        z[i][j].split(s);  // after this function z[i][j](old)=z[i][j](new) + s, but diam(z[i][j])=0
        z[i][j]+=init_growth_rate*s;
      }
    }
  }



  interval bZ,wi,gi,temp;
  for(int i=1;i<=this->dim;i++){
    for(int j=1;j<=this->dim;j++){
      if(coord_dissipative[i][j]){

        NZ=0;
        for(int k=1;k<=this->dim;k++)
          NZ+=dN[i][k]*z[k][j];

        bZ=NZ/(-lambda[i]);
        this->compute_gi_wi(i,lambda,d[i][j],h,NZ,gi,wi);


        if(d[i][j].right_b() < bZ.right_b()){ // this means that the right end  of d
                     // is below the right end of isolation level and that d[i][j]
                      // might increase up to  bZ.right_b()


          temp=gi.right_b() + diss_growth*Right(wi);
          if(temp.right_b() > bZ.right_b()) temp=bZ.right_b();


          z[i][j]=interval(z[i][j].left_b(),temp.right_b());
        }
        if(d[i][j].left_b() > bZ.left_b()){ // this means that the left end  of d is above
                        // the left end of isolation level and x[i] might decrease
                        // down to iso.left_b()

           temp=gi.left_b() - diss_growth*Right(wi);
           if(temp.left_b() < bZ.left_b()) temp=bZ.left_b();


           z[i][j]=interval(temp.left_b(),z[i][j].right_b());
        }

      }
    }
  }
  // end of building of the initial guess




  // the validation (check) and new guess
  int counter=1;
  while(!retval  && (counter <= limit)){
    for(int i=1; i<=this->dim;i++){
      for(int j=1;j<=this->dim;j++){
        if(!coord_dissipative[i][j]){
          interval t=0;
          for(int k=1;k<=dim;k++){
            t+=Df[i][k]*z[k][j];
          }
          Vencl[i][j]=d[i][j]+h_0h*t;
        }else{
          Vencl[i][j]=z[i][j];
        }
      }
    }
    retval=true;
    for(int i=1;i<=this->dim; i++){
      for(int j=1;j<=this->dim;j++){
        if(!coord_dissipative[i][j]){
          if(!in(Vencl[i][j],z[i][j],1)){
            retval=false;
            if(counter < limit){ // we update z[i][j] for a new attempt
                             //  in case of counter==limit
                             //  we want to leave z unchanged for diagnostic reasons
              z[i][j]=Ihull(z[i][j],Vencl[i][j]);
              z[i][j].split(s); // z[i][j](old)=z[i][j](new) + s, but diam(z[i][j])=0
              z[i][j]+=s*growth_rate;
            }
          }
        } else{ // this is dissipative coordinate
          NZ=0;
          for(int k=1;k<=dim;k++)
            NZ+=dN[i][k]*z[k][j];

          bZ=NZ/(-lambda[i]);

          if(!bZ.subset(d[i][j])){ // in fact since now d=Id - then this condition is
                            // always true, but later I might consider different <d>
            this->compute_gi_wi(i,lambda,d[i][j],h,NZ,gi,wi);


            if(z[i][j].right_b() < bZ.right_b()){ // this means that the right end  of z
                     // is below the right end of isolation level

              if(gi.right_b() > bZ.right_b()) gi=bZ.right_b(); // this will correct
                   // for  overestimates produced by imperfect arithmetics
                   //   because the result is less than bZ.right_b()
              Vencl[i][j]=interval(Vencl[i][j].left_b(),gi.right_b());
              if(gi.right_b() > z[i][j].right_b()){
                retval=false;
                if(counter < limit){
                  double rend= gi.right_b() + diss_growth*wi.right_b();
                  if(rend > bZ.right_b()) rend=bZ.right_b();
                  z[i][j]=interval(z[i][j].left_b(),rend );
                }
              }
            }
            if(z[i][j].left_b() > bZ.left_b()){ // this means that the left end  of z is above
                          // the left end of isolation level

              if(gi.left_b() < bZ.left_b()) gi=bZ.left_b(); // a corection - see above
              Vencl[i][j]=interval(gi.left_b(),Vencl[i][j].right_b());
              if(gi.left_b() < z[i][j].left_b()){
                retval=false;
                if(counter < limit){
                  double lend= gi.left_b() - diss_growth*wi.right_b();
                  if(lend < bZ.left_b()) lend=bZ.left_b();
                  z[i][j]=interval(lend,z[i][j].right_b());
                }
              }
            }
          }
        }
      }
    }
    counter++;
  }
  //  end of checking and setting a new guess

 // refinement - isoimprvarenclosure

  delete [] dN;
  delete [] Df;
  delete [] z;
  delete [] d;
  delete [] lambda;
  return retval;
}

void KuraShi::c1enclosure(interval h,const interval xencl[], interval Vencl[][MAX_DIM+1])
    // computation of enclosure for variational part
  //  \frac{\partial \varphi}{\partial x} ([0,h],x)=V([0,h],x)
  //  i.e starting from identity
  //  this estimate is given  by |Vestm| \leq exp(t*lognorm(df))
  // then we intersect this estimate with Id + [0,h]df(xencl)*Vestm
  // next if h>0 we try to use the isolation to get better enclosure
{
  bool donemixed=false;
  if(c1rough_encl_func_type==ks_mixed_enclosure){
    donemixed=varmixed_enclosure(h,xencl,Vencl);
  }
  if(!donemixed){
    this->Ode::c1enclosure(h,xencl,Vencl);
  }
  if(c1rough_encl_func_type != ks_ode_enclosure){
    if(h>0){
      isoimprvvarenclosure(h,xencl,Vencl);
    }
  }
}


bool checkdimpar(int dim, int T1_m, int T2_m, int T1_big_m, int T2_big_m)
{
  bool ret=true;
  ret=ret && (dim==T1_m);
  ret=ret && (dim==T2_m);
  ret=ret && (T1_big_m==T2_big_m);
  return ret;
}


#if defined(DIFF_INCL)
void KuraShi::compute_bk_gk(interval h,const interval z[],const KuraShi_Tail &T0,
          const KuraShi_Tail &T, KuraShi_Tail &b,interval gk[])
// computes the isolation segments over z \oplus T and stores them in  b
// and gk[i] - for   T0.m <=i<= T0.big_m  - used later in the computation
//         of enclosures for T[0,h] and Th
{
  interval *Nk,*bk;
  Nk=new interval[T0.big_m+1];
  bk=new interval[T0.big_m+1];
  interval cb,cN;
  int sN,sb;
  int p=4; // this is the power in lambda_k

  rigkscompbilinearterm(Nk,T.a,T.m,T.big_m,T.c,T.s,cN,z);
  sN=T.s-2;
  // now |N_k| \leq cN/k^{sN} for k > M (big_m)


  // for k>M we have b_k^\pm = \pm cb/k^{sb}
  interval Mp1_2=interval(T.big_m+1)*(T.big_m+1);
  cb=cN/(imu - 1.0/Mp1_2);
  sb=sN+p;

  // computation of b_k
  for(int i=T.m+1;i<=T.big_m;i++){
    bk[i]=-Nk[i]/lambda(i);
  }
  b.newvalues(T.m,T.big_m,sb,cb,bk);



  // computation of g_k
  double l,r;
  interval t;
  for(int i=T.m+1;i<=T.big_m;i++){
    #if defined(EXP_FROM_TABLE)
    t=(Right(T0(i))-Right(b(i)))*ks_exp.e_to_k_l(h,imu,i)+ Right(b(i));
    #else
    t=(Right(T0(i))-Right(b(i)))*exp(h*lambda(i))+ Right(b(i));
    #endif
    r=t.right_b();

    #if defined(EXP_FROM_TABLE)
    t=(Left(T0(i))-Left(b(i)))*ks_exp.e_to_k_l(h,imu,i)+ Left(b(i));
    #else
    t=(Left(T0(i))-Left(b(i)))*exp(h*lambda(i))+ Left(b(i));
    #endif
    l=t.left_b();
    gk[i]=interval(l,r);
  }

  delete [] Nk;
  delete [] bk;
}
#endif


#if defined(DIFF_INCL)

void infoT0subsetT(const KuraShi_Tail &T0, const KuraShi_Tail &T)
{
  for(int i=T0.m+1;i<=T0.big_m+1;i++){
    if(!in(T0(i),T(i),0))
        cerr << "violation for i=" << i << "  T0(i)=" << T0(i) << "  T(i)="<< T(i) << endl;
  }
  cerr << "powers: T0.s=" << T0.s << "   T.s="<< T.s << endl;
  cerr << "C - coefficients T0.c=" << T0.c << "  T.c="<< T.c << endl;
  cerr << "T(M+1)^+ - T0(M+1)^+= " << (T(T0.big_m+1).right_b() - T0(T0.big_m+1).right_b())
       << endl;
}

#if !defined(ENCL_TEST)
bool KuraShi::validate_fartail(interval h,const KuraShi_Tail &T0,const KuraShi_Tail &T,
                                          KuraShi_Tail &b)
#else
bool KuraShi::validate_fartail(interval h,const KuraShi_Tail &T0,const KuraShi_Tail &T,
                            KuraShi_Tail &b, ostream &report)
#endif
  // 1 is returned if T is validated, otherwise 0 is returned.
{

  bool fartailvalidated=(T0.subset(T));
  if(!fartailvalidated){
    cerr << "in validate_fartail, T0 is not a subset of T " << endl;
    infoT0subsetT(T0,T);
     #if defined(KRAK) && defined(WIN95)
                   waitBt();
    #endif
    exit(EXIT_FAILURE);
    //return 0;
  }
  if((b.s > T0.s) && (T0.c!=0)){ // this is the first case
    interval L;
    if(b.s-T0.s==2)
      L=sqrt(b.c/T0.c);
    else
      L=exp(log(b.c/T0.c)/(b.s - T0.s));

    L=Right(L);

    #if defined(ENCL_TEST)
    report << "CASE I : in the far tail validation L= " << L.right_b() << "  b.s=" <<  b.s <<
        " T0.s=" << T0.s << endl;
    #endif


    fartailvalidated=true;

    int k=T.big_m+1;
    interval gkk;
    double tkplus,bkplus;
    while(fartailvalidated && (k<=L.right_b())){
      tkplus=T(k).right_b();
      bkplus=b(k).right_b();
      if(tkplus < bkplus){ // if this does not hold, we don't have to make the comparison with g_k^+
 // because in this case T_k^+ > b_k+ > g_k^+
        #if defined(EXP_FROM_TABLE)
        gkk=(Right(T0(k))-bkplus)*ks_exp.e_to_k_l(h,imu,k)+ bkplus;
        #else
        gkk=(Right(T0(k))-bkplus)*exp(h*lambda(k))+ bkplus;
        #endif
        fartailvalidated=(tkplus >= gkk.right_b());
      }
      #if defined(ENCL_TEST)
      if(!fartailvalidated){
        report << " not validated for k=" << k << "  bk=" << bkplus << "  gk="
         << gkk.right_b() << "  T(k)=" << tkplus << "   T0(k)=" << T0(k).right_b() << endl;
      }
      #endif
      k++;
    }
  }
  if((b.s==T0.s) || (T0.c==0)){ // this is the second case in my notes
    fartailvalidated=true;
    if((T0.c).right_b() < (b.c).right_b())
      fartailvalidated=(T.s <= b.s) && (Right(T(T.big_m+1)) >= Right(b(T.big_m+1)) );

     #if defined(ENCL_TEST)
      if(!fartailvalidated){
        report << " far tail validation Case II failed: b.s=T0.s=" << b.s << "  T.s=" << T.s
        << "  T(M+1)^+=" << T(T.big_m+1).right_b() << "   b(M+1)^+=" <<
                b(T.big_m+1).right_b() << endl;
      }
      #endif

  }
  if((b.s < T0.s) && (T0.c!=0)){ // this is the third case
    interval L;
    if(T0.s-b.s==2)
      L=sqrt(T0.c/b.c);
    else
      L=exp(log(T0.c/b.c)/(T0.s-b.s));
    L=Right(L);

    #if defined(ENCL_TEST)
    report << "CASE III : in the far tail validation L= " << L.right_b() << "  b.s=" <<  b.s <<
        " T0.s=" << T0.s << endl;
    #endif
    int L1=int(L.right_b()); // now L1 is the integer part of L
    if(T.big_m > L1) L1=T.big_m;
    fartailvalidated=(T.s <= b.s) && (T(L1+1).right_b() >= b(L1+1).right_b());
    #if defined(ENCL_TEST)
      if(!fartailvalidated){
        report << " not validated for L1=" << L1 << "  b(L1+1)=" <<  b(L1+1).right_b()
          << "  T(L1+1)=" << T(L1+1).right_b() << "   T0(L1+1)=" << T0(L1+1).right_b()
          << endl;
      }
    #endif


  }

  #if defined(ENCL_TEST)
  report.flush();
  #endif



  return(fartailvalidated);
}
#endif


#if defined(DIFF_INCL)
#if !defined(ENCL_TEST)
bool KuraShi::validate_b_fartail(interval h,interval z[],const KuraShi_Tail &T0,const KuraShi_Tail &T)
#else
bool KuraShi::validate_b_fartail(interval h,interval z[],const KuraShi_Tail &T0,const KuraShi_Tail &T, ostream &report)
#endif
  // 1 is returned if T is validated, otherwise 0 is returned.
{



  interval *gk;
  gk=new interval[T0.big_m+1];
  KuraShi_Tail b;

  compute_bk_gk(h,z,T0,T,b,gk);
  delete [] gk;



  #if !defined(ENCL_TEST)
  return validate_fartail(h,T0,T,b);
  #else
  return validate_fartail(h,T0,T,b,report);
  #endif

}
#endif


#if defined(DIFF_INCL)
void KuraShi::new_cand_tailenclosure(const KuraShi_Tail &T0,KuraShi_Tail &T,
             const KuraShi_Tail &b,const interval g[],
             int validated,const int i_c_valid[],int fartailvalidated)
  // the function generates a new guess for the enclosure -
  // and on output stores it in T
{
  // as a rule we allow for refinement (a decrease of T) only if it
  // is already validated
  // we also don't change the power (T.s) in T



  if(validated){
    double l,r ;
    interval temp;
    for(int k=T.m+1;k<=T.big_m;k++){
      // computing the right end of T.a[k]
      if(b(k).right_b() <= T0(k).right_b())
        r=T0(k).right_b();
      else r=g[k].right_b();


      //  now we compute the left end of T.a[k]
      if(b(k).left_b() >= T0(k).left_b())
        l=T0(k).left_b();
      else
        l=g[k].left_b();

      T.a[k]=interval(l,r);
    }

    // we generate the far tail for T, but we want to keep s(T) unchanged!!!
    if(T0.c !=0){
      if((T0.s <= b.s) &&
            (T0(T0.big_m+1).right_b() >= b(T0.big_m+1).right_b())){
        T.newfartail(Right(T0(T0.big_m+1)),T.s);
      }
    }

  }
  if(!validated){  // we use the same approach as in mix_iso
    double d2=1.01;
    interval dg=0.1;
    double l,r;
    interval temp;

    for(int k=T.m+1;k<=T.big_m;k++){
      if(!i_c_valid[k]){  // we don't want to shrink the directions were we already
                          // have the validation, as expansion in other directions
                          // might put it in danger

        // computing the right end of T.a[k]
        r=T(k).right_b();
        if(r < b(k).right_b()){
          if(r < g[k].right_b()){
            temp=(1 - dg)*Right(g[k]) + dg*Right(b(k));
            r=temp.right_b();
            if(r > b(k).right_b()) // this can never happen for perfect arithmetic
                             // but it can with overestimates
               r=b(k).right_b();
          }
        }

        //  now we compute the left end of T.a[k]
        l=T(k).left_b();
        if(l > b(k).left_b()){
          if(l > g[k].left_b()){
            temp=(1-dg)*Left(g[k]) + dg * Left(b(k));
            l=temp.left_b();
            if(l< b(k).left_b())
              l=b(k).left_b();
          }
        }
        T.a[k]=interval(l,r);
        T.a[k]=inflate(T.a[k],d2);
      }
    }


    // we generate the far tail for T, but we want to keep s(T) unchanged!!!
    if(fartailvalidated){
      if(T0.c!=0){
        if((T0.s <= b.s) &&
                  (T0(T0.big_m+1).right_b() >= b(T0.big_m+1).right_b())){
          T.newfartail(Right(T0(T0.big_m+1)),T.s);
        }
      }
    }
    if(!fartailvalidated){
      interval aMp1=Right(d2*b(T.big_m+1));
      interval aMp1_2=Right(T0(T.big_m+1));
      if(aMp1_2 > aMp1) aMp1=aMp1_2;
      T.newfartail(aMp1,T.s);
    }
  }
}
#endif



#if defined(DIFF_INCL)
#if !defined(ENCL_TEST)
int KuraShi::validate_tail(interval h,const interval z[],const KuraShi_Tail &T0,KuraShi_Tail &T,
    int gen_new_T)
#else
int KuraShi::validate_tail(interval h,const interval z[],const KuraShi_Tail &T0,KuraShi_Tail &T,
    int gen_new_T, ostream &report)
#endif
  // 1 is returned if T is validated, otherwise 0 is returned.
  // Additionally if gen_new_T , then T is updated as follows : in case it
  //  is validated then we find better T([0,h]) (T is smaller)
  // otherwise we produce the new guess for $T=T([0,h])$. If !gen_new_T
  // then T is left unchanged.
{



  if(!checkdimpar(dim,T0.m,T.m,T0.big_m,T.big_m)){
    cerr << "bad dimensions in  <KuraShi::validate_tail> " << endl;
    #if defined(WIN95) && defined(KRAK)
      waitBt();
    #endif
    exit(EXIT_FAILURE);
  }




  int validated;

  interval *gk;
  gk=new interval[T0.big_m+1];
  KuraShi_Tail b;

  int *i_c_valid;
  i_c_valid=new int[T0.big_m+1];

  compute_bk_gk(h,z,T0,T,b,gk);


  //  the validation test
  validated=T0.subset(T);
  #if defined(ENCL_TEST)
  if(validated)
    report << "T0 \\subset T " << endl;
  else
    report << "T0 is not contained in T " << endl;
  #endif
  if(!validated){
    cerr << "validate_tail - T0 is not contained in T " << endl;
    infoT0subsetT(T0,T);
    exit(EXIT_FAILURE);
    return 0;
  }


  int k=T.m+1;
  while(k <= T.big_m){
    i_c_valid[k]=1;
    // we first check upper bounds:
    if(T0(k).right_b() < b(k).right_b())
      i_c_valid[k]=(T(k).right_b() >= gk[k].right_b());

    // now we check lower bounds
    if(i_c_valid[k]){
      if(T0(k).left_b() > b(k).left_b())
        i_c_valid[k]=(T(k).left_b() <= gk[k].left_b());
    }
    if(!i_c_valid[k])      validated=0;
    k++;
  }

  #if defined(ENCL_TEST)
  if(validated)
    report << "near tail - validated " << endl;
  else
    report << "near tail - NOT validated " << endl;
  #endif

  #if defined(ENCL_TEST)
    report << "bk  gk  bk \\subset T0   bk \\subset T   k-th dir. validated" << endl;
    for(int k=T0.m+1;k<=T0.big_m;k++){
      report << "k=" << k << "   " << b(k) << "    " << gk[k] <<
 "  " << in(b(k),T0(k),0) << "   " << in(b(k),T(k),0) << "  " << i_c_valid[k] << endl;
    }
    report << "far tail for b " << (b.c).right_b() << "/k^" << b.s
    << "   b[" << T0.big_m+1 << "]=" << b(T0.big_m+1);
    report << endl;
  #endif


  int fartailvalidated;
  // now we check the far tail:

  #if defined(ENCL_TEST)
   fartailvalidated=validate_fartail(h,T0,T,b,report);
  #else
    fartailvalidated=validate_fartail(h,T0,T,b);
  #endif

  #if defined(ENCL_TEST)
  if(fartailvalidated)
    report << "far tail - validated " << endl;
  else
    report << "far tail - NOT validated " << endl;
  #endif
  validated=validated && fartailvalidated;


  if(gen_new_T)
    new_cand_tailenclosure(T0,T,b,gk,validated,i_c_valid,fartailvalidated);

  delete [] gk;
  delete [] i_c_valid;

  #if defined(ENCL_TEST)
  report.flush();
  #endif



  return validated;
}
#endif

#if defined(DIFF_INCL)
void KuraShi::galerkin_errors(const interval x[],const KuraShi_Tail &T, interval galerr[])
// computes the Galerkin projection error (proj. dim=kurashi.dim)
// the result is store in galerr
{
  interval *q;
  q=new interval[T.big_m+1];

  copyvector(x,q,dim);  // dim=T.m
  for(int i=T.m+1;i<=T.big_m;i++)
    q[i]=(T.a)[i];

  ksgalerr_point(q,T.c,T.s,T.m,T.big_m,galerr);
  delete [] q;
}
#endif


#if defined(DIFF_INCL)
#if defined(ENCL_TEST)
int KuraShi::incl_enclosure_m_var(interval h,const interval x[],KuraShi_Tail &T,
     interval enclx[],int useappriorigalerror,interval apriorigal_err[],
     int nomessages,ostream &report)
#else
int KuraShi::incl_enclosure_m_var(interval h,const interval x[],KuraShi_Tail &T,
     interval enclx[],int useappriorigalerror,interval apriorigal_err[],
     int nomessages)
#endif
  // function seeks the enclosure for differential inclusion
  //    x' \in P_m(F(x,T)) for t \in [0,h]; this means that Tail is in T for t \in [0,h]
  // On output: if 1 is returned, then enclx is the desired enclosure
  //    otherwise enclx is unchanged
  // the variable <useappriorigalerror> changes a bit the meaning of the output 1 -
  // namely if it is true, then in the generation of Galerkin errors we take intersection
  //  with locally computed bounds and local bounds - this is very useful in tests
  // runs with large sets
 // nomessages=0 -  if the function  fails then a message is send
  //             to a screen and files
  // nomessages=1 -  the failure does not result in message to screen
{
  // first we have to compute the size of differential inclusion - it will be
  // located in <galerr>
  interval  *galerr;
  galerr=new interval[dim+1];
  galerkin_errors(x,T,galerr);
  if(useappriorigalerror){
     if(!commonpart(galerr,apriorigal_err,galerr,dim)){
       cerr << "KuraShi::incl_e.._m_var - empty intersection of local and ariori gal. err" << endl;
       #if defined(WIN95) && defined(KRAK)
         cerr << "press any key" << endl;
         waitBt();
       #endif
       exit(EXIT_FAILURE);
     }
  }

  int useperturbation=1;
  #if defined(ENCL_TEST)
  int retval=mixodeiso_pert_enclosure(x,h,enclx,useperturbation,galerr,
          nomessages,report);
  #else
  int retval=pert_enclosure(x,h,enclx,useperturbation,galerr,nomessages);
  #endif


  delete [] galerr;
  return retval;
}
#endif


#if defined(DIFF_INCL)
#if defined(ENCL_TEST)
int ksdiagtailenclosure(interval imu,const interval x[],
       KuraShi_Tail &T, interval *smax,interval *CL,ostream &report)
#else
int ksdiagtailenclosure(interval imu,const interval x[],
       KuraShi_Tail &T, interval *smax,interval *CL)
#endif
  // this heuristic   function tries find out whether it is possible
  // to find the enclosure for the far tail (if T.s !=0), for s=T.s
  // over the set defined by a_i \in x[i] for i<=T.m and
  //                         a_i \in T(i) for T.m <i <= T.big_m
  ////////
  // if T.s==0, this means that the far tail is zero, and  we just look for
  //         the largest possible power
  ///////
  // On output:
  //  we update smax and CL,  see below for an explanation
  //   1 - is returned when the heuristic tells that for C larger than CL
  //     one can hope for the enclosure (in this case also T.s <= smax)
  //   0 - if the test tells that no enclosure is possible (T.s > smax)
  //   the <smax> is an upper bound for the power 's' it will make
  //   sense to look for the enclosure
  //

{


  int emptyfartail=(T.c==0);
  int prevTs=T.s;

  interval *a;
  a=new interval[T.big_m+1];
  for(int i=1;i<=T.m;i++)
    a[i]=x[i];
  for(int i=T.m+1;i<=T.big_m;i++)
    a[i]=T(i);

  interval sumabsa=0;
  for(int k=1;k<=T.big_m;k++)
    sumabsa+=abs(a[k]);
  sumabsa=Right(sumabsa);

  interval lh=(imu *power(interval(T.big_m+1),2) - 1)*(2*T.big_m+1)/sumabsa;
  lh=Right(lh);
  *smax=log(lh)/log(2.0) -1;

  int encl_possible;
  if(!emptyfartail)
    encl_possible=(T.s <= *smax) ;
  else{
    encl_possible=(*smax >= 6);
    T.s=int(smax->right_b());
  }


  #if defined(ENCL_TEST)
  report << "FAR TAIL DIAGNOSE BEGIN " << endl;
  report << "s=" << T.s << endl;
  #endif



  // computation of FS_1(k) and FS_C(k) etc


  int kmaxFS1;
  interval maxFS1; // maxFS1=max(FS_1(k), k=M=1,..,2M)=FS_1(kmaxFS1)
  int kmaxksFS1;
  interval  maxksFS1;
      // maxksFS1=max(k^{s-1}FS_1(k), k=M=1,..,2M)=kmaxksFS_1^{s-1}(kmaxksFS1)
  int kmaxFSC;
  interval maxFSC;
  int kmaxksFSC;
  interval maxksFSC;

  int kmaxksFS;
  interval maxksFS;

  for(int k=T.big_m+1;k<=2*T.big_m;k++){
    interval fsk1=0;
    interval ksfsk1=0;

    for(int n=k-T.big_m;2*n < k;n++)
      fsk1+=a[n]*a[k-n];

    fsk1*=2;
    if(k%2==0) fsk1+=power(a[k/2],2);
    fsk1=Right(abs(fsk1));

    ksfsk1=Right(power(interval(k),T.s-1)*fsk1);

    interval fskc=0;
    interval ksfskc=0;
    for(int n=1;n<=k-T.big_m - 1; n++)
      fskc+=Right(abs(a[n])*power(interval(k-n),-T.s));
    fskc*=2;

    ksfskc=Right(power(interval(k),T.s-1)*fskc);


    interval ksfs=Right(ksfsk1 + T.c*ksfskc);

    if(k==T.big_m+1){
      kmaxFS1=k;
      maxFS1=fsk1;
      kmaxksFS1=k;
      maxksFS1=ksfsk1;

      kmaxFSC=k;
      maxFSC=fskc;
      kmaxksFSC=k;
      maxksFSC=ksfskc;

      kmaxksFS=k;
      maxksFS=ksfs;

    }else{
      if(fsk1 > maxFS1){
        kmaxFS1=k;
        maxFS1=fsk1;
      }
      if(ksfsk1 > maxksFS1){
        kmaxksFS1=k;
        maxksFS1=ksfsk1;
      }

      if(fskc > maxFSC){
        kmaxFSC=k;
        maxFSC=fskc;
      }
      if(ksfskc > maxksFSC){
        kmaxksFSC=k;
        maxksFSC=ksfskc;
      }

      if(ksfs > maxksFS){
        kmaxksFS=k;
        maxksFS=ksfs;
      }

    }

    /*
      #if defined(ENCL_TEST)
      report << "k=" << k << "  FS_1=" << fsk1.right_b() << "   k^{s-1}FS_1=" <<
      ksfsk1.right_b() <<  "  FS_C=" << (fskc*T.c).right_b() <<
      "  k^{s-1}FS_C=" << (ksfskc*T.c).right_b() << endl;
      #endif

    */
  }
  interval D1mediumk=maxksFS;
  interval D1largeK;

  D1largeK=sumabsa*power(interval(2),T.s+1)/(2*T.big_m +1);

  if(T.s <= *smax)
    *CL=maxksFS1*(2*T.big_m+1)/(sumabsa*power(interval(2),T.s+1));


  #if defined(ENCL_TEST)
  report << "max FS_1=" << maxFS1.right_b() << "  for k=" << kmaxFS1 << endl;
  report << "max k^{s-1}FS_1=" << maxksFS1.right_b() << "  for k=" <<
            kmaxksFS1 << endl;
  report << "max FS_C/C=" << maxFSC.right_b() << "  for k=" << kmaxFSC << endl;
  report << "max k^{s-1}FS_C/C=" << maxksFSC.right_b() << "  for k=" <<
            kmaxksFSC << endl;

  report << "max k^{s-1} FS(k)= " << maxksFS.right_b() << " for k="
  << kmaxksFS << endl;

  report << endl;


  report << "\\sum_{k=1}^M |a_k|=" << sumabsa.right_b() << endl;

  report << endl;
  report << "dominant term for D1(k>=2M) = " << D1largeK.right_b() << endl;;


  report << "s=" << T.s << "  smax = " << smax->right_b() << endl;
  if(T.s <= *smax)
    report << "large enclosures should exists, because s<= smax " << endl;
  else
    report << "large enclosures impossible, should be s<= smax " << endl;



  if(T.s <= *smax){
    report << "the linear range for enclosure C > " << CL->right_b() ;
    if(!emptyfartail)
      report << " currently T.c=" << T.c.right_b() ;
    report << endl;
    if(!emptyfartail){
      if(T.c <= *CL)
        report << "  outside linear range - enclosure not likely, need to increase C"
               << endl;
      else
        report << "  in linear range - enclosure possible" << endl;
    }

  }
  report << "FAR TAIL DIAGNOSE END " << endl;
  report.flush();
  #endif


  delete [] a;

  if(emptyfartail) T.s=prevTs; // we restore previous value

  return encl_possible;
}

#endif

#if defined(DIFF_INCL)
#if defined(ENCL_TEST)
int geninitfartail(interval imu,const interval x[],KuraShi_Tail &T,ostream& report)
#else
int geninitfartail(interval imu,const interval x[],KuraShi_Tail &T)
#endif
  // this heuristic functions tries to generate a good guess for the far tail enclosure
  // over the set <x> and a tail which is contained in T ; T - is here crude intial
  // guess which hopefully will be improved  (as the enclosure) (by changing the far tail of T)
  // so that T(input) \subset T(output)
  // dim(x)=T.m
  // 1 - is returned, when the diagnositic function finds C and s, such that the enclosure
  //     for the far tail is probable and T is updated accordingly
  // 0 - is returned, when the enclosure for far tail seems to be impossible, with current
  //        settings x,T.m,T.big_m
{
  interval smax,cl;
  int enclpos=0;
  int emptyfartail=(T.s==0);
  #if defined(ENCL_TEST)
  enclpos=ksdiagtailenclosure(imu,x,T,&smax,&cl,report);
  #else
  enclpos=ksdiagtailenclosure(imu,x,T,&smax,&cl);
  #endif
  if(enclpos){
    if(emptyfartail){
      T.s=int(smax.right_b());
      T.c=3*cl;
    }
    else if(3*cl > T.c) T.c=Right(3*cl);
  }else{
    if(!emptyfartail){
      int s_n=int(smax.right_b());
      if(s_n >= 6){ // this is an quite arbitrary requirement, but for
              //sure we need s_n >=4  for various sums to converge
        KuraShi_Tail Tn(T);
        Tn.newfartail(T(T.big_m+1),s_n);
        #if defined(ENCL_TEST)
        enclpos=ksdiagtailenclosure(imu,x,Tn,&smax,&cl,report);
        #else
        enclpos=ksdiagtailenclosure(imu,x,Tn,&smax,&cl);
        #endif
        if(enclpos){
          if(3*cl > Tn.c) Tn.c=Right(3*cl);
          T=Tn;
        }
      }
    }
  }
  return enclpos;
}
#endif


#if defined(DIFF_INCL)
#if !defined(ENCL_TEST)
int KuraShi::enclosure_with_tail(interval h,const interval x[],KuraShi_Tail &T0,
            interval enclx[],KuraShi_Tail &T,int Tinitialguess,
            int useappriorigalerror,interval apriorigal_err[],
            int nomessages )
#else
int KuraShi::enclosure_with_tail(interval h,const interval x[],KuraShi_Tail &T0,
            interval enclx[],KuraShi_Tail &T,int Tinitialguess,
            int useappriorigalerror,interval apriorigal_err[],
            int nomessages, ostream &report )
#endif
  // the function used for the dynamic tail evolution
  //  Input: h - time step, x \oplus  T0 - the inital condition
  //         if(useappriorigalerr) then we always intersect locally computed
  //               Galerkin errors with apriorigal_err - USED in some preparatory(test)
  //              computations when the set as too large
  //  Output: we find <enclx> and T, such that
  //     if(!useappriorigalerr) \varphi([0,h], x\oplus T0) \subset enclx \oplus T
  //     oterwise we have \varphi([0,h], x\oplus T0) \subset enclx \oplus T
  //                under the additional condition that the local Galerkin errors
  //                are contained in apriorigal_err
  //     if this is possible then we return 1
  //   otherwise 0 is returned
  //  //   if Tinitialguess - then inside the function a procedure
  //     building the enclosure for the tail stats with T(input)
  //         otherwise it  starts with T0

  //
    // nomessages=0 -  if the function  fails then the message is send
  //             to a screen and files
  // nomessages=1 -  the failure does not result in the message to screen
{

  if(Left(h)<0){
    cerr << "negative ho - the times step in  'KuraShi::enclosure_with_tail' " << endl;
     #if defined(KRAK) && defined(WIN95)
                   waitBt();
    #endif
    exit(EXIT_FAILURE);
  }

  #if defined(ENCL_TEST)
  report << "BEGIN OF THE enclosure_with_tail FUNCTION " << endl;
  #endif

  int maxdcount=3;  // this number tells that we will try decay rates in from
             // the following range s(T)-maxdcount, ..,s(T)

  interval *initenclx;
  initenclx= new interval[dim+1];

  interval hoh=interval(0,h.right_b());
  for(int i=1;i<=dim;i++)
    initenclx[i]=x[i]+hoh*vfield(i,x);


  KuraShi_Tail Tguess(T0);
  if(Tinitialguess)
    Tguess=T;
  else{
    #if defined(ENCL_TEST)
      geninitfartail(imu,x,Tguess,report);
    #else
      geninitfartail(imu,x,Tguess);  // one may think of using initenclx as the second parameter
      // here
    #endif
  }

  int gen_new=1;

  int validated=0;
  int dcount=0;

  while(!validated && (dcount <= maxdcount)){
    for(int i=1;i<=dim;i++){
      enclx[i]=initenclx[i];
    }
    T=Tguess;
    T.decpower(dcount);


    #if defined(ENCL_TEST)
    report << "dcount=" << dcount << ",  now : T.s=" << T.s << "  T.c=" <<
          T.c.right_b() << endl;
    report.flush();
    #endif


    #if defined(ENCL_TEST)
    if(validate_b_fartail(h,enclx,T0,T,report)){
    #else
    if(validate_b_fartail(h,enclx,T0,T)){
    #endif

      #if defined(ENCL_TEST)
      report << "generating a new  T - validation ignored " << endl;
      validate_tail(h,enclx,T0,T,gen_new,report);
      #else
      validate_tail(h,enclx,T0,T,gen_new);
      #endif


      #if defined(ENCL_TEST)
      report << "*****" << endl;
      report << "after initialization: " << endl;
      report << "enclx:" << endl;
      saveAllCoord(report,enclx,dim,2,1,1);
      T.save(report,0,"T:");
      report << "x \\subset enclx " << MultiIn(x,enclx,0,dim) << endl;
      report << "T0 \\subset T " << T0.subset(T) << endl;
      #endif


      int max_iter=T0.big_m/2;
      int iter=1;
      #if defined(KS_ENCL_ITER_TEST)
      while((!validated || in_enclmaxiter ) && (iter <= max_iter)){
      #else
      while((!validated) && (iter <= max_iter)){
      #endif
        #if defined(ENCL_TEST)
        int oldvalidated=validated;
        report << "*****" << endl;
        report << "iteration = " << iter << endl;
        #endif
        #if defined(ENCL_TEST)
        if(incl_enclosure_m_var(h,x,T,enclx,useappriorigalerror,apriorigal_err,
               nomessages,report)){
        #else
        if(incl_enclosure_m_var(h,x,T,enclx,useappriorigalerror,apriorigal_err,
               nomessages)){
        #endif

          #if defined(ENCL_TEST)
          report << "enclosure for m-variables OK" << endl;
          report << "enclx:" << endl;
          saveAllCoord(report,enclx,dim,2,1,1);
          report << "x \\subset enclx " << MultiIn(x,enclx,0,dim) << endl;
          #endif

          #if defined(ENCL_TEST)
          validated=validate_tail(h,enclx,T0,T,gen_new,report);
          #else
          validated=validate_tail(h,enclx,T0,T,gen_new);
          #endif


          /// if useappriorigalerror - then it will also make sense
          // in case the tail is not validated to take an intersection
          //  with some apriori given tail or just to take this apriori given tail
          //  instead - anyway this might require another parameter for an apriori
          // tail

          #if defined(ENCL_TEST)
          report << "tail validation - " << validated << endl;
          T.save(report,0,"T:");
          report << "T0 \\subset T " << T0.subset(T) << endl;
          #endif
        }
        #if defined(ENCL_TEST)
        else{
          report << "enclosure for m-variables not generated " << endl;
        }
        #endif

        #if defined(ENCL_TEST)
        if(oldvalidated && !validated){
          report << "IT NOT VALIDATED NOW< but it was BEFORE " << endl;
          return 0;
        }
        #endif
        iter=iter+1;
      }
    } // endif fartailvalidated
    #if defined(ENCL_TEST)
    report << "dcount=" << dcount << "  validated=" << validated << endl;
    report << "----------------" << endl;
    #endif
    dcount++;
  }



  delete [] initenclx;
  initenclx=NULL;


  if(validated){
    // refinement loop
    int max_iter=1;
    int iter=1;
    while(iter <= max_iter){
      #if defined(ENCL_TEST)
      int okencl=incl_enclosure_m_var(h,x,T,enclx,useappriorigalerror,apriorigal_err,
               nomessages,report);
      #else
      int okencl=incl_enclosure_m_var(h,x,T,enclx,useappriorigalerror,apriorigal_err,nomessages);
      #endif
      #if defined(ENCL_TEST)
      int oktail=validate_tail(h,enclx,T0,T,gen_new,report);
      #else
      int oktail=validate_tail(h,enclx,T0,T,gen_new);
      #endif

      #if defined(ENCL_TEST)
       report << " After refinement iter=" << iter << endl;
       report << "enclosure in x-var " << okencl << endl;
       report << "enclx:" << endl;
       saveAllCoord(report,enclx,dim,2,1,1);
       report << "x \\subset enclx " << MultiIn(x,enclx,0,dim) << endl;

       report << "Tail validation - " << oktail << endl;
       T.save(report,0,"T:");
       report << "T0 \\subset T " << T0.subset(T) << endl;

      #endif

      iter=iter+1;
    }
  }
 #if defined(ENCL_TEST)
  report << "END OF THE enclosure_with_tail FUNCTION " << endl;
  report.flush();
  #endif
  return validated;
}
#endif

#if defined(DIFF_INCL)
int KuraShi::move_tail(interval h,interval W2[],KuraShi_Tail &T0,
        KuraShi_Tail &T, KuraShi_Tail &Th)
// moves forward the tail - the resulting tail is Th
// we assume that :
//  T0 - is the initial tail
//   T - is the enclosure for tail for t \in [0,h]
//   W2 - the enclosure for leading variables for t \in [0,h]
// 1- is returned if everything is ok
// 0 - otherwise
{
  int validated;

  KuraShi_Tail b;
  KuraShi_Tail TTh(T0.m,T0.big_m);

  compute_bk_gk(h,W2,T0,T,b,TTh.a); // this takes care of the
                // near tail  Th(i) for m < i <= M(big_m)

  interval tmp;
  tmp=-4*h*imu*power(interval(T0.big_m+1),4);
  tmp+=2*h*power(interval(T0.big_m+1),2);
  tmp+=b.s - T0.s;  // now temp= - 4h*mu*(M+1)^4 + 2h*(M+1)^2 + s(b)-s(T(0))
  validated=(tmp <= 0);

  if(validated)
    validated=(4*imu*power(interval(T0.big_m+1),2) >= 1);

  if(validated){
    TTh.s=b.s;
    interval E;
    #if defined(EXP_FROM_TABLE)
    E=ks_exp.e_to_k_l(h,imu,T0.big_m+1);
    #else
    E=exp(h*lambda(T0.big_m+1));
    #endif
    E*=power(interval(T0.big_m +1),b.s - T0.s);
    TTh.c=T0.c*E + b.c;
  }
  if(validated)
    Th=TTh;

  if(validated){
    if(!Th.subset(T)){
      commonpart(Th,T,Th);
    }
  }

  return validated;
}
#endif

// DOUBLE versions



double KuraShi::vfield(int k,const double x[])
    /* int k; */ /* an index of coordinate */
   /*  double x[]; */ /* the argument of the vector field, coordinates start
     form 1 */
{
  double retval; /* returned value */
  double sum=0.0; /* \Sum_{m=-\infty}^{m=\infty} a_m a_{m-k} */
  int m,j1,j2,sign;
  for(m=-dim;m <= dim;m++){
    sign=1;
    j1=m;
    if (j1 < 0) {j1=-j1;sign=-sign;}

    j2=k-m;
    if (j2 < 0) {j2=-j2; sign=-sign;}

    if ((j1 > 0) && (j1 <= dim) && (j2 > 0) && (j2 <= dim)){
      double xj1,xj2;
      xj1=x[j1];

      xj2=x[j2];

      if(j1!=j2)  sum+= sign*xj1*xj2;
      else sum+=sign*pow(xj1,2);
    }
  }
  int k2=k*k;
  retval=k2*(1 - mu*k2)*x[k] - k*sum;
  return(retval);
}


double KuraShi::vfieldndiagt(int k,const double x[])
 // returns f_k - k^2(1-mu k^2)a_k
{
  int k2=k*k;
  return(vfield(k,x) - k2*(1-mu*k2)*x[k]);
}



double KuraShi::dvfield(int k, int i,const double x[])
     /* calculates Df_k/Dx_i (x) */
{
  double retval; /* returned value */
  double sum=0.0; /* D/Dx_i \Sum_{m=-\infty}^{m=\infty} a_m a_{m-k} */
  int m,j1,j2,sign;
  for(m=-dim;m<=dim;m++){
    sign=1;
    j1=m;
    if (j1 < 0) {j1=-j1;sign=-sign;}

    j2=k-m;
    if (j2 < 0) {j2=-j2; sign=-sign;}

    if ((j1 > 0) && (j1 <= dim) && (j2 > 0) && (j2 <= dim)){
      double xj1,xj2;
      xj1=x[j1];

      xj2=x[j2];


      if (j1==i) sum+=  sign*xj2;
      if (j2==i) sum+=  sign*xj1;
    }
  }
  retval= - k*sum;

  if (k==i) {
     int k2=k*k;
     retval=k2*(1 - mu*k2) + retval;
  }
  return(retval);
}

double KuraShi::d2vfield(int k,int i,int j,const double[] /*x[]*/)
     /* calculates D^2f_k/Dx_iDx_j (x) */
{
  double retval=0.0;
  if ((-i+j==k) || (i-j==k)) retval=-2.0;
  if (i+j==k) retval=2.0;

  retval*=-k;
  return(retval);
}

double KuraShi::parameter_value(const char par[])
  // returns the value of the parameter par
{
  double retval=0;
  if(strcmp(par,"mu")==0)
     retval=mu;

  return retval;
}

void KuraShi::writeparameters(ostream &out)
{
  int oldprecision=out.precision();
  out.precision(10);
  out << infostr << "  ";
  if(diameter(imu)==0){
    out << "mu=" <<   mu << "  ";
    out << "alpha=" <<  mutoalpha(mu) << endl;
  } else{
     writeinterval(out,imu,2);
     out << "   alpha=" << mutoalpha(imu) << endl;
  }
  out << "dim=" << dim << endl;
  char buffer[80];
  out << "Rough enclosure: " ;
  out << ksencl_name(rough_encl_func_type,buffer,80) << endl;
  out << "C1 -rough enclosure: ";
  out << ksencl_name(c1rough_encl_func_type,buffer,80) << endl;
  #if defined(KS_ENCL_ITER_TEST)
  if(in_enclmaxiter)
    out << "forcing the same number of iterations in enclosures " << endl;
  else
    out << "enclosures iterated till validation only" << endl;
  #endif
  out.precision(oldprecision);
}

#ifdef KRAK
void KuraShi::writeparameters(Frame &out)
{
  out << infostr << "  ";
  out << "mu=" << imu << "  ";
  out << "dim=" << dim << "\n";
}
#endif



int chooseksp(double mu)
// the choice of an initial condition for KS-equations
// the returned value should be later feeded to function <KuramotoPoint>
// mu - is suggesting the interesting range of parameters
{

  cout << "-----------------------------"<< endl;
  cout << "select the initial point " << endl;
  if(mu < 0.1)
    cout << " 1 - from an attractor for mu=0.02991 " << endl;
  if(mu > 0.09){
    cout << " 2 - from the symmetric attracting  cycle for mu=0.125 "
               << endl;
    cout << " 3 - from the symmetric attracting  cycle for mu=0.127 "
               << endl;
    cout << " 4 - from the symmetric attracting  cycle for mu=0.128 "
  << endl;
    cout << " 5 - from the non-symmetric attracting cycle for mu=0.1215 (past bif.)  "
  << endl;
    cout << " 6 - from the symmetric unstable cycle for mu=0.1215 (past bif.)  "
      << endl;
    cout << " 7 - from nonsymmetric unstable cycle for mu=0.1212 on chaotic attractor "
      << endl;
    cout << " 8 - from symmetric unstable cycle for mu=0.1212  NOT on  attractor "
      << endl;
    cout << " 9 - from unstable period-2 for mu=0.1212   on  attractor "
      << endl;
    cout << " 10 - from unstable period-3 for mu=0.1212 on attractor(????)" << endl;
  }
  if(mu < 0.1){
    cout << " 21 - from non-sym. per. orbit (unstable) for mu=0.02991 on the attractor very problematic" << endl;
    cout << " 22 - from the attracting period two orbit for mu=0.03 (doubious) " << endl;
    cout << " 23 - an attracting period three orbit for mu=0.029924 " << endl;
    cout << " 24 - from symmetric attracting periodic orbit for mu=0.032 "
         << endl;
    cout << " 25 - from symmetric unstable periodic orbit for mu=0.02991  "
         << endl;
  }
  int pno;
  cin >> pno;
  #if defined(WIN95) && defined(KRAK)
  cout << endl;
  #endif
  int ret=0;
  if(pno==1) ret=201;
  if(pno==2) ret=101;
  if(pno==3) ret=109;
  if(pno==4) ret=102;
  if(pno==5) ret=108;
  if(pno==6) ret=110;
  if(pno==7) ret=112;
  if(pno==8) ret=113;
  if(pno==9) ret=114;
  if(pno==10) ret=115;
  if(pno==21) ret=103;
  if(pno==22) ret=105;
  if(pno==23) ret=106;
  if(pno==24) ret=107;
  if(pno==25) ret=111;
  return(ret);
}




void KuramotoPoint(interval y[],int ydim,int numpt )
  //  numpt=201 - mu=0.02991 - a point from the attractor on the 8-th section
  //  numpt= 101 - mu \approx 0.125 - from limit cycle
  //  numpt=3 - attracting fixed point fixed mu=0.125 bimodal stable
  //        31 - a fixed point  mu=0.127 bimodal unstable
  //       5 - an attracting fixed point, mu=0.08 - causing problems
  //          computed in 12D; differes from the one computed in 10D
  //               on third digit
  //       6 - an 'negative' attracting point form the 'giant' branch for
  //           mu=0.06 - may be used as a starting point for the search for
  //           the nearby values
  ///////////
  //  numbers from 3-99 - fixed points
  //               101 - 199 - periodic orbits
  //               201 - 299 - interesting initial cond. for attractor
{
  zero(y,ydim);
  if(numpt==201){ // a point from the attractor on the 8-th section
                   // mu=0.02991
    interval x[17];

    x[1]=-8.104437e-01;
    x[2]=-1.328997e-01;
    x[3]= 8.104437e-01;
    x[4]= -2.312598e+00;
    x[5]= -8.737371e-01;
    x[6]=  -4.433101e-01;
    x[7]=  7.625193e-01;
    x[8]=  -4.352889e-01;
    x[9]=  -2.543029e-01;
    x[10]=  -2.011455e-01;
    x[11]=  1.111975e-01;
    x[12]=  -1.074697e-02;
    x[13]=  -1.756767e-02;
    x[14]=  -3.699506e-02;
    x[15]=  6.922592e-03;
    x[16]=  1.809564e-03;
    for(int i=1;i<=ydim;i++){
      if(i<=16) y[i]=x[i];
      else y[i]=0;
    }
  }
  if(numpt==101){     //   mu \approx 0.125 - from limit cycle
    interval x[10];
    x[1]=0.0;
    x[2]=1.02655;
    x[3]=-0.63229;
    x[4]=-0.213039;
    x[5]=0.122888;
    x[6]=0.000432719;
    x[7]=-0.0144715;
    x[8]=0.00200219;
    x[9]=0.000999897;
    for(int i=1;i<=ydim;i++){
      if(i<=9) y[i]=x[i];
      else y[i]=0.0;
    }

  }
  if(numpt==31){ //  fixed point  mu=0.127 bimodal unstable
    interval x[11];
    for(int i=1;i<=10;i++)  x[i]=0.0;
    x[2]= 1.46196;
    x[4]=-0.471281;
    x[6]=0.0632959;
    x[8]=-0.0071052;
    x[10]=0.000687481;
    for(int i=1;i<=ydim;i++){
      if(i<=10) y[i]=x[i];
      else y[i]=0.0;
    }
  }

  if(numpt==3){ // attracting fixed point fixed mu=0.125
                // bimodal stable
    interval x[11];
    for(int i=1;i<=10;i++)  x[i]=0.0;
    x[2]=-1.4523;
    x[4]=-0.478232;
    x[6]=-0.0650864;
    x[8]= -0.00742211;
    x[10]=-0.00072879;
    for(int i=1;i<=ydim;i++){
      if(i<=10) y[i]=x[i];
      else y[i]=0.0;
    }
  }


  if(numpt==5){ // attracting fixed point, mu=0.08 - causing problems
       // computed in 12D; differs from the one computed in 10D on third digit
    interval x[13];
    for(int i=1;i<=12;i++)  x[i]=0.0;

    x[1]=-0.0528592;
    x[2]= 0.144535;
    x[3]=-2.15928;
    x[4]=-0.182685;
    x[5]=0.114066;
    x[6]=-0.393727;
    x[7]=-0.0418124;
    x[8]=0.0171228;
    x[9]=-0.0331338;
    x[10]=-0.00495295;
    x[11]=0.00165249;
    x[12]=-0.00222362 ;

    /*    x[1]=-0.0520911;
    x[2]=0.142486;
    x[3]=-2.16006;
    x[4]=-0.179964;
    x[5]=0.112626;
    x[6]=-0.39428;
    x[7]=-0.0411408;
    x[8]=0.0171219;
    x[9]=-0.0334302;
    x[10]=-0.00486702;  */
    for(int i=1;i<=ydim;i++){
      if(i<=12) y[i]=x[i];
      else y[i]=0.0;
    }
  }
  if(numpt==6){ // an 'negative' attracting point form the 'giant' branch for
                //  mu=0.06 - may be used as a starting point for the search for
                //  the nearby values
    double x[21];
    x[1]=-0.810822;
    x[2]=-1.24773;
    x[3]=0.334719;
    x[4]=-0.968871;
    x[5]=-0.0839611;
    x[6]=-0.350698;
    x[7]=-0.00116337;
    x[8]=-0.0751995;
    x[9]=-0.000733089;
    x[10]=-0.017283;
    x[11]=-0.000513207;
    x[12]=-0.0033864;
    x[13]=-7.63527e-05;
    x[14]=-0.000625608;
    x[15]=-1.78304e-05;
    x[16]=-0.000112751;
    x[17]=-3.61616e-06;
    x[18]=-1.94846e-05;
    x[19]=-6.70907e-07;
    x[20]=-3.29675e-06;
    for(int i=1;i<=ydim;i++){
      if(i<=20) y[i]=x[i];
        else y[i]=0.0;
    }
  }
  if(numpt==102){ // symmetric periodic orbit for mu=0.128
                  // on 1-st section a_1=0, a_1' >0
    double x[17];
    x[1]=0;
    x[2]=1.23336;
    x[3]=0.535385;
    x[4]=-0.311235;
    x[5]=-0.122156;
    x[6]=0.0207885;
    x[7]=0.0171866;
    x[8]=-0.000246737;
    x[9]=-0.00166447;
    x[10]=-0.000169709;
    x[11]=0.000126062;
    x[12]=2.78657e-05;
    x[13]=-7.2107e-06;
    x[14]=-2.97039e-06;
    x[15]=2.44863e-07;
    x[16]=2.50716e-07;

    for(int i=1;i<=ydim;i++){
      if(i<=16) y[i]=x[i];
        else y[i]=0.0;
    }
  }
  if(numpt==103){ // an unstable non-symmetric periodic orbit on
                  // the attractor for mu=0.02991 on 5-th section a_1=0, a_1' <0
    double x[17];
    x[1]=0;
    x[2]=0.00358636;
    x[3]=-0.667183;
    x[4]=-1.42152;
    x[5]=1.35111;
    x[6]=-1.4484;
    x[7]=-0.3835;
    x[8]=-0.00738814;
    x[9]=0.123488;
    x[10]=-0.316915;
    x[11]=0.0997181;
    x[12]=-0.0246127;
    x[13]=-0.0221222;
    x[14]=-0.0183195;
    x[15]=0.0169374;
    x[16]=-0.0112737;
    for(int i=1;i<=ydim;i++){
      if(i<=16) y[i]=x[i];
        else y[i]=0.0;
    }
  }
  if(numpt==105){ // an attracting  periodic period 2 orbit for mu=0.03 on the first section
                       // the leading eigenvalue is 0.991, 21.IV.2005
                       // when I recomputed it on LINUX (sec=1, standard) -0.6094
                       // this orbit appears to be rather DOUBIOUS - when using perprendicular
                       // section it could not be found
    double x[17];
    x[1]=0.0;
    x[2]=0.125831;
    x[3]=0.521673;
    x[4]=-1.13984;
    x[5]=-0.944662;
    x[6]=-1.30027;
    x[7]=0.244631;
    x[8]=0.0156347;
    x[9]=-0.0462797;
    x[10]=-0.200199;
    x[11]=-0.0672693;
    x[12]=-0.0289011;
    x[13]=0.0147617;
    x[14]=-0.00687439;
    x[15]=-0.00726163;
    x[16]=-0.00662688;
    for(int i=1;i<=ydim;i++){
      if(i<=16) y[i]=x[i];
        else y[i]=0.0;
    }
  }
  if(numpt==106){ // an attracting periodic period 3 orbit for mu=0.029924
                  // on the first section, cannot be found using Newton function
                  /// must be very weakly attracting
    double x[17];
    x[1]=0.0;
    x[2]=0.15662;
    x[3]=0.497611;
    x[4]=-1.16707;
    x[5]=-0.913971;
    x[6]=-1.26025;
    x[7]=0.251415;
    x[8]=0.00944373;
    x[9]=-0.0535157;
    x[10]=-0.196929;
    x[11]=-0.0609044;
    x[12]=-0.0263872;
    x[13]=  0.0140049;
    x[14]=  -0.00751349;
    x[15]=  -0.00713512;
    x[16]=  -0.00617119;
    for(int i=1;i<=ydim;i++){
      if(i<=16) y[i]=x[i];
        else y[i]=0.0;
    }
  }

  if(numpt==107){ // an attracting symmetric periodic orbit for mu=0.032
    // on the first section a_1=0, a_1' > 0
    double x[21];
    x[1]=0.0;
    x[2]=-0.0527352;
    x[3]= 0.748426 ;
    x[4]= -1.69971 ;
    x[5]= -1.37841 ;
    x[6]= -1.14589;
    x[7]=  0.467857;
    x[8]=  -0.0816581 ;
    x[9]= -0.184337 ;
    x[10]= -0.291846;
    x[11]=  -0.0488261 ;
    x[12]= -0.00203389 ;
    x[13]= 0.0116754 ;
    x[14]= -0.0245463;
    x[15]=  -0.0140966 ;
    x[16]= -0.00581978 ;
    x[17]= 0.001412 ;
    x[18]= -0.000415157;
    x[19]=  -0.000962116 ;
    x[20]= -0.000982133 ;
    for(int i=1;i<=ydim;i++){
      if(i<=20) y[i]=x[i];
        else y[i]=0.0;
    }
  }

  if(numpt==108){ // the attracting nonsymmetric periodic orbit for mu=0.1215
    // not long after the bifurcation (period doubling for half-Poincare)
    //  from the symmetric attracting periodic orbit
    // on the  first section
    double x[21];
    x[1]=0;
    x[2]=0.85927;
    x[3]=0.681639;
    x[4]=-0.149051;
    x[5]=-0.114404;
    x[6]=-0.0114424;
    x[7]=0.0114077;
    x[8]=0.00285685;
    x[9]=-0.000470889;
    x[10]=-0.000331671;
    x[11]=-1.56431e-05;
    x[12]=2.29122e-05;
    x[13]=4.94772e-06;
    x[14]=-9.0299e-07;
    x[15]=-4.96895e-07;
    x[16]=-1.45976e-08;
    x[17]=3.22344e-08;
    x[18]=5.97201e-09;
    x[19]=-1.25442e-09;
    x[20]=-5.8994e-10;
    for(int i=1;i<=ydim;i++){
      if(i<=20) y[i]=x[i];
        else y[i]=0.0;
    }
  }

  if(numpt==109){ // the attracting symmetric periodic orbit for mu=0.127
    // on the  second section
    double x[17];
    x[1]=0.201211;  x[2]=1.28998;  x[3]=  0.201211;  x[4]=  -0.377866;
    x[5]=  -0.0423095;  x[6]=  0.0431616;  x[7]=  0.00694021;
    x[8]=  -0.00415649;  x[9]=  -0.00079449;  x[10]=  0.000331606;
    x[11]=  7.93919e-05;  x[12]=  -2.39081e-05;  x[13]=  -7.08168e-06;
    x[14]=  1.56836e-06;  x[15]=  5.84928e-07;  x[16]=  -9.42032e-08;
    for(int i=1;i<=ydim;i++){
      if(i<=16) y[i]=x[i];
        else y[i]=0.0;
    }
  }


  if(numpt==110){ // the  unstable  symmetric periodic orbit for mu=0.1215
    // not long after the bifurcation (period doubling for half-Poincare)
    // on the  second section section
    double x[14];
    x[1]=0.245003;  x[2]=1.04151; x[3]=0.245003; x[4]=-0.276077;
    x[5]=-0.0437142; x[6]=0.025314;  x[7]=0.00634611;  x[8]=-0.00199681;
    x[9]=-0.000617746;  x[10]=0.000118483; x[11]=5.26952e-05;
    x[12]=-5.77491e-06; x[13]=-3.94999e-06;
    for(int i=1;i<=ydim;i++){
      if(i<=13) y[i]=x[i];
        else y[i]=0.0;
    }
  }





  if(numpt==111){ // the unstable symmetric periodic orbit for mu=0.02991
    // on the  second section
    double x[26];
    x[1]=0.26893;  x[2]=0.217827; x[3]=0.26893; x[4]=-1.73395; x[5]=-1.15829;
    x[6]=-0.745116;  x[7]=0.451722;  x[8]=-0.217103; x[9]=-0.263588;
    x[10]=-0.203271;  x[11]=0.00102835; x[12]=-0.000706661; x[13]=-0.0109868;
    x[14]=-0.0267955; x[15]=-0.00765627; x[16]=-0.000913738;
    x[17]=0.000396579; ; x[18]=-0.00173248;  ; x[19]=-0.00112683;
    x[20]=-0.000417885; x[21]=3.42432e-05; x[22]=-5.50541e-05;
    x[23]=-8.24281e-05; x[24]=-5.72439e-05; ; x[25]=-9.69029e-06;
    for(int i=1;i<=ydim;i++){
      if(i<=25) y[i]=x[i];
        else y[i]=0.0;
    }
  }
  if(numpt==112){ // non-symmetric unstable orbit  for mu=0.1212 on attractor
                // on the second section

    double x[14];
    x[1]=0.260827;  x[2]=1.11512;  x[3]=0.260827;  ;  x[4]=-0.32086;
    x[5]=-0.0495389;  x[6]=0.0319916;  x[7]=0.00780235;  x[8]=-0.00276601;
    x[9]=-0.000819602;  x[10]=0.000182703;  x[11]=7.57513e-05;
    x[12]=-1.02432e-05;  x[13]=-6.16247e-06;


/*    double x[15];
    x[1]=0.261008; x[2]=1.11628; x[3]=0.261008; x[4]=-0.321566;
    x[5]=-0.0496197; x[6]=0.0321027; x[7]=0.00782415; x[8]=-0.00277937;
    x[9]=-0.000822814; x[10]=0.000183884; x[11]=7.61368e-05; x[12]=-1.03253e-05;
    x[13]=-6.19632e-06; ; x[14]=4.34984e-07; */
    for(int i=1;i<=ydim;i++){
      if(i<=13) y[i]=x[i];
        else y[i]=0.0;
    }
  }
  if(numpt==113){ // symmetric unstable orbit  for mu=0.1212 not on attractor
                // on the second section

    double x[14];
    x[1]=0.245372; x[2]=1.02694; x[3]=0.245372; x[4]=-0.27012;
    x[5]=-0.0433005; x[6]=0.0243929; x[7]=0.00622941; x[8]=-0.001897;
    x[9]=-0.000599527; x[10]=0.000110037; x[11]=5.05721e-05; x[12]=-5.17827e-06;
    x[13]=-3.74419e-06;
    for(int i=1;i<=ydim;i++){
      if(i<=13) y[i]=x[i];
        else y[i]=0.0;
    }
  }

  if(numpt==114){ // unstable period 2 orbit  for mu=0.1212 on attractor
                // on the second section

    double x[15];
   /* x[1]=0.257165;  x[2]=1.07881;  x[3]=0.257165;  x[4]=-0.300173;
    x[5]=-0.0473921;  x[6]=0.0287246;  x[7]=0.00721166;  x[8]=-0.00238033;
    x[9]=-0.000731568; x[10]=0.000148836; x[11]=6.52231e-05; x[12]=-7.74406e-06;
    x[13]=-5.10615e-06; */


   x[1]=0.257292; x[2]=1.07967; x[3]=0.257292; x[4]=-0.300664; x[5]=-0.0474495;
   x[6]=0.0287995; x[7]=0.00722656; x[8]=-0.00238898; x[9]=-0.000733698;
   x[10]=0.000149569; x[11]=6.54716e-05; x[12]=-7.79684e-06; x[13]=-5.13336e-06;
   x[14]=2.81989e-07;


    for(int i=1;i<=ydim;i++){
      if(i<=14) y[i]=x[i];
        else y[i]=0.0;
    }
  }

  if(numpt==115){ // unstable period 3 orbit  for mu=0.1212 I do not know whether it is on attractor
                  // on the second section

    double x[14];
    x[1]=0.257318;  x[2]=1.0768;  x[3]=0.257318; x[4]=-0.299169; x[5]=-0.0473338; x[6]=0.0285561; x[7]=0.00719115;
    x[8]=-0.00236067; x[9]=-0.000728081; x[10]=0.000147068; x[11]=6.47882e-005; x[12]=-7.61464e-006;
    x[13]=-5.06513e-006;
    for(int i=1;i<=ydim;i++){
      if(i<=13) y[i]=x[i];
      else y[i]=0.0;
    }

  }


  if(numpt==202){ // an initial condition which 'converges' to
    // second (non-symmetric) transient attractor for mu=0.02991
    //- there must be something invariant and unstable there
 /*   double x[4];
    x[1]=-1.5;
    x[2]=0;
    x[3]=-1.5 + 3.0/8;
    for(int i=1;i<=ydim;i++){
      if(i<=3) y[i]=x[i];
      else y[i]=0.0;
    }  */
     double x[17];
     x[1]=-1.01877;
     x[2]=-0.234377;
     x[3]=0.904268;
     x[4]=-1.76824;
     x[5]=-0.961397;
     x[6]=-1.26838;
     x[7]=0.143954;
     x[8]=-0.306959;
     x[9]=-0.121087;
     x[10]=-0.284552;
     x[11]=-0.0603894;
     x[12]=-0.0610997;
     x[13]=-0.00596099;
     x[14]=-0.0290116;
     x[15]=-0.0105687;
     x[16]=-0.0106402;

     for(int i=1;i<=ydim;i++){
      if(i<=16) y[i]=x[i];
        else y[i]=0.0;
    }
  }
}


void KuramotoPoint(double y[],int ydim,int numpt)
{
  interval x[MAX_DIM+1];
  KuramotoPoint(x,ydim,numpt);
  for(int i=1; i<=ydim; i++)
   y[i]=(x[i]).right_b();
}


void KuramotoSection(interval c_sec[],int sec_no,int pdim)
///// All sections are in 1-3 plane
///// the trajectories rotate in clockwise direction
//  1 - x[1]=0, x'[1] >0,               angle Pi/2
//  2 - x[1]-x[3]=0  ; x'[1]-x[3]' >0   angle Pi/4
//  3 - x[3]=0,   x'[3] <0              angle 0
//  4 - x[1]+x[3]=0, x'[1]+x'[3] <0     angle -Pi/4
//  5 - x[1]=0,  x'[1]<0                      -Pi/2
//  6 - -x[1]+x[3]=0,  -x'[1]+x'[3] >0        -3Pi/4
//  7 - x[3]=0,   x'[3] >0                    -Pi
//  8 - x[1]+x[3]=0, x'[1]+x'[3] >0           -5Pi/4
{
  for(int i=0;i<=pdim;i++)
    c_sec[i]=0.0;
  switch(sec_no){
    case 1: c_sec[1]=1.0;break;  /* x[1]=0, x[1] grows , the angle is decreasing  */
    case 2: c_sec[1]=1.0;  /* angle= Pi/4 and decreases */
            c_sec[3]=-1.0; break;
    case 3: c_sec[3]=-1.0; break; /* we cut the axis  x[3] from above*/
    case 4: c_sec[1]= -1.0; /* angle= - Pi/4 */
            c_sec[3]=-1.0; break;
    case 5: c_sec[1]=-1.0; break; /* a OY-axis - a bottom part */
    case 6: c_sec[1]=-1.0;  /* an angle= -3Pi/4 */
            c_sec[3]=1.0; break;
    case 7: c_sec[3]=1.0; break;  /* a left half of OX-axis */
    case 8: c_sec[1]=1.0; /* an angle= 3Pi/4 */
            c_sec[3]=1.0; break;
  }

}

void KuramotoSection(double c_sec[],int sec_no,int pdim)
{
  interval x[MAX_DIM+1];
  KuramotoSection(x,sec_no,pdim);
  for(int i=0; i<=pdim; i++)
    c_sec[i]=(x[i]).right_b();
}


void KuramotoRadialDir(interval vect[],int numsec,int pdim)
//  in the plane 1-3 , we produce the radial vector,
//   numsec=1, coresponds  to the angle Pi/2 (upper half of OX axis)
//  and we continue clockwise
//  with an step -Pi/4  for numsec=1,...,8
{
  for(int i=0;i<=pdim;i++)
    vect[i]=0.0;
  switch(numsec){
  case 1: vect[3]=1.0;break;  // Pi/2
  case 2: vect[1]=1.0;  //  Pi/4
          vect[3]=1.0; break;
  case 3: vect[1]=1.0; break; // 0
  case 4: vect[1]= 1.0; // -Pi/4
          vect[3]=-1.0; break;
  case 5: vect[3]=-1.0; break; //   OY - bottom part
  case 6: vect[1]=-1.0;  // -3Pi/4
          vect[3]=-1.0; break;
  case 7: vect[1]=-1.0; break;  // a left-half axis OX
  case 8: vect[1]=-1.0; //  angle 3Pi/4
          vect[3]=1.0; break;
  }
}

void KuramotoRadialDir(double vect[],int numsec,int pdim)
{
  interval *ivect;
  ivect=new interval[pdim+1];
  KuramotoRadialDir(ivect,numsec,pdim);
  todouble(ivect,vect,pdim);
  delete [] ivect;
}


void KuramotoColVectDir(double colvect[][MAX_DIM+1],int *numcolvect,
      int sec_no,int pdim,int betweenS)
  // procedure returns directions in which we want the coherent coordinates
  // to point (i.e. to have positive scalar product with them)
  // these directions are given by colvect[1]  and eventually colvect[2]
  // if betweenS - then colvect[1] is the section direction
  //          and colvect[2] is radial direction in the plane (a1,a3)
  //          on output also *numcolvect=2
  // if !betweenS - we set colvect[1] only, as the radial direction
  //    in the plane (a1,a3), in this case *numcolvect=1
{
  *numcolvect=0;
  if(betweenS){
    *numcolvect=1;
    KuramotoSection(colvect[1],sec_no,pdim);
  }
  (*numcolvect)++;
  KuramotoRadialDir(colvect[*numcolvect],sec_no,pdim);

}

/// symmetry functions

void kssymmetry(interval x[],int dim)
//  the symmetry changing signs on odd modes is applied to <x>
{
  for(int i=1;i<=dim;i++)
    if((i % 2) ==1) x[i]=-x[i];
}

void kssymmetry(interval Dp[][MAX_DIM+1],int dim)
//  the symmetry changing signs on odd modes is applied to the square
//   dim x dim matrix Dp
{
   for(int i=1;i<=dim;i++){
     int isign=1;
     if((i % 2) ==1) isign=-1;
     for(int j=1;j<=dim;j++){
       int jsign=1;
       if((j % 2) ==1) jsign=-1;
       Dp[i][j]*=(isign*jsign);
     }
   }
}



void kssymmetry(double x[],int dim)
//  the symmetry changing signs on odd modes is applied to <x>
{
  for(int i=1;i<=dim;i++)
    if((i % 2) ==1) x[i]=-x[i];
}

void kssymmetry(double Dp[][MAX_DIM+1],int dim)
//  the symmetry changing signs on odd modes is applied to the square
//   dim x dim matrix Dp
{
   for(int i=1;i<=dim;i++){
     int isign=1;
     if((i % 2) ==1) isign=-1;
     for(int j=1;j<=dim;j++){
       int jsign=1;
       if((j % 2) ==1) jsign=-1;
       Dp[i][j]*=(isign*jsign);
     }
   }
}

void kssymmmatrix(double R[][MAX_DIM+1],int dim)
// returns the symmetry matrix:
//  the symmetry is changing signs on odd modes
{
  for(int i=1;i<=dim;i++)
    for(int j=1;j<=dim;j++){
      if(i==j){
        if((i%2)==1) R[i][j]=-1.0;
        else R[i][j]=1.0;
      }
      else R[i][j]=0.0;
    }
}


void kssymmmatrix(interval R[][MAX_DIM+1],int dim)
// returns the symmetry matrix:
//  the symmetry is changing signs on odd modes
{
  for(int i=1;i<=dim;i++)
    for(int j=1;j<=dim;j++){
      if(i==j){
        if((i%2)==1) R[i][j]=-1.0;
        else R[i][j]=1.0;
      }
      else R[i][j]=0.0;
    }
}




interval ksl2norm(int normtype,interval a[],int n,interval c, int s)
//  for a set \Pi_{i=1}^n a[k] \times \Pi_{k=n+1}&\infty C/k^s
// representing the functions u=\Sum_{k=1}^{\infty} - 2a_k \sin(kx)
//   this function computes L_2 on (0,2\Pi) if normtype==0
//  or L_2 norm of gradient u, when normtype==1
{
  interval ret=0.0;
  if(normtype==0){  // L-2 norm
    for(int k=1;k<=n;k++){
      ret+=power(a[k],2);
    }
    ret+=power(c,2)/(power(interval(n),2*s-1)*(2*s-1));
  }
  if(normtype==1){ // L_2 norm for gradient
    for(int k=1;k<=n;k++){
      ret+=power(k*a[k],2);
    }
    ret+=power(c,2)/(power(interval(n),2*s-3)*(2*s-3));
  }
  ret=sqrt(ret);
  ret*=2*pi;
  return(Right(ret));
}



interval ksuniformnorm(int normtype,interval a[],int n,interval c, int s)
//  for a set \Pi_{i=1}^n a[k] \times \Pi_{k=n+1}&\infty C/k^s
// representing the functions u=\Sum_{k=1}^{\infty} - 2a_k \sin(kx)
//   this function computes C^0 on (0,2\Pi) if normtype==0
//  or C^0 norm of gradient u, when normtype==1
{
  interval ret=0.0;
  if(normtype==0){  // C^0 norm
    for(int k=1;k<=n;k++){
      ret+=abs(a[k]);
    }
    ret+=c/(power(interval(n),s-1)*(s-1));
  }
  if(normtype==1){ // C^0 norm for gradient
    for(int k=1;k<=n;k++){
      ret+=abs(k*a[k]);
    }
    ret+=c/(power(interval(n),s-2)*(s-2));
  }
  ret*=2;
  return(Right(ret));
}


int kssugestedorder(int dim,double mu,double h)
// the function computes a minimal <order> for given times step h - so that
// the numerical scheme becomes stable
// ----------------
// the function is based on the following considerations
//  for equation x'=Lx, where L < 0
//  we look at how many terms are need in the partial sum:
//        exp(Lh)= \sum_{i=0}^n \frac{(Lh)^i}{i!}
//   we want this partial sum + estimate for the remainder (for e^{|Lh|})
//   to be less than 1
{
  int order=0;
  int maxorder=30;
  double L=(dim*dim*(1 - mu*dim*dim));

  if(L <0){
    double expsum=1.0;
    double nextterm=1.0;
    double remestm;
    int enough=0;
    while(!enough){
      order++;
      nextterm*=((L*h)/order);
      expsum+=nextterm;
      if(fabs(L*h) < order + 1){
        remestm=fabs(nextterm)/(1-(fabs(L*h)/(order+1)));
        enough=(fabs(expsum) + remestm <1.0);
      }
      if(order > maxorder){
        order=0;
        enough=1;
      }
    }
  }
  return order;
}


int ksstablepar(int dim,double mu,double ih,int iorder,double *newh,int *neworder)
// the function tries to propose the values of the time step (*newh) and order (*neworder)
// which should lead to a stable nonrigorous computation using explicite methods
{
  *newh=ih;
  *neworder=iorder;
  int ord_suggested=kssugestedorder(dim,mu,ih);
  int found=((ord_suggested <=iorder) && (ord_suggested!=0));
  if(*neworder > MAX_ORDER) found=0;
  int maxiter=10;
  int i=1;
  while(!found && (i<=maxiter)){
    *newh = (*newh)/2;
    *neworder=kssugestedorder(dim,mu,*newh);
    if((*neworder > MAX_ORDER) || (*neworder==0)) found=0;
    else found=1;
    i++;
  }
  if(!found){
   *newh=0.5/(dim*dim*(1 - mu*dim*dim));
   *newh=fabs(*newh);
   *neworder=4;
  }
  return 1;
}


KS_Exp::KS_Exp()
{
  par_defined=0;
  tabledim=0;
  inittable=NULL;
  tabexp=NULL;
}

KS_Exp::KS_Exp(int pdimtable)
{
  par_defined=0;
  tabledim=pdimtable;
  inittable=new int[pdimtable+1];
  tabexp=new double[pdimtable+1];
}

KS_Exp::~KS_Exp()
{
  delete [] inittable;
  delete [] tabexp;
}


void KS_Exp::inflatetable(int pdim)
{
  if(pdim > tabledim){
    int* ninittable;
    ninittable=new int[pdim+1];

    double *ntabexp;
    ntabexp=new double[pdim+1];

    for(int i=1;i<=pdim;i++){
      ninittable[i]=0;
      ntabexp[i]=0;
    }
    if(par_defined){
      for(int i=1;i<=tabledim;i++){
        if(inittable[i]){
          ninittable[i]=1;
          ntabexp[i]=tabexp[i];
        }
      }
    }
    tabledim=pdim;
    delete [] inittable;
    delete [] tabexp;
    inittable=ninittable;
    tabexp=ntabexp;
  }
}

void KS_Exp::inittabexp(double ph,double pmu)
{
  if(par_defined){
    if((ph!=h) || (mu!=pmu)) par_defined=0;
  }

  if(!par_defined){
    h=ph;
    mu=pmu;
    for(int i=1;i<=tabledim;i++)
      inittable[i]=0;
    par_defined=1;
  }
}

interval  KS_Exp::e_to_k_l(interval& h, interval& pmu, int k)
{
  double retval;
  inittabexp(h.left_b(),pmu.left_b()); // the smallest values of h and \mu
                  // give rise to a maximum of exp(h*\lambda_k)
  int done=0;
  if(k <= tabledim){
    if(inittable[k]){
      retval=tabexp[k];
      done=1;
    }
  }
  if(!done){
    interval k2=interval(k)*k;
    interval lambda=(1 - pmu* k2)*k2;
    retval=exp(h*lambda).right_b();
    if(k > tabledim) inflatetable(k+5); // we add 5 to insure that the inflation
                   // will happen less often
    tabexp[k]=retval;
    inittable[k]=1;
  }
  return retval;
}



